!
! Module: fits_utils
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains some routines for dealing with keywords and columns of FITS files.
! 
! Comments: TODO:
!*********** Look through ALL software for unnecessary pointer allocations!
!*********** Look through ALL software and declare all ptrs => null()!
! 
! $Id: fits_utils.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- fred
module fits_utils

  use dal
  use exprcolgen

  use ssc_misc
!  use get_free_io_unit_mod
  use sort_mod ! used in sortTable.

  implicit none 

  private 

  public addOrOpenColumnDataReal32
  public addOrOpenColumnDataInt32
  public addOrOpenColumn
  public getRealKwd
  public getIntKwd
  public getStringKwd
  public copyAttribute
  public readColData
  public myAddColumn
  public minNonNullValue
  public maxNonNullValue
  public loadFitsArray
  public mergeTables
!  public saveOutImage
  public copyVitalArrayKwds
  public addTables
  public copyColumnBetweenTables
  public copyRowsBetweenTables
  public setValuesToNull
  public checkAndCopyKeywords
  public attributeValuesAreEqual
  public andKeywords
  public splitSetTabName
  public sortTable

!  interface saveOutImage
!    module procedure saveOutImageReal32, saveOutImageInt32
!  end interface

  interface andKeywords
    module procedure andKeywordsSetScalar, andKeywordsTabScalar
    module procedure andKeywordsSetVector, andKeywordsTabVector
  end interface

  interface setValuesToNull
    module procedure setColValuesToNull
  end interface

  interface copyAttribute
    module procedure copyAttribute_arrayToArray
    module procedure copyAttribute_arrayToSet
    module procedure copyAttribute_setToSet
    module procedure copyAttributeSetToArray
    module procedure copyAttributeTabToTab
  end interface

  interface readColData
    module procedure readColData_real32_wn
    module procedure readColData_real64_wn
    module procedure readColData_int8_wn
    module procedure readColData_int16_wn
    module procedure readColData_int32_wn
    module procedure readColData_real32
    module procedure readColData_real64
    module procedure readColData_int8
    module procedure readColData_int16
    module procedure readColData_int32
  end interface

  interface myAddColumn
    module procedure myAddColumn_int32
    module procedure myAddColumn_real32
    module procedure myAddColumn_real64
  end interface

  interface minNonNullValue
    module procedure minNonNullValue_real32
    module procedure minNonNullValue_real64
  end interface

  interface maxNonNullValue
    module procedure maxNonNullValueInt16
    module procedure maxNonNullValue_real32
    module procedure maxNonNullValue_real64
  end interface

  interface loadFitsArray
    module procedure loadFitsArray_1d_r32, loadFitsArray_1d_wa_r32
    module procedure loadFitsArray_2d_r32, loadFitsArray_2d_wa_r32
    module procedure loadFitsArray_3d_r32, loadFitsArray_3d_wa_r32
    module procedure loadFitsArray_4d_r32, loadFitsArray_4d_wa_r32
    module procedure loadFitsArray_1d_r64, loadFitsArray_1d_wa_r64
    module procedure loadFitsArray_2d_r64, loadFitsArray_2d_wa_r64
    module procedure loadFitsArray_3d_r64, loadFitsArray_3d_wa_r64
    module procedure loadFitsArray_4d_r64, loadFitsArray_4d_wa_r64
    module procedure loadFitsArray2dBoolean
  end interface

contains 

  !---------------------------------------------------------------------
  subroutine splitSetTabName(setTabName, setName, tabName)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in)  :: setTabName
    character(*), intent(out) :: setName, tabName

    integer(int16) :: posOfColon
    type(DataSetT) :: set

    !-------------------------------------------------------------------
    ! Start

    posOfColon = index(setTabName, ':')
    if (posOfColon .le. 0) then ! no colon in the set-tab name. User
    ! has forgotten to include the table name.
      call warning('missingBlockName', "No colon+block name was &
        &found after the dataset name. I'm going to use the first block &
        &in the dataset. No cause for alarm - provided that this &
        &really is the block you want!")
      setName = setTabName
      set = dataSet(setName, READ)
      tabName = name(block(set, 0))
      call release(set)

    else
      setName = setTabName(:posOfColon - 1)
      tabName = setTabName(posOfColon + 1:)
    end if

  end subroutine splitSetTabName

  !---------------------------------------------------------------------
  subroutine loadFitsArray2dBoolean(setName, fitsArray, blockNumber)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: setName
    integer, intent(in), optional :: blockNumber
    logical, pointer :: fitsArray(:,:)

    type(DataSetT) :: set
    type(ArrayT) :: inArray
    integer :: localBlockNumber
    integer(int32), dimension(:), pointer ::&
      dimVector
!!!    real(single), dimension(:,:,:), pointer :: cube
!!!    real(single), dimension(:,:,:,:), pointer :: hyperCube
    integer :: numDimensions, arrayShape(4)
!!!, biggestDimensions(2, 1), imageShape(2)
    logical :: mask(4)
    integer(int8), dimension(:,:), pointer ::&
      arrayDataInt8
    integer(int16), dimension(:,:), pointer ::&
      arrayDataInt16
    integer(int32), dimension(:,:), pointer ::&
      arrayDataInt32
    real(single), dimension(:,:), pointer ::&
      arrayDataReal32
    real(double), dimension(:,:), pointer ::&
      arrayDataReal64
!    logical(bool), dimension(:,:), pointer ::&
!      arrayDataBoolean

    !-------------------------------------------------------------------
    ! Start

    if (present(blockNumber)) then
      localBlockNumber = blockNumber
    else
      localBlockNumber = 0
    end if

    set = dataSet(setName, READ)
    inArray = array(set, localBlockNumber, READ)
    dimVector => dimensions(inArray)
    numDimensions = size(dimVector)
    arrayShape = 0 ! vector op, default
    mask = .true.  ! vector op, default

    select case(numDimensions)
    case(1)
call error('arrayRankTooSmall', '')
    case(2)
      allocate(fitsArray(dimVector(0), dimVector(1)))

      fitsArray = .false.
      if (dataType(inArray) .eq. INTEGER8) then
        arrayDataInt8 => int8Array2Data(inArray)
        where(arrayDataInt8 > 0) fitsArray = .true.
      else if (dataType(inArray) .eq. INTEGER16) then
        arrayDataInt16 => int16Array2Data(inArray)
        where(arrayDataInt16 > 0) fitsArray = .true.
      else if (dataType(inArray) .eq. INTEGER32) then
        arrayDataInt32 => int32Array2Data(inArray)
        where(arrayDataInt32 > 0) fitsArray = .true.
      else if (dataType(inArray) .eq. REAL32) then
        arrayDataReal32 => real32Array2Data(inArray)
        where(arrayDataReal32 > 0) fitsArray = .true.
!      else if (dataType(inArray) .eq. BOOLEAN) then
!        arrayDataBoolean => boolArray2Data(inArray)
!        fitsArray = arrayDataBoolean
      else ! this gives a DAL error if data type is none of the above.
        arrayDataReal64 => real64Array2Data(inArray)
        where(arrayDataReal64 > 0) fitsArray = .true.
      end if

    case(3)
!      call loadFitsArray_3d_wa_r32(inArray, cube)!!!, dataType)
!      arrayShape(:3) = shape(cube)
!if (count(arrayShape .gt. 1) .gt. 2) call error('badCubeShape','')
!      biggestDimensions(1,:) = maxloc(arrayShape)
!      mask(biggestDimensions(1,1)) = .false.
!      biggestDimensions(2,:) = maxloc(arrayShape, mask)
!      imageShape(1) = size(cube, biggestDimensions(1,1))
!      imageShape(2) = size(cube, biggestDimensions(2,1))
!      allocate(fitsArray(imageShape(1), imageShape(2)))
!      fitsArray = reshape(cube, imageShape)
call error('notSupported', '')

    case(4)
!      call loadFitsArray_4d_wa_r32(inArray, hyperCube)!!!, dataType)
!      arrayShape = shape(hyperCube)
!if (count(arrayShape .gt. 1) .gt. 2) call error('badHyperCubeShape','')
!      biggestDimensions(1,:) = maxloc(arrayShape)
!      mask(biggestDimensions(1,1)) = .false.
!      biggestDimensions(2,:) = maxloc(arrayShape, mask)
!      imageShape(1) = size(cube, biggestDimensions(1,1))
!      imageShape(2) = size(cube, biggestDimensions(2,1))
!      allocate(fitsArray(imageShape(1), imageShape(2)))
!      fitsArray = reshape(hyperCube, imageShape)
call error('notSupported', '')

    case default
call error('arrayRankTooBig', '')
    end select

call release(set)
!*** very important! Make sure these releases are in the subroutines when you get around to putting them in a library.

  end subroutine loadFitsArray2dBoolean

  !---------------------------------------------------------------------
  subroutine addTables(masterTab, slaveTab, indexColName, indexValue)
  !---------------------------------------------------------------------
  ! This takes the contents of slaveTab and appends them to the end of
  ! masterTab. (MasterTab is permitted initially to contain 0 rows.) If
  ! slaveTab has columns that masterTab doesn't have, columns of these
  ! names are added to masterTab. Any non-slaveTab rows of these new
  ! columns should be set to null (numeric or logical) or empty (string)
  ! but this is not done properly as yet. Same goes for any rows of
  ! slaveTab for columns not present in slaveTab.
  !
  ! A column with a name specified by the argument indexColName is used
  ! to keep track of the different components within masterTab. If this
  ! was not previously present within masterTab, it is created (as type
  ! int16). If a column of this name is already present within the
  ! slaveTab, a warning is generated. Any values of such a column will
  ! not be copied to indexColName in masterTab.
  !
  ! If a column of this name is present within masterTab, but the data
  ! type of this column is not int16, an error is generated.
  !
  ! If indexValue is not supplied, and indexColName existed previously
  ! in masterTab, the index value for the present slaveCol rows is
  ! calculated by adding 1 to the largest previous value of indexCol.
  ! If indexValue is not supplied, and indexCol is but newly added in
  ! masterTab, it is set to null (or rather, should be) for any
  ! previous non-slaveTab rows and 0 for the slaveTab rows.
  !
  ! If indexValue is supplied, naturally this is the value used; however
  ! a warning is given if it matches any previous value.
  !---------------------------------------------------------------------

    implicit none

    type(TableT),   intent(in)           :: masterTab, slaveTab
    character(*),   intent(in), optional :: indexColName
    integer(int16), intent(in), optional :: indexValue

    integer(int32) :: numMasterRows, numSlaveRows, colNum
    type(ColumnT) :: slaveCol, masterCol, indexCol
    integer(int16), pointer :: index(:)
    integer(int16) :: maxValue!, fred
    logical :: allRowsNull
logical :: testflag = .true.

    !-------------------------------------------------------------------
    ! Start:

    numMasterRows = numberOfRows(masterTab)
    numSlaveRows  = numberOfRows(slaveTab)
    call insertRows(masterTab, numMasterRows, numSlaveRows)
if (testflag) print *, 'inserted', numSlaveRows, 'rows at starting row', numMasterRows, 'in master table.'

!**** now call copyRowsBetweenTables instead of the below???

    ! For each column in slaveTab, (i) check to see if masterTab has it;
    ! (ii) confirm data types are the same if so, create new column in
    ! masterTab if not; (iii) copy over relevant slaveTab data.

    do colNum = 0, numberOfColumns(slaveTab) - 1
      slaveCol = column(slaveTab, colNum)
      if (present(indexColName)) then
        if (name(slaveCol) == indexColName) then
          call warning('indexColAlreadyInSourceTab','')
    cycle
        end if
      end if

      if (hasColumn(masterTab, name(slaveCol))) then
        masterCol = column(masterTab, name(slaveCol))

        if (dataType(slaveCol) .ne. dataType(masterCol))&
call error('mixedDataTypes','')

      else ! master table doesn't have such a column.
        masterCol = addColumn(masterTab, name(slaveCol), dataType(slaveCol))
        if (dataType(masterCol) .ne. STRING&
        .and. dataType(masterCol) .ne. BOOLEAN&
        .and. .not. nullDefined(masterCol))&
          call setNullValue(masterCol, -999)
        call setValuesToNull(masterCol) ! sets everything to null.
      end if

!if (dataType(slaveCol) .ne. dataType(masterCol)) then
!  print *, 'name(slaveCol), dataType(slaveCol), dataType(masterCol): '&
!  , trim(name(slaveCol)), dataType(slaveCol), dataType(masterCol)
!call error('mixedDataTypes','')
!end if
      call copyColumnBetweenTables(slaveCol, 0, numSlaveRows&
        , masterCol, numMasterRows)

      call release(slaveCol)
      call release(masterCol)
    end do

    ! If master has a col that slave does not, add nulls to relevant
    ! rows in master:
    do colNum = 0, numberOfColumns(masterTab) - 1
      masterCol = column(masterTab, colNum)

      if (present(indexColName)) then
    if (name(masterCol) == indexColName) cycle
      end if

      if (.not. hasColumn(slaveTab, name(masterCol))) then
        if (dataType(masterCol) .ne. STRING&
        .and. dataType(masterCol) .ne. BOOLEAN&
        .and. .not. nullDefined(masterCol))&
          call setNullValue(masterCol, -999)
        call setValuesToNull(masterCol, numMasterRows, numSlaveRows)
        ! sets everything to null.
      end if
    end do

    if (present(indexColName)) then
      ! If index col doesn't exist in master, create it
      if (.not. hasColumn(masterTab, indexColName)) then
        indexCol = addColumn(masterTab, indexColName, INTEGER16)
        call setNullValue(indexCol, -999)
        call setValuesToNull(indexCol)

        index => int16Data(indexCol)
        if (present(indexValue)) then
!print *, 'setting rows', numMasterRows, ' to'&
!, numMasterRows + numSlaveRows - 1, ' of column '//trim(indexColName)&
!//' to', indexValue
          index(numMasterRows:numMasterRows + numSlaveRows - 1) = indexValue
        else
!print *, 'setting rows', numMasterRows, ' to'&
!, numMasterRows + numSlaveRows - 1, ' of column '//trim(indexColName)&
!//' to', 0
          index(numMasterRows:numMasterRows + numSlaveRows - 1) = 0
        end if

      else if (dataType(column(masterTab, indexColName)) .ne. INTEGER16) then
call error('indexColWrongDataType','')

      else
        indexCol = column(masterTab, indexColName)
        if (.not. nullDefined(indexCol)) call setNullValue(indexCol, -999)

        index => int16Data(indexCol)
        if (present(indexValue)) then
          if (any(index == indexValue))&
            call warning('indexValueAlreadyPresent','')
          index(numMasterRows:numMasterRows + numSlaveRows - 1) = indexValue

        else
          call maxNonNullValue(masterTab, indexColName, maxValue&
            , allRowsNull, int(0, int32), numMasterRows - 1)
          if (allRowsNull) then
!print *, 'setting rows', numMasterRows, ' to'&
!, numMasterRows + numSlaveRows - 1, ' of column '//trim(indexColName)&
!//' to', 0
            index(numMasterRows:numMasterRows + numSlaveRows - 1) = 0
          else
!print *, 'setting rows', numMasterRows, ' to'&
!, numMasterRows + numSlaveRows - 1, ' of column '//trim(indexColName)&
!//' to (maxvalue + 1)', maxValue + 1
            index(numMasterRows:numMasterRows + numSlaveRows - 1) = maxValue + 1
          end if
        end if
      end if

      call release(indexCol)
    end if

  end subroutine addTables

  !---------------------------------------------------------------------
  subroutine addTables_20021001(masterTab, slaveTab, indexColName, indexValue)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: masterTab, slaveTab
    character(*), intent(in) :: indexColName
    integer(int16), intent(in) :: indexValue

    integer(int32) :: numMasterRows, numSlaveRows, colNum
    type(ColumnT) :: slaveCol, masterCol, indexCol
    integer(int16), pointer :: index(:)
logical :: testflag = .true.

    !-------------------------------------------------------------------
    ! Start:

    numMasterRows = numberOfRows(masterTab)
    numSlaveRows  = numberOfRows(slaveTab)
    call insertRows(masterTab, numMasterRows, numSlaveRows)
if (testflag) print *, 'inserted', numSlaveRows, 'rows at starting row', numMasterRows, 'in master table.'

!**** now call copyRowsBetweenTables instead of the below???

    do colNum = 0, numberOfColumns(slaveTab) - 1
      slaveCol = column(slaveTab, colNum)
    if (name(slaveCol) == indexColName) cycle

      if (hasColumn(masterTab, name(slaveCol))) then
        masterCol = column(masterTab, name(slaveCol))

        if (dataType(slaveCol) .ne. dataType(masterCol))&
call error('mixedDataTypes','')

      else ! master table doesn't have such a column.
        masterCol = addColumn(masterTab, name(slaveCol), dataType(slaveCol))
        call setValuesToNull(masterCol) ! sets everything to null.
      end if

!if (dataType(slaveCol) .ne. dataType(masterCol)) then
!  print *, 'name(slaveCol), dataType(slaveCol), dataType(masterCol): '&
!  , trim(name(slaveCol)), dataType(slaveCol), dataType(masterCol)
!call error('mixedDataTypes','')
!end if
      call copyColumnBetweenTables(slaveCol, 0, numSlaveRows&
        , masterCol, numMasterRows)

      call release(slaveCol)
      call release(masterCol)
    end do

    ! If master has a col that slave does not, add nulls to relevant
    ! rows in master:
    do colNum = 0, numberOfColumns(masterTab) - 1
      masterCol = column(masterTab, colNum)
    if (name(masterCol) == indexColName) cycle

      if (.not. hasColumn(slaveTab, name(masterCol))) then
        call setValuesToNull(masterCol, numMasterRows, numSlaveRows)
        ! sets everything to null.
      end if
    end do

    ! If index col doesn't exist in master, create it
    if (.not. hasColumn(masterTab, indexColName)) then
      indexCol = addColumn(masterTab, indexColName, INTEGER16)
!*** set to null or blank ??? Prob not nec here.
    else if (dataType(column(masterTab, indexColName)) .ne. INTEGER16) then
      call deleteColumn(masterTab, indexColName)
      indexCol = addColumn(masterTab, indexColName, INTEGER16)
!*** set to null or blank ??? Prob not nec here.
    else
      indexCol = column(masterTab, indexColName)
    end if

    ! Load index col with indexValue
    index => int16Data(indexCol)
    index(numMasterRows:numMasterRows + numSlaveRows - 1) = indexValue
    call release(indexCol)

  end subroutine addTables_20021001

  !---------------------------------------------------------------------
  subroutine copyRowsBetweenTables(fromTab, fromRowStart, numFromRowsToCopy&
  , toTab, toRowStart)
  !---------------------------------------------------------------------
!*** put into fits_utils

    implicit none

    type(TableT),   intent(in) ::&
      fromTab, toTab
    integer(int32), intent(in) ::&
      fromRowStart, numFromRowsToCopy, toRowStart

    integer(int32) :: numFromRows, numToRows, colNum
    type(ColumnT) :: fromCol, toCol
logical :: testflag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testflag) print *, 'copy rows:'
if (testflag) print *, '  fromRowStart, numFromRowsToCopy, toRowStart: '&
, fromRowStart, numFromRowsToCopy, toRowStart

    numFromRows = numberOfRows(fromTab)
    numToRows   = numberOfRows(toTab)
if (fromRowStart < 0) call error('fromStartRowTooLow','')
if (toRowStart   < 0) call error('toStartRowTooLow','')
    if (fromRowStart + numFromRowsToCopy > numFromRows)&
call error('tooFewFromRows','')
if (toRowStart + numFromRowsToCopy > numToRows) call error('tooFewToRows','')

    ! If fromTab has columns that toTab does not, add these columns to toTab:

    do colNum = 0, numberOfColumns(fromTab) - 1
      fromCol = column(fromTab, colNum)

      if (hasColumn(toTab, name(fromCol))) then
        toCol = column(toTab, name(fromCol))

if (dataType(fromCol) .ne. dataType(toCol)) call error('mixedDataTypes','')

      else ! toTab doesn't have such a column.
        toCol = addColumn(toTab, name(fromCol), dataType(fromCol))
        call setValuesToNull(toCol) ! sets everything to null.
      end if

      call copyColumnBetweenTables(fromCol, fromRowStart, numFromRowsToCopy&
        , toCol, toRowStart)

      call release(fromCol)
      call release(toCol)
    end do

  end subroutine copyRowsBetweenTables

  !---------------------------------------------------------------------
  subroutine copyColumnBetweenTables(fromCol, fromColStartRow, numFromRows&
  , toCol, toColStartRow)
  !---------------------------------------------------------------------
!*** put into fits_utils

    implicit none

    type(ColumnT), intent(in) :: fromCol, toCol
    integer(int32), intent(in) :: fromColStartRow, numFromRows, toColStartRow
logical :: testflag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testflag) print *, '  copy row', fromColStartRow, 'to row'&
, fromColStartRow+numFromRows-1, ' of column name '//trim(name(fromCol))&
//' of table A to rows', toColStartRow, '-', toColStartRow+numFromRows-1&
, 'of table B.'

    if (fromColStartRow < 0)&
call error('fromStartRowTooLow','')
    if (toColStartRow < 0)&
call error('toStartRowTooLow','')
    if (numberOfRows(fromCol) < fromColStartRow + numFromRows)&
call error('tooFewFromRows','')
    if (numberOfRows(toCol) < toColStartRow + numFromRows)&
call error('tooFewToRows','')

    select case(dataType(fromCol))
    case(INTEGER8)
      call copyColsTweenInt8(fromCol, fromColStartRow, numFromRows&
        , toCol, toColStartRow)
    case(INTEGER16)
      call copyColsTweenInt16(fromCol, fromColStartRow, numFromRows&
        , toCol, toColStartRow)
    case(INTEGER32)
      call copyColsTweenInt32(fromCol, fromColStartRow, numFromRows&
        , toCol, toColStartRow)
    case(REAL32)
      call copyColsTweenSingle(fromCol, fromColStartRow, numFromRows&
        , toCol, toColStartRow)
    case(REAL64)
      call copyColsTweenDouble(fromCol, fromColStartRow, numFromRows&
        , toCol, toColStartRow)
    case(STRING)
      call copyColsTweenString(fromCol, fromColStartRow, numFromRows&
        , toCol, toColStartRow)
    case(BOOLEAN)
      call copyColsTweenBool(fromCol, fromColStartRow, numFromRows&
        , toCol, toColStartRow)
    case default
call error('badColDataType','')
    end select

  end subroutine copyColumnBetweenTables

  !---------------------------------------------------------------------
  subroutine copyColsTweenInt8(fromCol, fromColStartRow, numFromRows&
  , toCol, toColStartRow)
  !---------------------------------------------------------------------
  ! This should NOT be made a public subroutine.
  !
  ! The checks on size of the columns have already been done in
  ! copyColumnBetweenTables().

    implicit none

    type(ColumnT), intent(in) :: fromCol, toCol
    integer(int32), intent(in) :: fromColStartRow, numFromRows, toColStartRow

    integer(int8), pointer :: toVector(:), fromVector(:)

    !-------------------------------------------------------------------
    ! Start

    fromVector => int8Data(fromCol)
    toVector   => int8Data(toCol)
    toVector(toColStartRow:toColStartRow + numFromRows - 1)&
      = fromVector(fromColStartRow:fromColStartRow + numFromRows - 1)
    call release(fromCol)
    call release(toCol)

  end subroutine copyColsTweenInt8

  !---------------------------------------------------------------------
  subroutine copyColsTweenInt16(fromCol, fromColStartRow, numFromRows&
  , toCol, toColStartRow)
  !---------------------------------------------------------------------
  ! This should NOT be made a public subroutine.
  !
  ! The checks on size of the columns have already been done in
  ! copyColumnBetweenTables().

    implicit none

    type(ColumnT), intent(in) :: fromCol, toCol
    integer(int32), intent(in) :: fromColStartRow, numFromRows, toColStartRow

    integer(int16), pointer :: toVector(:), fromVector(:)

    !-------------------------------------------------------------------
    ! Start

    fromVector => int16Data(fromCol)
    toVector   => int16Data(toCol)
    toVector(toColStartRow:toColStartRow + numFromRows - 1)&
      = fromVector(fromColStartRow:fromColStartRow + numFromRows - 1)
    call release(fromCol)
    call release(toCol)

  end subroutine copyColsTweenInt16

  !---------------------------------------------------------------------
  subroutine copyColsTweenInt32(fromCol, fromColStartRow, numFromRows&
  , toCol, toColStartRow)
  !---------------------------------------------------------------------
  ! This should NOT be made a public subroutine.
  !
  ! The checks on size of the columns have already been done in
  ! copyColumnBetweenTables().

    implicit none

    type(ColumnT), intent(in) :: fromCol, toCol
    integer(int32), intent(in) :: fromColStartRow, numFromRows, toColStartRow

    integer(int32), pointer :: toVector(:), fromVector(:)

    !-------------------------------------------------------------------
    ! Start

    fromVector => int32Data(fromCol)
    toVector   => int32Data(toCol)
    toVector(toColStartRow:toColStartRow + numFromRows - 1)&
      = fromVector(fromColStartRow:fromColStartRow + numFromRows - 1)
    call release(fromCol)
    call release(toCol)

  end subroutine copyColsTweenInt32

  !---------------------------------------------------------------------
  subroutine copyColsTweenSingle(fromCol, fromColStartRow, numFromRows&
  , toCol, toColStartRow)
  !---------------------------------------------------------------------
  ! This should NOT be made a public subroutine.
  !
  ! The checks on size of the columns have already been done in
  ! copyColumnBetweenTables().

    implicit none

    type(ColumnT), intent(in) :: fromCol, toCol
    integer(int32), intent(in) :: fromColStartRow, numFromRows, toColStartRow

    real(single), pointer :: toVector(:), fromVector(:)

    !-------------------------------------------------------------------
    ! Start

    fromVector => real32Data(fromCol)
    toVector   => real32Data(toCol)
    toVector(toColStartRow:toColStartRow + numFromRows - 1)&
      = fromVector(fromColStartRow:fromColStartRow + numFromRows - 1)
    call release(fromCol)
    call release(toCol)

  end subroutine copyColsTweenSingle

  !---------------------------------------------------------------------
  subroutine copyColsTweenDouble(fromCol, fromColStartRow, numFromRows&
  , toCol, toColStartRow)
  !---------------------------------------------------------------------
  ! This should NOT be made a public subroutine.
  !
  ! The checks on size of the columns have already been done in
  ! copyColumnBetweenTables().

    implicit none

    type(ColumnT), intent(in) :: fromCol, toCol
    integer(int32), intent(in) :: fromColStartRow, numFromRows, toColStartRow

    real(double), pointer :: toVector(:), fromVector(:)

    !-------------------------------------------------------------------
    ! Start

    fromVector => real64Data(fromCol)
    toVector   => real64Data(toCol)
    toVector(toColStartRow:toColStartRow + numFromRows - 1)&
      = fromVector(fromColStartRow:fromColStartRow + numFromRows - 1)
    call release(fromCol)
    call release(toCol)

  end subroutine copyColsTweenDouble

  !---------------------------------------------------------------------
  subroutine copyColsTweenString(fromCol, fromColStartRow, numFromRows&
  , toCol, toColStartRow)
  !---------------------------------------------------------------------
  ! This should NOT be made a public subroutine.
  !
  ! The checks on size of the columns have already been done in
  ! copyColumnBetweenTables().

    implicit none

    type(ColumnT), intent(in) :: fromCol, toCol
    integer(int32), intent(in) :: fromColStartRow, numFromRows, toColStartRow

    integer(int32) :: row

    !-------------------------------------------------------------------
    ! Start

    do row = 0, numFromRows - 1
      call setStringCell(toCol, row + toColStartRow&
        , stringCell(fromCol, row + fromColStartRow))
    end do

  end subroutine copyColsTweenString

  !---------------------------------------------------------------------
  subroutine copyColsTweenBool(fromCol, fromColStartRow, numFromRows&
  , toCol, toColStartRow)
  !---------------------------------------------------------------------
  ! This should NOT be made a public subroutine.
  !
  ! The checks on size of the columns have already been done in
  ! copyColumnBetweenTables().

    implicit none

    type(ColumnT), intent(in) :: fromCol, toCol
    integer(int32), intent(in) :: fromColStartRow, numFromRows, toColStartRow

    logical(bool), pointer :: toVector(:), fromVector(:)

    !-------------------------------------------------------------------
    ! Start

    fromVector => boolData(fromCol)
    toVector   => boolData(toCol)
    toVector(toColStartRow:toColStartRow + numFromRows - 1)&
      = fromVector(fromColStartRow:fromColStartRow + numFromRows - 1)
    call release(fromCol)
    call release(toCol)

  end subroutine copyColsTweenBool

  !---------------------------------------------------------------------
  subroutine setColValuesToNull(col, startRow, numRowsToNull)
  !---------------------------------------------------------------------
  ! Sets the column entries to nulls where possible, blanks or .false. otherwise. For integer columns, if a null value has not been defined by calling setNullValue, an error will result.

    implicit none

    type(ColumnT), intent(in) :: col
    integer(int32), intent(in), optional :: startRow, numRowsToNull

    integer(int32) :: finisRow, localStartRow, numRows, row
    logical(bool), pointer :: boolPointer(:)

    !-------------------------------------------------------------------
    ! Start

    if (present(startRow)) then
if (startRow < 0) call error('negativeStartRow','ssclib:')
      localStartRow = startRow
    else
      localStartRow = 0
    end if

    numRows = numberOfRows(col)

    if (present(numRowsToNull)) then
      finisRow = localStartRow + numRowsToNull - 1
if (finisRow > numRows) call error('numLastRowTooLarge','ssclib:')
    else
      finisRow = numRows - 1
    end if

    if (dataType(col) .eq. BOOLEAN) then
      boolPointer => boolData(col)
      boolPointer(localStartRow:finisRow) = .false.
deallocate(boolPointer)

    else if (dataType(col) .eq. STRING) then
      do row = localStartRow, finisRow
        call setStringCell(col, row, '')
      end do

    else
      do row = localStartRow, finisRow
        call setToNull(col, row)
      end do
    end if

  end subroutine setColValuesToNull

  !---------------------------------------------------------------------
  subroutine checkAndCopyKeywords(fromTabs, toTab, keywordNames&
  , keywordIsUniform)
  !---------------------------------------------------------------------
  ! This sub checks each keyword in the list to see whether it is present and of the same value in all the input tables. If so, it copies it into the output table.

!*** see also the routine(s) andKeywords().

    implicit none

    type(TableT), intent(in) :: fromTabs(:), toTab
    character(*), intent(in) :: keywordNames(:)
    logical, intent(out), optional ::&
      keywordIsUniform(size(keywordNames))

    integer(int16) :: numTabs, numKeywords, keywordNum, tabNum
    logical :: localKwdIsUniform(size(keywordNames))!!! = .false.
    type(AttributeT) :: attA, attB

    !-------------------------------------------------------------------
    ! Start

localKwdIsUniform = .false.
    numTabs = size(fromTabs)
    numKeywords = size(keywordNames)

    KEYWORD: do keywordNum = 1, numKeywords
      localKwdIsUniform(keywordNum) = .true. ! temporary default.

      ! Note that there are two ways in which a keyword may be uniform:
      ! either it is present in all tables, and of the same value; or it
      ! is not present in all tables. These two cases are tested separately
      ! in the two branches of the following 'if'.

      if (hasAttribute(fromTabs(1), keywordNames(keywordNum))) then
        do tabNum = 2, numTabs
          if (.not. hasAttribute(fromTabs(tabNum), keywordNames(keywordNum)))&
          then
            localKwdIsUniform(keywordNum) = .false.
    cycle KEYWORD
          end if

          attA = attribute(fromTabs(1),      keywordNames(keywordNum))
          attB = attribute(fromTabs(tabNum), keywordNames(keywordNum))
          if (.not. attributeValuesAreEqual(attA, attB)) then
            localKwdIsUniform(keywordNum) = .false.
    cycle KEYWORD
          end if
        end do

        if (localKwdIsUniform(keywordNum)) then ! copy it:
          call copyAttribute(fromTabs(1), keywordNames(keywordNum), toTab)
        end if

      else ! table 1 doesn't have the attribute.
        do tabNum = 2, numTabs
          if (hasAttribute(fromTabs(tabNum), keywordNames(keywordNum))) then
            localKwdIsUniform(keywordNum) = .false.
    cycle KEYWORD
          end if
        end do
      end if
    end do KEYWORD

    if (present(keywordIsUniform)) keywordIsUniform = localKwdIsUniform

  end subroutine checkAndCopyKeywords

  !---------------------------------------------------------------------
  logical function attributeValuesAreEqual(attA, attB)
  !---------------------------------------------------------------------

!*** put into fits_utils

    implicit none

    type(AttributeT), intent(in) :: attA, attB

    !-------------------------------------------------------------------
    ! Start

    attributeValuesAreEqual = .false. ! default

  if (dataType(attA) .ne. dataType(attB)) return

    select case(dataType(attA))
    case(INTEGER8)
      if (int8Attribute(attA) == int8Attribute(attB))&
        attributeValuesAreEqual = .true.
    case(INTEGER16)
      if (int16Attribute(attA) == int16Attribute(attB))&
        attributeValuesAreEqual = .true.
    case(INTEGER32)
      if (int32Attribute(attA) == int32Attribute(attB))&
        attributeValuesAreEqual = .true.
    case(REAL32)
      if (real32Attribute(attA) == real32Attribute(attB))&
        attributeValuesAreEqual = .true.
    case(REAL64)
      if (real64Attribute(attA) == real64Attribute(attB))&
        attributeValuesAreEqual = .true.
    case(STRING)
      if (stringAttribute(attA) == stringAttribute(attB))&
        attributeValuesAreEqual = .true.
    case(BOOLEAN)
      if (booleanAttribute(attA) .eqv. booleanAttribute(attB))&
        attributeValuesAreEqual = .true.
    end select

  end function attributeValuesAreEqual

  !---------------------------------------------------------------------
  subroutine getRealKwd(keyword, inDataSet, imageArray, value, status)
  !---------------------------------------------------------------------
  ! Gets FITS real32 or 64 attribute from main dataset or primary array.
  !
  ! These routines were adapted from Clive Page's implot.

    implicit none

    character*(*),  intent(in)            :: keyword
    real(double),   intent(out)           :: value
    integer,        intent(out), optional :: status
    type(DataSetT), intent(in)            :: inDataSet
    type(ArrayT),   intent(in)            :: imageArray

    integer :: localStatus
    logical :: fatalFail

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0
    value = 0.0d0 ! default

    if (present(status)) then
      fatalFail = .false.
    else
      fatalFail = .true.
    end if

    if(hasAttribute(inDataSet, keyword)) then
      if (dataType(attribute(inDataSet, keyword)) .eq. REAL32) then
        value = real32Attribute(inDataSet, keyword)
      else
        if (fatalFail&
        .or. dataType(attribute(inDataSet, keyword)) .eq. REAL64) then
          value = real64Attribute(inDataSet, keyword)
        else
          localStatus = -1
        end if
      end if
    else if (fatalFail .or. hasAttribute(imageArray, keyword)) then
      if (dataType(attribute(imageArray, keyword)) .eq. REAL32) then
        value = real32Attribute(imageArray, keyword)
      else
        if (fatalFail&
        .or. dataType(attribute(imageArray, keyword)) .eq. REAL64) then
          value = real64Attribute(imageArray, keyword)
        else
          localStatus = -2
        end if
      end if
    else
      localStatus = -3
    end if

    if (present(status)) status = localStatus

  end subroutine getRealKwd

  !---------------------------------------------------------------------
  subroutine getIntKwd(keyword, inDataSet, imageArray, value, status)
  !---------------------------------------------------------------------
  ! Gets FITS int8, 16 or 32 attribute from main dataset or primary array.
  !
  ! These routines were adapted from Clive Page's implot.

    implicit none

    character*(*),  intent(in)            :: keyword
    integer(int32), intent(out)           :: value
    integer,        intent(out), optional :: status
    type(DataSetT), intent(in)            :: inDataSet
    type(ArrayT),   intent(in)            :: imageArray

    integer :: localStatus
    logical :: fatalFail

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0
    value = 0 ! default.

    if (present(status)) then
      fatalFail = .false.
    else
      fatalFail = .true.
    end if

    if(hasAttribute(inDataSet, keyword)) then
      if (dataType(attribute(inDataSet, keyword)) .eq. INTEGER8) then
        value = int8Attribute(inDataSet, keyword)
      else if (dataType(attribute(inDataSet, keyword)) .eq. INTEGER16) then
        value = int16Attribute(inDataSet, keyword)
      else
        if (fatalFail&
        .or. dataType(attribute(inDataSet, keyword)) .eq. INTEGER32) then
          value = int32Attribute(inDataSet, keyword)
        else
          localStatus = -1
        end if
      end if
    else if (fatalFail .or. hasAttribute(imageArray, keyword)) then
      if (dataType(attribute(imageArray, keyword)) .eq. INTEGER8) then
        value = int8Attribute(imageArray, keyword)
      else if (dataType(attribute(imageArray, keyword)) .eq. INTEGER16) then
        value = int16Attribute(imageArray, keyword)
      else
        if (fatalFail&
        .or. dataType(attribute(imageArray, keyword)) .eq. INTEGER32) then
          value = int32Attribute(imageArray, keyword)
        else
          localStatus = -2
        end if
      end if
    else
      localStatus = -3
    end if

    if (present(status)) status = localStatus

  end subroutine getIntKwd

  !---------------------------------------------------------------------
  subroutine getStringKwd(keyword, inDataSet, imageArray, value, status)
  !---------------------------------------------------------------------
  ! Gets FITS string attribute from main dataset or primary array.
  !
  ! These routines were adapted from Clive Page's implot.

!****Here it is assumed that a keyword of any type can be accessed as string?? Should check this.

    implicit none

    character*(*),  intent(in)            :: keyword
    character*(*),  intent(out)           :: value
    integer,        intent(out), optional :: status
    type(DataSetT), intent(in)            :: inDataSet
    type(ArrayT),   intent(in)            :: imageArray

    integer :: localStatus
    logical :: fatalFail

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0
    value = "??" ! default.

    if (present(status)) then
      fatalFail = .false.
    else
      fatalFail = .true.
    end if

    if(hasAttribute(inDataSet, keyword)) then
!print *, 'hasAttribute(inDataSet, keyword)'
      value = stringAttribute(inDataSet, keyword)
    else if (fatalFail .or. hasAttribute(imageArray, keyword)) then
!print *, 'check the array'
      value = stringAttribute(imageArray, keyword)
    else
      localStatus = -3
    end if

    if (present(status)) status = localStatus

  end subroutine getStringKwd

  !---------------------------------------------------------------------
  subroutine copyAttribute_arrayToArray(inArray, kwdName, outArray, status)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT), intent(in) :: inArray, outArray
    character(*), intent(in) :: kwdName
    integer, intent(out), optional :: status

    !-------------------------------------------------------------------
    ! Start

    if (present(status) .and. .not. hasAttribute(inArray, kwdName)) then
      status = -1
  return
    end if

    call setAttribute(outArray, attribute(inArray, kwdName))
    if (present(status)) status = 0

  end subroutine copyAttribute_arrayToArray

  !---------------------------------------------------------------------
  subroutine copyAttribute_arrayToSet(inArray, kwdName, outSet, status)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT), intent(in) :: inArray
    type(DataSetT), intent(in) :: outSet
    character(*), intent(in) :: kwdName
    integer, intent(out), optional :: status

    !-------------------------------------------------------------------
    ! Start

    if (present(status) .and. .not. hasAttribute(inArray, kwdName)) then
      status = -1
  return
    end if

    call setAttribute(outSet, attribute(inArray, kwdName))
    if (present(status)) status = 0

  end subroutine copyAttribute_arrayToSet

  !---------------------------------------------------------------------
  subroutine copyAttributeSetToArray(inSet, kwdName, outArray, status)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT), intent(in) :: inSet
    type(ArrayT), intent(in) :: outArray
    character(*), intent(in) :: kwdName
    integer, intent(out), optional :: status

    !-------------------------------------------------------------------
    ! Start

    if (present(status) .and. .not. hasAttribute(inSet, kwdName)) then
      status = -1
  return
    end if

    call setAttribute(outArray, attribute(inSet, kwdName))
    if (present(status)) status = 0

  end subroutine copyAttributeSetToArray

  !---------------------------------------------------------------------
  subroutine copyAttribute_setToSet(inSet, kwdName, outSet, status)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT), intent(in) :: inSet, outSet
    character(*), intent(in) :: kwdName
    integer, intent(out), optional :: status

    !-------------------------------------------------------------------
    ! Start

    if (present(status) .and. .not. hasAttribute(inSet, kwdName)) then
      status = -1
  return
    end if

    call setAttribute(outSet, attribute(inSet, kwdName))
    if (present(status)) status = 0

  end subroutine copyAttribute_setToSet

  !---------------------------------------------------------------------
  subroutine copyAttributeTabToTab(inTab, kwdName, outTab, status)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: inTab, outTab
    character(*), intent(in) :: kwdName
    integer, intent(out), optional :: status

    !-------------------------------------------------------------------
    ! Start

    if (present(status) .and. .not. hasAttribute(inTab, kwdName)) then
      status = -1
  return
    end if

    call setAttribute(outTab, attribute(inTab, kwdName))
    if (present(status)) status = 0

  end subroutine copyAttributeTabToTab

  !---------------------------------------------------------------------
  subroutine readColData_real32_wn(tab, colName, colData, status)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in a real32 vector colData.

    implicit none 

    type(TableT), intent(in)            :: tab
    character(*), intent(in)            :: colName
    real(single), intent(out)           :: colData(:)
    integer,      intent(out), optional :: status

    type(ColumnT) :: col
    integer :: localStatus
    integer(int8), dimension(:), pointer ::&
      colDataInt8
    integer(int16), dimension(:), pointer ::&
      colDataInt16
    integer(int32), dimension(:), pointer ::&
      colDataInt32
    real(single), dimension(:), pointer ::&
      colDataReal32
    real(double), dimension(:), pointer ::&
      colDataReal64

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    colData = 0.0
    if (size(colData) .ne. numberOfRows(tab)) then
      if (present(status)) status = -1
  return
    end if

!    allocate(colData(0:numberOfRows(col) - 1))
    col = column(tab, colName)

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = real(colDataInt8, single)
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = real(colDataInt16, single)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = real(colDataInt32, single)
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = colDataReal32
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = real(colDataReal64, single)
    end select

    call release(col)

    if (present(status)) status = localStatus

  end subroutine readColData_real32_wn

  !---------------------------------------------------------------------
  subroutine readColData_real64_wn(tab, colName, colData, status)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in a real64 vector colData.

    implicit none 

    type(TableT), intent(in)            :: tab
    character(*), intent(in)            :: colName
    real(double), intent(out)           :: colData(:)
    integer,      intent(out), optional :: status

    type(ColumnT) :: col
    integer :: localStatus
    integer(int8), dimension(:), pointer ::&
      colDataInt8
    integer(int16), dimension(:), pointer ::&
      colDataInt16
    integer(int32), dimension(:), pointer ::&
      colDataInt32
    real(single), dimension(:), pointer ::&
      colDataReal32
    real(double), dimension(:), pointer ::&
      colDataReal64

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    colData = 0.0
    if (size(colData) .ne. numberOfRows(tab)) then
      if (present(status)) status = -1
  return
    end if

!    allocate(colData(0:numberOfRows(col) - 1))
    col = column(tab, colName)

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = real(colDataInt8, double)
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = real(colDataInt16, double)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = real(colDataInt32, double)
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = real(colDataReal32, double)
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = colDataReal64
    end select

    call release(col)

    if (present(status)) status = localStatus

  end subroutine readColData_real64_wn

  !---------------------------------------------------------------------
  subroutine readColData_int8_wn(tab, colName, colData, status)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in an int8 vector colData.

    implicit none 

    type(TableT), intent(in)            :: tab
    character(*), intent(in)            :: colName
    integer(int8), intent(out)          :: colData(:)
    integer,      intent(out), optional :: status

    type(ColumnT) :: col
    integer :: localStatus
    integer(int8), dimension(:), pointer ::&
      colDataInt8
    integer(int16), dimension(:), pointer ::&
      colDataInt16
    integer(int32), dimension(:), pointer ::&
      colDataInt32
    real(single), dimension(:), pointer ::&
      colDataReal32
    real(double), dimension(:), pointer ::&
      colDataReal64

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    colData = 0
    if (size(colData) .ne. numberOfRows(tab)) then
      if (present(status)) status = -1
  return
    end if

!    allocate(colData(0:numberOfRows(col) - 1))
    col = column(tab, colName)

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = colDataInt8
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = int(colDataInt16, int8)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = int(colDataInt32, int8)
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = int(colDataReal32, int8)
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = int(colDataReal64, int8)
    end select

    call release(col)

    if (present(status)) status = localStatus

  end subroutine readColData_int8_wn

  !---------------------------------------------------------------------
  subroutine readColData_int16_wn(tab, colName, colData, status)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in an int16 vector colData.

    implicit none 

    type(TableT), intent(in)            :: tab
    character(*), intent(in)            :: colName
    integer(int16), intent(out)         :: colData(:)
    integer,      intent(out), optional :: status

    type(ColumnT) :: col
    integer :: localStatus
    integer(int8), dimension(:), pointer ::&
      colDataInt8
    integer(int16), dimension(:), pointer ::&
      colDataInt16
    integer(int32), dimension(:), pointer ::&
      colDataInt32
    real(single), dimension(:), pointer ::&
      colDataReal32
    real(double), dimension(:), pointer ::&
      colDataReal64

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    colData = 0
    if (size(colData) .ne. numberOfRows(tab)) then
      if (present(status)) status = -1
  return
    end if

!    allocate(colData(0:numberOfRows(col) - 1))
    col = column(tab, colName)

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = int(colDataInt8, int16)
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = int(colDataInt16, int16)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = colDataInt32
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = int(colDataReal32, int16)
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = int(colDataReal64, int16)
    end select

    call release(col)

    if (present(status)) status = localStatus

  end subroutine readColData_int16_wn

  !---------------------------------------------------------------------
  subroutine readColData_int32_wn(tab, colName, colData, status)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in an int32 vector colData.

    implicit none 

    type(TableT), intent(in)            :: tab
    character(*), intent(in)            :: colName
    integer(int32), intent(out)         :: colData(:)
    integer,      intent(out), optional :: status

    type(ColumnT) :: col
    integer :: localStatus
    integer(int8), dimension(:), pointer ::&
      colDataInt8
    integer(int16), dimension(:), pointer ::&
      colDataInt16
    integer(int32), dimension(:), pointer ::&
      colDataInt32
    real(single), dimension(:), pointer ::&
      colDataReal32
    real(double), dimension(:), pointer ::&
      colDataReal64

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    colData = 0
    if (size(colData) .ne. numberOfRows(tab)) then
      if (present(status)) status = -1
  return
    end if

!    allocate(colData(0:numberOfRows(col) - 1))
    col = column(tab, colName)

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = int(colDataInt8, int32)
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = int(colDataInt16, int32)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = colDataInt32
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = int(colDataReal32, int32)
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = int(colDataReal64, int32)
    end select

    call release(col)

    if (present(status)) status = localStatus

  end subroutine readColData_int32_wn

  !---------------------------------------------------------------------
  subroutine readColData_real32(col, colData, status)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in a real32 vector colData.

    implicit none 

    type(ColumnT), intent(in)           :: col
    real(single), intent(out)           :: colData(:)
    integer,      intent(out), optional :: status

    integer :: localStatus
    integer(int8), dimension(:), pointer ::&
      colDataInt8
    integer(int16), dimension(:), pointer ::&
      colDataInt16
    integer(int32), dimension(:), pointer ::&
      colDataInt32
    real(single), dimension(:), pointer ::&
      colDataReal32
    real(double), dimension(:), pointer ::&
      colDataReal64

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    colData = 0.0
    if (size(colData) .ne. numberOfRows(col)) then
      if (present(status)) status = -1
  return
    end if

!    allocate(colData(0:numberOfRows(col) - 1))

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = real(colDataInt8, single)
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = real(colDataInt16, single)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = real(colDataInt32, single)
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = colDataReal32
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = real(colDataReal64, single)
    end select

    if (present(status)) status = localStatus

  end subroutine readColData_real32

  !---------------------------------------------------------------------
  subroutine readColData_real64(col, colData, status)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in a real64 vector colData.

    implicit none 

    type(ColumnT), intent(in)           :: col
    real(double), intent(out)           :: colData(:)
    integer,      intent(out), optional :: status

    integer :: localStatus
    integer(int8), dimension(:), pointer ::&
      colDataInt8
    integer(int16), dimension(:), pointer ::&
      colDataInt16
    integer(int32), dimension(:), pointer ::&
      colDataInt32
    real(single), dimension(:), pointer ::&
      colDataReal32
    real(double), dimension(:), pointer ::&
      colDataReal64

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    colData = 0.0
    if (size(colData) .ne. numberOfRows(col)) then
      if (present(status)) status = -1
  return
    end if

!    allocate(colData(0:numberOfRows(col) - 1))

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = real(colDataInt8, double)
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = real(colDataInt16, double)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = real(colDataInt32, double)
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = real(colDataReal32, double)
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = colDataReal64
    end select

    if (present(status)) status = localStatus

  end subroutine readColData_real64

  !---------------------------------------------------------------------
  subroutine readColData_int8(col, colData, status)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in an int8 vector colData.

    implicit none 

    type(ColumnT), intent(in)           :: col
    integer(int8), intent(out)          :: colData(:)
    integer,      intent(out), optional :: status

    integer :: localStatus
    integer(int8), dimension(:), pointer ::&
      colDataInt8
    integer(int16), dimension(:), pointer ::&
      colDataInt16
    integer(int32), dimension(:), pointer ::&
      colDataInt32
    real(single), dimension(:), pointer ::&
      colDataReal32
    real(double), dimension(:), pointer ::&
      colDataReal64

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    colData = 0
    if (size(colData) .ne. numberOfRows(col)) then
      if (present(status)) status = -1
  return
    end if

!    allocate(colData(0:numberOfRows(col) - 1))

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = colDataInt8
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = int(colDataInt16, int8)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = int(colDataInt32, int8)
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = int(colDataReal32, int8)
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = int(colDataReal64, int8)
    end select

    if (present(status)) status = localStatus

  end subroutine readColData_int8

  !---------------------------------------------------------------------
  subroutine readColData_int16(col, colData, status)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in an int16 vector colData.

    implicit none 

    type(ColumnT), intent(in)           :: col
    integer(int16), intent(out)         :: colData(:)
    integer,      intent(out), optional :: status

    integer :: localStatus
    integer(int8), dimension(:), pointer ::&
      colDataInt8
    integer(int16), dimension(:), pointer ::&
      colDataInt16
    integer(int32), dimension(:), pointer ::&
      colDataInt32
    real(single), dimension(:), pointer ::&
      colDataReal32
    real(double), dimension(:), pointer ::&
      colDataReal64

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    colData = 0
    if (size(colData) .ne. numberOfRows(col)) then
      if (present(status)) status = -1
  return
    end if

!    allocate(colData(0:numberOfRows(col) - 1))

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = int(colDataInt8, int16)
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = int(colDataInt16, int16)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = colDataInt32
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = int(colDataReal32, int16)
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = int(colDataReal64, int16)
    end select

    if (present(status)) status = localStatus

  end subroutine readColData_int16

  !---------------------------------------------------------------------
  subroutine readColData_int32(col, colData, status)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in an int32 vector colData.

    implicit none 

    type(ColumnT), intent(in)           :: col
    integer(int32), intent(out)         :: colData(:)
    integer,      intent(out), optional :: status

    integer :: localStatus
    integer(int8), dimension(:), pointer ::&
      colDataInt8
    integer(int16), dimension(:), pointer ::&
      colDataInt16
    integer(int32), dimension(:), pointer ::&
      colDataInt32
    real(single), dimension(:), pointer ::&
      colDataReal32
    real(double), dimension(:), pointer ::&
      colDataReal64

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    colData = 0
    if (size(colData) .ne. numberOfRows(col)) then
      if (present(status)) status = -1
  return
    end if

!    allocate(colData(0:numberOfRows(col) - 1))

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = int(colDataInt8, int32)
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = int(colDataInt16, int32)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = colDataInt32
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = int(colDataReal32, int32)
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = int(colDataReal64, int32)
    end select

    if (present(status)) status = localStatus

  end subroutine readColData_int32

  !---------------------------------------------------------------------
  subroutine myAddColumn_int32(tab, name, inData, unit, comment, nulls&
  , nullValue)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: tab
    character(*), intent(in) :: name
    character(*), intent(in), optional :: unit, comment
    logical, intent(in), optional :: nulls(:)
    integer(int32), intent(in), optional :: nullValue
    integer(int32), intent(in) :: inData(:)

    character(255) :: localUnit, localComment
    logical :: localNulls(size(inData))
    type(ColumnT) :: col
    integer(int32), pointer :: colData(:)
    integer(int32) :: i, localNullValue

    !-------------------------------------------------------------------
    ! Start

    if (present(unit)) then
      localUnit = unit
    else
      localUnit = ''
    end if
    if (present(comment)) then
      localComment = comment
    else
      localComment = ''
    end if
    if (present(nullValue)) then
      localNullValue = nullValue
    else
      localNullValue = -999
    end if
    if (present(nulls)) then
      localNulls = .true.
      if (size(nulls) .eq. size(inData)) then
        localNulls = nulls
      else if (size(nulls) .gt. size(inData)) then
        localNulls = nulls(1:size(inData))
      else
        localNulls(1:size(nulls)) = nulls
      end if
    else
      localNulls = .false.
    end if

    col = addColumn(tab, name, INTEGER32, trim(localUnit)&
    , comment=trim(localComment))

    colData => int32Data(col)
    colData = inData(1:size(inData))

    if (any(localNulls)) then
      call setNullValue(col, localNullValue)
      do i = 1, size(inData)
        if (localNulls(i)) call setToNull(col, i - 1)
      end do
    end if

    call release(col)

  end subroutine myAddColumn_int32

  !---------------------------------------------------------------------
  subroutine myAddColumn_real32(tab, name, inData, unit, comment, nulls)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: tab
    character(*), intent(in) :: name
    character(*), intent(in), optional :: unit, comment
    logical, intent(in), optional :: nulls(:)
    real(single), intent(in) :: inData(:)

    character(255) :: localUnit, localComment
    logical :: localNulls(size(inData))
    type(ColumnT) :: col
    real(single), pointer :: colData(:)
    integer(int32) :: i

    !-------------------------------------------------------------------
    ! Start

    if (present(unit)) then
      localUnit = unit
    else
      localUnit = ''
    end if
    if (present(comment)) then
      localComment = comment
    else
      localComment = ''
    end if
    if (present(nulls)) then
      localNulls = .true.
      if (size(nulls) .eq. size(inData)) then
        localNulls = nulls
      else if (size(nulls) .gt. size(inData)) then
        localNulls = nulls(1:size(inData))
      else
        localNulls(1:size(nulls)) = nulls
      end if
    else
      localNulls = .false.
    end if

    col = addColumn(tab, name, REAL32, trim(localUnit)&
    , comment=trim(localComment))

    colData => real32Data(col)
    colData = inData(1:size(inData))

    if (any(localNulls)) then
      do i = 1, size(inData)
        if (localNulls(i)) call setToNull(col, i - 1)
      end do
    end if

    call release(col)

  end subroutine myAddColumn_real32

  !---------------------------------------------------------------------
  subroutine myAddColumn_real64(tab, name, inData, unit, comment, nulls)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: tab
    character(*), intent(in) :: name
    character(*), intent(in), optional :: unit, comment
    logical, intent(in), optional :: nulls(:)
    real(double), intent(in) :: inData(:)

    character(255) :: localUnit, localComment
    logical :: localNulls(size(inData))
    type(ColumnT) :: col
    real(double), pointer :: colData(:)
    integer(int32) :: i

    !-------------------------------------------------------------------
    ! Start

    if (present(unit)) then
      localUnit = unit
    else
      localUnit = ''
    end if
    if (present(comment)) then
      localComment = comment
    else
      localComment = ''
    end if
    if (present(nulls)) then
      localNulls = .true.
      if (size(nulls) .eq. size(inData)) then
        localNulls = nulls
      else if (size(nulls) .gt. size(inData)) then
        localNulls = nulls(1:size(inData))
      else
        localNulls(1:size(nulls)) = nulls
      end if
    else
      localNulls = .false.
    end if

    col = addColumn(tab, name, REAL64, trim(localUnit)&
    , comment=trim(localComment))

    colData => real64Data(col)
    colData = inData(1:size(inData))

    if (any(localNulls)) then
      do i = 1, size(inData)
        if (localNulls(i)) call setToNull(col, i - 1)
      end do
    end if

    call release(col)

  end subroutine myAddColumn_real64

  !---------------------------------------------------------------------
  subroutine minNonNullValue_real32(tab, colName, minValue, allRowsNull)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: tab
    character(*), intent(in) :: colName
    real(single), intent(out) :: minValue
    logical, intent(out) :: allRowsNull

    type(ColumnT) :: col
    integer(int32) :: firstNonNullRow, row
    real(single), pointer :: colData(:)

    !-------------------------------------------------------------------
    ! Start

    minValue = 0.0

    col = column(tab, colName)
    allRowsNull = .true. ! default
    firstNonNullRow = -1
    do row = 0, numberOfRows(col) - 1
      if (.not. isNull(col, row)) then
        allRowsNull = .false.
        firstNonNullRow = row
    exit
      end if
    end do

    if (allRowsNull) then
      call release(col)
  return
    end if

    colData => real32Data(col)
    minValue = colData(firstNonNullRow)
    do row = firstNonNullRow + 1, numberOfRows(col) - 1
      if (.not. isNull(col, row)) then
        minValue = min(minValue, colData(row))
      end if
    end do

    call release(col)

  end subroutine minNonNullValue_real32

  !---------------------------------------------------------------------
  subroutine minNonNullValue_real64(tab, colName, minValue, allRowsNull)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: tab
    character(*), intent(in) :: colName
    real(double), intent(out) :: minValue
    logical, intent(out) :: allRowsNull

    type(ColumnT) :: col
    integer(int32) :: firstNonNullRow, row
    real(double), pointer :: colData(:)

    !-------------------------------------------------------------------
    ! Start

    minValue = 0.0

    col = column(tab, colName)
    allRowsNull = .true. ! default
    firstNonNullRow = -1
    do row = 0, numberOfRows(col) - 1
      if (.not. isNull(col, row)) then
        allRowsNull = .false.
        firstNonNullRow = row
    exit
      end if
    end do

    if (allRowsNull) then
      call release(col)
  return
    end if

    colData => real64Data(col)
    minValue = colData(firstNonNullRow)
    do row = firstNonNullRow + 1, numberOfRows(col) - 1
      if (.not. isNull(col, row)) then
        minValue = min(minValue, colData(row))
      end if
    end do

    call release(col)

  end subroutine minNonNullValue_real64

  !---------------------------------------------------------------------
  subroutine maxNonNullValueInt16(tab, colName, maxValue, allRowsNull, startRow, finisRow)
  !---------------------------------------------------------------------
  ! NOTE! If setNullValue(column, value) has not been called prior to entry into this subroutine, an error will result.

    implicit none

    type(TableT),   intent(in)           :: tab
    character(*),   intent(in)           :: colName
    integer(int32), intent(in), optional :: startRow, finisRow
    integer(int16), intent(out)          :: maxValue
    logical,        intent(out)          :: allRowsNull

    type(ColumnT) :: col
    integer(int32) ::&
      firstNonNullRow, row, numRows, localStartRow, localFinisRow
    integer(int16), pointer :: colData(:)

    !-------------------------------------------------------------------
    ! Start

    maxValue = 0.0

    col = column(tab, colName)

    if (present(startRow)) then
if (startRow < 0) call error('negativeStartRow','ssclib:')
      localStartRow = startRow
    else
      localStartRow = 0
    end if

    numRows = numberOfRows(col)

    if (present(finisRow)) then
      localFinisRow = finisRow
if (finisRow > numRows) call error('numLastRowTooLarge','ssclib:')
    else
      localFinisRow = numRows - 1
    end if

    allRowsNull = .true. ! default
    firstNonNullRow = -1
    do row = localStartRow, localFinisRow
      if (.not. isNull(col, row)) then
        allRowsNull = .false.
        firstNonNullRow = row
    exit
      end if
    end do

    if (allRowsNull) then
      call release(col)
  return
    end if

    colData => int16Data(col)
    maxValue = colData(firstNonNullRow)
    do row = firstNonNullRow + 1, localFinisRow
      if (.not. isNull(col, row)) then
        maxValue = max(maxValue, colData(row))
      end if
    end do

    call release(col)

  end subroutine maxNonNullValueInt16

  !---------------------------------------------------------------------
  subroutine maxNonNullValue_real32(tab, colName, maxValue, allRowsNull)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: tab
    character(*), intent(in) :: colName
    real(single), intent(out) :: maxValue
    logical, intent(out) :: allRowsNull

    type(ColumnT) :: col
    integer(int32) :: firstNonNullRow, row
    real(single), pointer :: colData(:)

    !-------------------------------------------------------------------
    ! Start

    maxValue = 0.0

    col = column(tab, colName)
    allRowsNull = .true. ! default
    firstNonNullRow = -1
    do row = 0, numberOfRows(col) - 1
      if (.not. isNull(col, row)) then
        allRowsNull = .false.
        firstNonNullRow = row
    exit
      end if
    end do

    if (allRowsNull) then
      call release(col)
  return
    end if

    colData => real32Data(col)
    maxValue = colData(firstNonNullRow)
    do row = firstNonNullRow + 1, numberOfRows(col) - 1
      if (.not. isNull(col, row)) then
        maxValue = max(maxValue, colData(row))
      end if
    end do

    call release(col)

  end subroutine maxNonNullValue_real32

  !---------------------------------------------------------------------
  subroutine maxNonNullValue_real64(tab, colName, maxValue, allRowsNull)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: tab
    character(*), intent(in) :: colName
    real(double), intent(out) :: maxValue
    logical, intent(out) :: allRowsNull

    type(ColumnT) :: col
    integer(int32) :: firstNonNullRow, row
    real(double), pointer :: colData(:)

    !-------------------------------------------------------------------
    ! Start

    maxValue = 0.0

    col = column(tab, colName)
    allRowsNull = .true. ! default
    firstNonNullRow = -1
    do row = 0, numberOfRows(col) - 1
      if (.not. isNull(col, row)) then
        allRowsNull = .false.
        firstNonNullRow = row
    exit
      end if
    end do

    if (allRowsNull) then
      call release(col)
  return
    end if

    colData => real64Data(col)
    maxValue = colData(firstNonNullRow)
    do row = firstNonNullRow + 1, numberOfRows(col) - 1
      if (.not. isNull(col, row)) then
        maxValue = max(maxValue, colData(row))
      end if
    end do

    call release(col)

  end subroutine maxNonNullValue_real64

  !---------------------------------------------------------------------
  subroutine loadFitsArray_1d_r32(imageSetName, vector)!!!, dataType)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: imageSetName
    real(single), dimension(:), pointer :: vector

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(imageSetName, READ)
    inArray = array(set, 0, READ)
    call loadFitsArray_1d_wa_r32(inArray, vector)!!!, dataType)
    call release(set)

  end subroutine loadFitsArray_1d_r32

  !---------------------------------------------------------------------
  subroutine loadFitsArray_2d_r32(imageSetName, image)!!!, dataType)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: imageSetName
    real(single), dimension(:,:), pointer :: image

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(imageSetName, READ)
    inArray = array(set, 0, READ)
    call loadFitsArray_2d_wa_r32(inArray, image)!!!, dataType)
    call release(set)

  end subroutine loadFitsArray_2d_r32

  !---------------------------------------------------------------------
  subroutine loadFitsArray_3d_r32(imageSetName, cube)!!!, dataType)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: imageSetName
    real(single), dimension(:,:,:), pointer :: cube

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(imageSetName, READ)
    inArray = array(set, 0, READ)
    call loadFitsArray_3d_wa_r32(inArray, cube)!!!, dataType)
    call release(set)

  end subroutine loadFitsArray_3d_r32

  !---------------------------------------------------------------------
  subroutine loadFitsArray_4d_r32(imageSetName, hyperCube)!!!, dataType)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: imageSetName
    real(single), dimension(:,:,:,:), pointer :: hyperCube

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(imageSetName, READ)
    inArray = array(set, 0, READ)
    call loadFitsArray_4d_wa_r32(inArray, hyperCube)!!!, dataType)
    call release(set)

  end subroutine loadFitsArray_4d_r32

  !---------------------------------------------------------------------
  subroutine loadFitsArray_1d_r64(imageSetName, vector)!!!, dataType)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: imageSetName
    real(double), dimension(:), pointer :: vector

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(imageSetName, READ)
    inArray = array(set, 0, READ)
    call loadFitsArray_1d_wa_r64(inArray, vector)!!!, dataType)
    call release(set)

  end subroutine loadFitsArray_1d_r64

  !---------------------------------------------------------------------
  subroutine loadFitsArray_2d_r64(imageSetName, image)!!!, dataType)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: imageSetName
    real(double), dimension(:,:), pointer :: image

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(imageSetName, READ)
    inArray = array(set, 0, READ)
    call loadFitsArray_2d_wa_r64(inArray, image)!!!, dataType)
    call release(set)

  end subroutine loadFitsArray_2d_r64

  !---------------------------------------------------------------------
  subroutine loadFitsArray_3d_r64(imageSetName, cube)!!!, dataType)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: imageSetName
    real(double), dimension(:,:,:), pointer :: cube

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(imageSetName, READ)
    inArray = array(set, 0, READ)
    call loadFitsArray_3d_wa_r64(inArray, cube)!!!, dataType)
    call release(set)

  end subroutine loadFitsArray_3d_r64

  !---------------------------------------------------------------------
  subroutine loadFitsArray_4d_r64(imageSetName, hyperCube)!!!, dataType)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: imageSetName
    real(double), dimension(:,:,:,:), pointer :: hyperCube

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(imageSetName, READ)
    inArray = array(set, 0, READ)
    call loadFitsArray_4d_wa_r64(inArray, hyperCube)!!!, dataType)
    call release(set)

  end subroutine loadFitsArray_4d_r64

  !---------------------------------------------------------------------
  subroutine loadFitsArray_1d_wa_r32(inArray, vector)!!!, dataType)
  !---------------------------------------------------------------------
!**** may need to deallocate vector!!

    implicit none

    type(ArrayT), intent(in) :: inArray
    real(single), dimension(:), pointer :: vector

    integer(int32), dimension(:), pointer ::&
      dimVector
    real(single), dimension(:,:), pointer :: image
    real(single), dimension(:,:,:), pointer :: cube
    real(single), dimension(:,:,:,:), pointer :: hyperCube
    integer :: numDimensions, arrayShape(4), biggestDimension(1), vectorShape(1)

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)

    numDimensions = size(dimVector)

    arrayShape = 0 ! vector op, default

    select case(numDimensions)
    case(1)
call error('arrayRankTooSmall', '')
    case(2)
      call loadFitsArray_2d_wa_r32(inArray, image)!!!, dataType)
      arrayShape(:2) = shape(image)
if (count(arrayShape .gt. 1) .gt. 1) call error('badImageShape','')
      biggestDimension = maxloc(arrayShape)
      vectorShape(1) = size(image, biggestDimension(1))
      allocate(vector(vectorShape(1)))
      vector = reshape(image, vectorShape)

    case(3)
      call loadFitsArray_3d_wa_r32(inArray, cube)!!!, dataType)
      arrayShape(:3) = shape(cube)
if (count(arrayShape .gt. 1) .gt. 1) call error('badCubeShape','')
      biggestDimension = maxloc(arrayShape)
      vectorShape(1) = size(cube, biggestDimension(1))
      allocate(vector(vectorShape(1)))
      vector = reshape(cube, vectorShape)

    case(4)
      call loadFitsArray_4d_wa_r32(inArray, hyperCube)!!!, dataType)
      arrayShape = shape(hyperCube)
if (count(arrayShape .gt. 1) .gt. 1) call error('badHyperCubeShape','')
      biggestDimension = maxloc(arrayShape)
      vectorShape(1) = size(hyperCube, biggestDimension(1))
      allocate(vector(vectorShape(1)))
      vector = reshape(hyperCube, vectorShape)

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine loadFitsArray_1d_wa_r32

  !---------------------------------------------------------------------
  subroutine loadFitsArray_2d_wa_r32(inArray, image)!!!, dataType)
  !---------------------------------------------------------------------
!**** may need to deallocate image!!

    implicit none

    type(ArrayT), intent(in) :: inArray
    real(single), dimension(:,:), pointer :: image

    integer(int32), dimension(:), pointer ::&
      dimVector
    real(single), dimension(:,:,:), pointer :: cube
    real(single), dimension(:,:,:,:), pointer :: hyperCube
    integer :: numDimensions, arrayShape(4), biggestDimensions(2, 1), imageShape(2)
    logical :: mask(4)

    integer(int8), dimension(:,:), pointer ::&
      arrayDataInt8
    integer(int16), dimension(:,:), pointer ::&
      arrayDataInt16
    integer(int32), dimension(:,:), pointer ::&
      arrayDataInt32
    real(single), dimension(:,:), pointer ::&
      arrayDataReal32
    real(double), dimension(:,:), pointer ::&
      arrayDataReal64

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)

    numDimensions = size(dimVector)

    arrayShape = 0 ! vector op, default
    mask = .true.  ! vector op, default

    select case(numDimensions)
    case(1)
call error('arrayRankTooSmall', '')
    case(2)
      allocate(image(dimVector(0), dimVector(1)))

      if (dataType(inArray) .eq. INTEGER8) then
        arrayDataInt8 => int8Array2Data(inArray)
        image = real(arrayDataInt8, single)
      else if (dataType(inArray) .eq. INTEGER16) then
        arrayDataInt16 => int16Array2Data(inArray)
        image = real(arrayDataInt16, single)
      else if (dataType(inArray) .eq. INTEGER32) then
        arrayDataInt32 => int32Array2Data(inArray)
        image = real(arrayDataInt32, single)
      else if (dataType(inArray) .eq. REAL32) then
        arrayDataReal32 => real32Array2Data(inArray)
        image = arrayDataReal32
      else ! this gives a DAL error if data type is none of the above.
        arrayDataReal64 => real64Array2Data(inArray)
        image = real(arrayDataReal64, single)
      end if

    case(3)
      call loadFitsArray_3d_wa_r32(inArray, cube)!!!, dataType)
      arrayShape(:3) = shape(cube)
if (count(arrayShape .gt. 1) .gt. 2) call error('badCubeShape','')
      biggestDimensions(1,:) = maxloc(arrayShape)
      mask(biggestDimensions(1,1)) = .false.
      biggestDimensions(2,:) = maxloc(arrayShape, mask)
      imageShape(1) = size(cube, biggestDimensions(1,1))
      imageShape(2) = size(cube, biggestDimensions(2,1))
      allocate(image(imageShape(1), imageShape(2)))
      image = reshape(cube, imageShape)

    case(4)
      call loadFitsArray_4d_wa_r32(inArray, hyperCube)!!!, dataType)
      arrayShape = shape(hyperCube)
if (count(arrayShape .gt. 1) .gt. 2) call error('badHyperCubeShape','')
      biggestDimensions(1,:) = maxloc(arrayShape)
      mask(biggestDimensions(1,1)) = .false.
      biggestDimensions(2,:) = maxloc(arrayShape, mask)
      imageShape(1) = size(cube, biggestDimensions(1,1))
      imageShape(2) = size(cube, biggestDimensions(2,1))
      allocate(image(imageShape(1), imageShape(2)))
      image = reshape(hyperCube, imageShape)

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine loadFitsArray_2d_wa_r32

  !---------------------------------------------------------------------
  subroutine loadFitsArray_3d_wa_r32(inArray, cube)!!!, dataType)
  !---------------------------------------------------------------------
!**** may need to deallocate cube!!

    implicit none

    type(ArrayT), intent(in) :: inArray
    real(single), dimension(:,:,:), pointer :: cube

    integer(int32), dimension(:), pointer ::&
      dimVector
    real(single), dimension(:,:,:,:), pointer :: hyperCube
    integer :: numDimensions, arrayShape(4), biggestDimensions(3, 1), cubeShape(3)
    logical :: mask(4)

    integer(int8), dimension(:,:,:), pointer ::&
      arrayDataInt8
    integer(int16), dimension(:,:,:), pointer ::&
      arrayDataInt16
    integer(int32), dimension(:,:,:), pointer ::&
      arrayDataInt32
    real(single), dimension(:,:,:), pointer ::&
      arrayDataReal32
    real(double), dimension(:,:,:), pointer ::&
      arrayDataReal64

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)

    numDimensions = size(dimVector)

    arrayShape = 0 ! vector op, default
    mask = .true.  ! vector op, default

    select case(numDimensions)
    case(1, 2)
call error('arrayRankTooSmall', '')
    case(3)
      allocate(cube(dimVector(0), dimVector(1), dimVector(2)))

      if (dataType(inArray) .eq. INTEGER8) then
        arrayDataInt8 => int8Array3Data(inArray)
        cube = real(arrayDataInt8, single)
      else if (dataType(inArray) .eq. INTEGER16) then
        arrayDataInt16 => int16Array3Data(inArray)
        cube = real(arrayDataInt16, single)
      else if (dataType(inArray) .eq. INTEGER32) then
        arrayDataInt32 => int32Array3Data(inArray)
        cube = real(arrayDataInt32, single)
      else if (dataType(inArray) .eq. REAL32) then
        arrayDataReal32 => real32Array3Data(inArray)
        cube = arrayDataReal32
      else ! this gives a DAL error if data type is none of the above.
        arrayDataReal64 => real64Array3Data(inArray)
        cube = real(arrayDataReal64, single)
      end if

    case(4)
      call loadFitsArray_4d_wa_r32(inArray, hyperCube)!!!, dataType)
      arrayShape = shape(hyperCube)
if (count(arrayShape .gt. 1) .gt. 3) call error('badHyperCubeShape','')
      biggestDimensions(1,:) = maxloc(arrayShape)
      mask(biggestDimensions(1,1)) = .false.
      biggestDimensions(2,:) = maxloc(arrayShape, mask)
      mask(biggestDimensions(2,1)) = .false.
      biggestDimensions(3,:) = maxloc(arrayShape, mask)
      cubeShape(1) = size(hyperCube, biggestDimensions(1,1))
      cubeShape(2) = size(hyperCube, biggestDimensions(2,1))
      cubeShape(3) = size(hyperCube, biggestDimensions(3,1))
      allocate(cube(cubeShape(1), cubeShape(2), cubeShape(3)))
      cube = reshape(hyperCube, cubeShape)

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine loadFitsArray_3d_wa_r32

  !---------------------------------------------------------------------
  subroutine loadFitsArray_4d_wa_r32(inArray, hypercube)!!!, dataType)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT), intent(in) :: inArray
    real(single), dimension(:,:,:,:), pointer :: hyperCube

    integer(int32), dimension(:), pointer ::&
      dimVector
    integer :: numDimensions

!    integer(int8),  dimension(:,:,:,:), pointer ::&
!      arrayDataInt8
!    integer(int16), dimension(:,:,:,:), pointer ::&
!      arrayDataInt16
!    integer(int32), dimension(:,:,:,:), pointer ::&
!      arrayDataInt32
!    real(single),   dimension(:,:,:,:), pointer ::&
!      arrayDataReal32
!    real(double),   dimension(:,:,:,:), pointer ::&
!      arrayDataReal64

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)

    numDimensions = size(dimVector)

    select case(numDimensions)
    case(1, 2, 3)
call error('arrayRankTooSmall', '')
    case(4)
call error('notAvailableYet', "Reading of 4-dimensional arrays doesn't &
&seem to be available in the DAL yet.")

!      allocate(hypercube(dimVector(0), dimVector(1), dimVector(2)&
!      , dimVector(3)))
!
!      if (dataType(inArray) .eq. INTEGER8) then
!        arrayDataInt8    => int8Array4Data(inArray)
!        hyperCube = real(arrayDataInt8, double)
!      else if (dataType(inArray) .eq. INTEGER16) then
!        arrayDataInt16   => int16Array4Data(inArray)
!        hyperCube = real(arrayDataInt16, double)
!      else if (dataType(inArray) .eq. INTEGER32) then
!        arrayDataInt32   => int32Array4Data(inArray)
!        hyperCube = real(arrayDataInt32, double)
!      else if (dataType(inArray) .eq. REAL32) then
!        arrayDataReal32 => real32Array4Data(inArray)
!        hyperCube = real(arrayDataReal32, double)
!      else ! this gives a DAL error if data type is none of the above.
!        arrayDataReal64 => real64Array4Data(inArray)
!        hyperCube = arrayDataReal64
!      end if
hyperCube = 0.0 ! keeps the compiler from issuing a warning.

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine loadFitsArray_4d_wa_r32

  !---------------------------------------------------------------------
  subroutine loadFitsArray_1d_wa_r64(inArray, vector)!!!, dataType)
  !---------------------------------------------------------------------
!**** may need to deallocate vector!!

    implicit none

    type(ArrayT), intent(in) :: inArray
    real(double), dimension(:), pointer :: vector

    integer(int32), dimension(:), pointer ::&
      dimVector
    real(double), dimension(:,:), pointer :: image
    real(double), dimension(:,:,:), pointer :: cube
    real(double), dimension(:,:,:,:), pointer :: hyperCube
    integer :: numDimensions, arrayShape(4), biggestDimension(1), vectorShape(1)

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)

    numDimensions = size(dimVector)

    arrayShape = 0 ! vector op, default

    select case(numDimensions)
    case(1)
call error('arrayRankTooSmall', '')
    case(2)
      call loadFitsArray_2d_wa_r64(inArray, image)!!!, dataType)
      arrayShape(:2) = shape(image)
if (count(arrayShape .gt. 1) .gt. 1) call error('badImageShape','')
      biggestDimension = maxloc(arrayShape)
      vectorShape(1) = size(image, biggestDimension(1))
      allocate(vector(vectorShape(1)))
      vector = reshape(image, vectorShape)

    case(3)
      call loadFitsArray_3d_wa_r64(inArray, cube)!!!, dataType)
      arrayShape(:3) = shape(cube)
if (count(arrayShape .gt. 1) .gt. 1) call error('badCubeShape','')
      biggestDimension = maxloc(arrayShape)
      vectorShape(1) = size(cube, biggestDimension(1))
      allocate(vector(vectorShape(1)))
      vector = reshape(cube, vectorShape)

    case(4)
      call loadFitsArray_4d_wa_r64(inArray, hyperCube)!!!, dataType)
      arrayShape = shape(hyperCube)
if (count(arrayShape .gt. 1) .gt. 1) call error('badHyperCubeShape','')
      biggestDimension = maxloc(arrayShape)
      vectorShape(1) = size(hyperCube, biggestDimension(1))
      allocate(vector(vectorShape(1)))
      vector = reshape(hyperCube, vectorShape)

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine loadFitsArray_1d_wa_r64

  !---------------------------------------------------------------------
  subroutine loadFitsArray_2d_wa_r64(inArray, image)!!!, dataType)
  !---------------------------------------------------------------------
!**** may need to deallocate image!!

    implicit none

    type(ArrayT), intent(in) :: inArray
    real(double), dimension(:,:), pointer :: image

    integer(int32), dimension(:), pointer ::&
      dimVector
    real(double), dimension(:,:,:), pointer :: cube
    real(double), dimension(:,:,:,:), pointer :: hyperCube
    integer :: numDimensions, arrayShape(4), biggestDimensions(2, 1), imageShape(2)
    logical :: mask(4)

    integer(int8), dimension(:,:), pointer ::&
      arrayDataInt8
    integer(int16), dimension(:,:), pointer ::&
      arrayDataInt16
    integer(int32), dimension(:,:), pointer ::&
      arrayDataInt32
    real(single), dimension(:,:), pointer ::&
      arrayDataReal32
    real(double), dimension(:,:), pointer ::&
      arrayDataReal64

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)

    numDimensions = size(dimVector)

    arrayShape = 0 ! vector op, default
    mask = .true.  ! vector op, default

    select case(numDimensions)
    case(1)
call error('arrayRankTooSmall', '')
    case(2)
      allocate(image(dimVector(0), dimVector(1)))

      if (dataType(inArray) .eq. INTEGER8) then
        arrayDataInt8 => int8Array2Data(inArray)
        image = real(arrayDataInt8, double)
      else if (dataType(inArray) .eq. INTEGER16) then
        arrayDataInt16 => int16Array2Data(inArray)
        image = real(arrayDataInt16, double)
      else if (dataType(inArray) .eq. INTEGER32) then
        arrayDataInt32 => int32Array2Data(inArray)
        image = real(arrayDataInt32, double)
      else if (dataType(inArray) .eq. REAL32) then
        arrayDataReal32 => real32Array2Data(inArray)
        image = real(arrayDataReal32, double)
      else ! this gives a DAL error if data type is none of the above.
        arrayDataReal64 => real64Array2Data(inArray)
        image = arrayDataReal64
      end if

    case(3)
      call loadFitsArray_3d_wa_r64(inArray, cube)!!!, dataType)
      arrayShape(:3) = shape(cube)
if (count(arrayShape .gt. 1) .gt. 2) call error('badCubeShape','')
      biggestDimensions(1,:) = maxloc(arrayShape)
      mask(biggestDimensions(1,1)) = .false.
      biggestDimensions(2,:) = maxloc(arrayShape, mask)
      imageShape(1) = size(cube, biggestDimensions(1,1))
      imageShape(2) = size(cube, biggestDimensions(2,1))
      allocate(image(imageShape(1), imageShape(2)))
      image = reshape(cube, imageShape)

    case(4)
      call loadFitsArray_4d_wa_r64(inArray, hyperCube)!!!, dataType)
      arrayShape = shape(hyperCube)
if (count(arrayShape .gt. 1) .gt. 2) call error('badHyperCubeShape','')
      biggestDimensions(1,:) = maxloc(arrayShape)
      mask(biggestDimensions(1,1)) = .false.
      biggestDimensions(2,:) = maxloc(arrayShape, mask)
      imageShape(1) = size(cube, biggestDimensions(1,1))
      imageShape(2) = size(cube, biggestDimensions(2,1))
      allocate(image(imageShape(1), imageShape(2)))
      image = reshape(hyperCube, imageShape)

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine loadFitsArray_2d_wa_r64

  !---------------------------------------------------------------------
  subroutine loadFitsArray_3d_wa_r64(inArray, cube)!!!, dataType)
  !---------------------------------------------------------------------
!**** may need to deallocate cube!!

    implicit none

    type(ArrayT), intent(in) :: inArray
    real(double), dimension(:,:,:), pointer :: cube

    integer(int32), dimension(:), pointer ::&
      dimVector
    real(double), dimension(:,:,:,:), pointer :: hyperCube
    integer :: numDimensions, arrayShape(4), biggestDimensions(3, 1), cubeShape(3)
    logical :: mask(4)

    integer(int8), dimension(:,:,:), pointer ::&
      arrayDataInt8
    integer(int16), dimension(:,:,:), pointer ::&
      arrayDataInt16
    integer(int32), dimension(:,:,:), pointer ::&
      arrayDataInt32
    real(single), dimension(:,:,:), pointer ::&
      arrayDataReal32
    real(double), dimension(:,:,:), pointer ::&
      arrayDataReal64

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)

    numDimensions = size(dimVector)

    arrayShape = 0 ! vector op, default
    mask = .true.  ! vector op, default

    select case(numDimensions)
    case(1, 2)
call error('arrayRankTooSmall', '')
    case(3)
      allocate(cube(dimVector(0), dimVector(1), dimVector(2)))

      if (dataType(inArray) .eq. INTEGER8) then
        arrayDataInt8 => int8Array3Data(inArray)
        cube = real(arrayDataInt8, double)
      else if (dataType(inArray) .eq. INTEGER16) then
        arrayDataInt16 => int16Array3Data(inArray)
        cube = real(arrayDataInt16, double)
      else if (dataType(inArray) .eq. INTEGER32) then
        arrayDataInt32 => int32Array3Data(inArray)
        cube = real(arrayDataInt32, double)
      else if (dataType(inArray) .eq. REAL32) then
        arrayDataReal32 => real32Array3Data(inArray)
        cube = real(arrayDataReal32, double)
      else ! this gives a DAL error if data type is none of the above.
        arrayDataReal64 => real64Array3Data(inArray)
        cube = arrayDataReal64
      end if

    case(4)
      call loadFitsArray_4d_wa_r64(inArray, hyperCube)!!!, dataType)
      arrayShape = shape(hyperCube)
if (count(arrayShape .gt. 1) .gt. 3) call error('badHyperCubeShape','')
      biggestDimensions(1,:) = maxloc(arrayShape)
      mask(biggestDimensions(1,1)) = .false.
      biggestDimensions(2,:) = maxloc(arrayShape, mask)
      mask(biggestDimensions(2,1)) = .false.
      biggestDimensions(3,:) = maxloc(arrayShape, mask)
      cubeShape(1) = size(hyperCube, biggestDimensions(1,1))
      cubeShape(2) = size(hyperCube, biggestDimensions(2,1))
      cubeShape(3) = size(hyperCube, biggestDimensions(3,1))
      allocate(cube(cubeShape(1), cubeShape(2), cubeShape(3)))
      cube = reshape(hyperCube, cubeShape)

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine loadFitsArray_3d_wa_r64

  !---------------------------------------------------------------------
  subroutine loadFitsArray_4d_wa_r64(inArray, hypercube)!!!, dataType)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT), intent(in) :: inArray
    real(double), dimension(:,:,:,:), pointer :: hyperCube

    integer(int32), dimension(:), pointer ::&
      dimVector
    integer :: numDimensions

!    integer(int8),  dimension(:,:,:,:), pointer ::&
!      arrayDataInt8
!    integer(int16), dimension(:,:,:,:), pointer ::&
!      arrayDataInt16
!    integer(int32), dimension(:,:,:,:), pointer ::&
!      arrayDataInt32
!    real(single),   dimension(:,:,:,:), pointer ::&
!      arrayDataReal32
!    real(double),   dimension(:,:,:,:), pointer ::&
!      arrayDataReal64

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)

    numDimensions = size(dimVector)

    select case(numDimensions)
    case(1, 2, 3)
call error('arrayRankTooSmall', '')
    case(4)
call error('notAvailableYet', "Reading of 4-dimensional arrays doesn't &
&seem to be available in the DAL yet.")

!      allocate(hypercube(dimVector(0), dimVector(1), dimVector(2)&
!      , dimVector(3)))
!
!      if (dataType(inArray) .eq. INTEGER8) then
!        arrayDataInt8    => int8Array4Data(inArray)
!        hyperCube = real(arrayDataInt8, double)
!      else if (dataType(inArray) .eq. INTEGER16) then
!        arrayDataInt16   => int16Array4Data(inArray)
!        hyperCube = real(arrayDataInt16, double)
!      else if (dataType(inArray) .eq. INTEGER32) then
!        arrayDataInt32   => int32Array4Data(inArray)
!        hyperCube = real(arrayDataInt32, double)
!      else if (dataType(inArray) .eq. REAL32) then
!        arrayDataReal32 => real32Array4Data(inArray)
!        hyperCube = real(arrayDataReal32, double)
!      else ! this gives a DAL error if data type is none of the above.
!        arrayDataReal64 => real64Array4Data(inArray)
!        hyperCube = arrayDataReal64
!      end if
hyperCube = 0.0 ! keeps the compiler from issuing a warning.

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine loadFitsArray_4d_wa_r64

  !---------------------------------------------------------------------
  subroutine mergeTables_old(inSetNames, inTabNames, outSetName, outTabName&
  , indexColName, indexFileName, status)
  !---------------------------------------------------------------------
  ! Merges up to 10 tables into 1. At present the input tables must have the same number of rows.

!    use sscmisc

    implicit none

    character(*), intent(in)            :: inSetNames(:),&
                                           inTabNames(size(inSetNames))
    character(*), intent(out)           :: outSetName
    character(*), intent(out), optional :: outTabName
    character(*), intent(in),  optional :: indexColName,&
                                           indexFileName
    integer,      intent(out), optional :: status

    character(255) ::&
      localOutTabName
    character(10) ::&
      setNumStr
    integer ::&
      numSets,&
      maxNumSets,&
      numColsInFirstTab,&
      numRowsInFirstTab,&
      localStatus,&
      colNum,&
      setNum,&
      numColsTakenFromThisSet
    integer, parameter ::&
      numColNameChars = 20,&
      numDigits = 1
    character(numColNameChars - (1 + numDigits)), allocatable ::&
      maxListOfColNames(:)
    ! The funny length of these strings is because we are going to construct the column names in the outset by adding an underscore plus the set number to the input column names, and we want output names that are (i) less than numColNameChars long and (ii) unique within the table.
    character(numColNameChars) ::&
      outColName
    logical, allocatable ::&
      columnIsPresentInAll(:)
    type(DataSetT) ::&
      firstSet,&
      set,&
      outSet
    type(TableT) ::&
      firstTab,&
      tab,&
      outTab
    integer, allocatable ::&
      indices(:),&
      allIndices(:,:)
integer, pointer :: xi(:)

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    numSets = size(inSetNames)
    if (numSets .le. 1) then
      if (present(status)) status = -1
  return
    end if
!    maxNumSets = 10**numDigits
maxNumSets = 10
    if (numSets .gt. maxNumSets) then
      if (present(status)) status = -2
  return
    end if

    if (.not. present(outTabName)) localOutTabName = 'COMBINED'

    firstSet = dataSet(inSetNames(1), READ)
    firstTab = table(firstSet, inTabNames(1))
    numColsInFirstTab = numberOfColumns(firstTab)
    numRowsInFirstTab = numberOfRows(firstTab)
    allocate(maxListOfColNames(numColsInFirstTab)&
    , columnIsPresentInAll(numColsInFirstTab))

    columnIsPresentInAll = .true. ! default
    maxListOfColNames   = ''     ! default

    do colNum = 1, numColsInFirstTab
      if (len(name(column(firstTab, colNum - 1))) .gt. numColNameChars) then
        localStatus = localStatus + 1
        columnIsPresentInAll(colNum) = .false.
      else
        maxListOfColNames(colNum) = name(column(firstTab, colNum - 1))
      end if
    end do
    call release(firstSet)

    do setNum = 2, numSets
      set = dataSet(inSetNames(setNum), READ)
      tab = table(set, inTabNames(setNum))

      do colNum = 1, numColsInFirstTab
      if (.not. columnIsPresentInAll(colNum)) cycle
        if (.not. hasColumn(tab, maxListOfColNames(colNum)))&
          columnIsPresentInAll(colNum) = .false.
      end do

      call release(set)
    end do

    if (.not. any(columnIsPresentInAll)) then
      if (present(status)) status = -3
  return
    end if

    if (.not. present(indexColName) .and. present(indexFileName)) then
      allocate(indices(numColsInFirstTab)&
      , allIndices(numColsInFirstTab, numSets))
      call loadIndices(indexFileName, allIndices, localStatus)
      if (localStatus .lt. 0) then
        if (present(status)) status = localStatus
  return
      end if

    else
!      indices = indicize(numRowsInFirstTab) - 1
xi => indicize(numRowsInFirstTab)
indices = xi - 1
deallocate(xi)
    end if

    outSet = dataSet(outSetName, CREATE)
    outTab = addTable(outSet, trim(localOutTabName), numRowsInFirstTab)
    do setNum = 1, numSets
      write(setNumStr, '(i10.10)') setNum-1
      set = dataSet(inSetNames(setNum), READ)
      tab = table(set, inTabNames(setNum))
      if (numberOfRows(tab) .ne. numRowsInFirstTab) then
        if (present(status)) status = -4
  return
      end if

      if (present(indexColName)) then
        call readColData(tab, indexColName, indices)
      else if (present(indexFileName)) then
        indices = allIndices(:, setNum)
      end if

      numColsTakenFromThisSet = 0
      do colNum = 1, numColsInFirstTab
      if (.not. columnIsPresentInAll(colNum)) cycle
        numColsTakenFromThisSet = numColsTakenFromThisSet + 1
        outColName = trim(maxListOfColNames(colNum))//'_'&
        //setNumStr(len(setNumStr) + 1 - numDigits:)
        call copyColumnInCorrectOrder(tab, maxListOfColNames(colNum)&
        , indices, outTab, outColName)
      end do

      if (numColsTakenFromThisSet .le. 0) localStatus = localStatus + 2

      call release(set)
    end do
    call release(outSet)

    if (present(status)) status = localStatus

!deallocate(indices)

  end subroutine mergeTables_old

  !---------------------------------------------------------------------
  subroutine mergeTables(inSetNames, inTabNames, outSetName&
  , outTabName, inColNames, indexColName, indexFileName, status)
  !---------------------------------------------------------------------
  ! Merges up to 10 tables into 1. At present the input tables must have the same number of rows.

!    use sscmisc

    implicit none

    character(*), intent(in)            :: inSetNames(:),&
                                           inTabNames(size(inSetNames))
    character(*), intent(in)            :: outSetName
    character(*), intent(in),  optional :: outTabName
    character(*), intent(in),  optional :: indexColName,&
                                           indexFileName,&
                                           inColNames(:)
    integer,      intent(out), optional :: status

    character(255) ::&
      localOutTabName
    character(10) ::&
      setNumStr
    integer ::&
      numSets,&
      maxNumSets,&
      maxNumCols,&
      numRowsInFirstTab,&
      localStatus,&
      colNum,&
      setNum,&
      numColsTakenFromThisSet,&
      startSet
    integer, parameter ::&
      numColNameChars = 20,&
      numDigits = 1
    character(numColNameChars - (1 + numDigits)), allocatable ::&
      maxListOfColNames(:)
    ! The funny length of these strings is because we are going to construct the column names in the outset by adding an underscore plus the set number to the input column names, and we want output names that are (i) less than numColNameChars long and (ii) unique within the table.
    character(numColNameChars) ::&
      outColName
    logical, allocatable ::&
      columnIsPresentInAll(:)
    type(DataSetT) ::&
      firstSet,&
      set,&
      outSet
    type(TableT) ::&
      firstTab,&
      tab,&
      outTab
    integer, allocatable ::&
      indices(:),&
      allIndices(:,:)
integer, pointer :: xi(:)

    !-------------------------------------------------------------------
    ! Start

print *, '000'

    localStatus = 0
    numSets = size(inSetNames)
    if (numSets .le. 1) then
      if (present(status)) status = -1
  return
    end if
!    maxNumSets = 10**numDigits
maxNumSets = 10
    if (numSets .gt. maxNumSets) then
      if (present(status)) status = -2
  return
    end if

print *, '111'
    if (present(outTabName)) then
      localOutTabName = outTabName
    else
      localOutTabName = 'COMBINED'
    end if

    firstSet = dataSet(inSetNames(1), READ)
    firstTab = table(firstSet, inTabNames(1))
    numRowsInFirstTab = numberOfRows(firstTab)

    if (present(inColNames)) then
print *, '222'
      maxNumCols = size(inColNames)
      allocate(maxListOfColNames(maxNumCols)&
      , columnIsPresentInAll(maxNumCols))
  
      columnIsPresentInAll = .true. ! default
      maxListOfColNames    = inColNames
      startSet = 1

    else
      maxNumCols = numberOfColumns(firstTab)
      allocate(maxListOfColNames(maxNumCols)&
      , columnIsPresentInAll(maxNumCols))

      columnIsPresentInAll = .true. ! default
      maxListOfColNames   = ''     ! default

      do colNum = 1, maxNumCols
        if (len(name(column(firstTab, colNum - 1))) .gt. numColNameChars) then
          localStatus = localStatus + 1
          columnIsPresentInAll(colNum) = .false.
        else
          maxListOfColNames(colNum) = name(column(firstTab, colNum - 1))
        end if
      end do
      startSet = 2
    end if
    call release(firstSet)
print *, '333'

    do setNum = startSet, numSets
!print *, 'setNum, inSetNames(setNum), inTabNames(setNum): ', setNum, trim(inSetNames(setNum)), trim(inTabNames(setNum))
      set = dataSet(inSetNames(setNum), READ)
!print *, '  aaa'
      tab = table(set, inTabNames(setNum))
!print *, '  bbb'

      do colNum = 1, maxNumCols
!print *, '    ---'
!print *, 'colNum, maxListOfColNames(colNum): ', colNum, maxListOfColNames(colNum)
      if (.not. columnIsPresentInAll(colNum)) cycle
!print *, '    --+'
        if (.not. hasColumn(tab, maxListOfColNames(colNum)))&
          columnIsPresentInAll(colNum) = .false.
!print *, '    column ', maxListOfColNames(colNum), ' IsPresent = ', columnIsPresentInAll(colNum)
      end do

!print *, '  ccc'
      call release(set)
!print *, '  ddd'
    end do
print *, '444'

    if (.not. any(columnIsPresentInAll)) then
      if (present(status)) status = -3
  return
    end if
print *, '555'

    allocate(indices(numRowsInFirstTab))
print *, '666'
    if (.not. present(indexColName) .and. present(indexFileName)) then
print *, '777'
      allocate(allIndices(numRowsInFirstTab, numSets))
print *, '888'
      call loadIndices(indexFileName, allIndices, localStatus)
print *, '999'
      if (localStatus .lt. 0) then
print *, 'oooo! status=', status
        if (present(status)) status = localStatus
  return
      end if
print *, '000'

    else
!      indices = indicize(numRowsInFirstTab) - 1
xi => indicize(numRowsInFirstTab)
indices = xi - 1
deallocate(xi)
    end if

    outSet = dataSet(outSetName, CREATE)
print *, '111 localOutTabName=', '*'//trim(localOutTabName)//'*'
    outTab = addTable(outSet, trim(localOutTabName), numRowsInFirstTab)
print *, '222'
    do setNum = 1, numSets
print *, ''
print *, '  aaa'
      write(setNumStr, '(i10.10)') setNum-1
print *, '  bbb'
      set = dataSet(inSetNames(setNum), READ)
print *, '  ccc'
      tab = table(set, inTabNames(setNum))
print *, '  ddd'
      if (numberOfRows(tab) .ne. numRowsInFirstTab) then
        if (present(status)) status = -4
print *, '  oooo! numberOfRows(tab) (',numberOfRows(tab),') .ne. &
&numRowsInFirstTab (',numRowsInFirstTab,'). status=',status
  return
      end if
print *, '  eee'

      if (present(indexColName)) then
        call readColData(tab, indexColName, indices)
      else if (present(indexFileName)) then
        indices = allIndices(:, setNum)
      end if
print *, '  fff'

      numColsTakenFromThisSet = 0
print *, '  ggg'
      do colNum = 1, maxNumCols
      if (.not. columnIsPresentInAll(colNum)) cycle
        numColsTakenFromThisSet = numColsTakenFromThisSet + 1
        outColName = trim(maxListOfColNames(colNum))//'_'&
        //setNumStr(len(setNumStr) + 1 - numDigits:)
        call copyColumnInCorrectOrder(tab, maxListOfColNames(colNum)&
        , indices, outTab, outColName)
      end do

print *, '  hhh'
      if (numColsTakenFromThisSet .le. 0) localStatus = localStatus + 2

print *, '  iii'
      call release(set)
    end do
print *, '333'
    call release(outSet)

    if (present(status)) status = localStatus
print *, 'eeee nnnn dddd'
deallocate(indices)

  end subroutine mergeTables

  !---------------------------------------------------------------------
  subroutine loadIndices(indexFileName, allIndices, status)
  !---------------------------------------------------------------------

!    use get_free_io_unit_mod

    implicit none

    character(*), intent(in) :: indexFileName
    integer, intent(out) :: allIndices(:,:), status

    integer :: ioUnit, numRows, numSets, lineNum, readStatus
    logical :: fileExists, fileOpened
    character(255) :: buffer(size(allIndices, 1))

    !-------------------------------------------------------------------
    ! Start

!print *, '  start of loadIndices.'
    status = 0

    numRows = size(allIndices, 1)
    numSets = size(allIndices, 2)
    inquire(file=indexFileName, exist=fileExists, opened=fileOpened)
    if (.not. fileExists) then
      status = -5
  return
    else if (fileOpened) then
      status = -6
  return
    end if

!    call get_free_io_unit(ioUnit)
call getFreeIoUnit(ioUnit)
    open(ioUnit, file=indexFileName, status='old', form='formatted'&
    , action='read', pad='yes')
      lineNum = 0
      do
        lineNum = lineNum + 1
!print *, ''
!print *, '    lineNum=', lineNum
        read(ioUnit, '(a255)', iostat=readStatus) buffer(lineNum)
!        read(ioUnit, *, iostat=readStatus) buffer(lineNum)
      if (readStatus .ne. 0) exit
!print *, lineNum, '*'//buffer(lineNum)//'* **'//trim(buffer(lineNum))//'**'
        call pickBufferApart(trim(buffer(lineNum)), allIndices(lineNum,:), status)
!print *, '    pickBufferApart status=', status
!print *, '    1st index=', allIndices(lineNum,1)
!print *, '    2nd index=', allIndices(lineNum,2)
      if (status .lt. 0) exit
      end do    
    close(ioUnit)

!print *, ''
!print *, '  end of loadIndices. status=', status

  end subroutine loadIndices

  !---------------------------------------------------------------------
  subroutine pickBufferApart(buffer, indices, status)
  !---------------------------------------------------------------------

!    use sscmisc

    implicit none

    character(*), intent(in) :: buffer
    integer, intent(out) :: indices(:), status

    integer :: tabPosition, nonTabStart, nonTabEnd, numSets, setNum

    !-------------------------------------------------------------------
    ! Start

!print *, 'start of pickBufferApart'

    status = 0

!print *, '**'//trim(buffer)//'**'
    numSets = size(indices)

    if (len(buffer) .le. 0) then
      status = -1
  return
    end if
    tabPosition = index(buffer, achar(9))
    if (tabPosition .le. 0) then ! no tabs found.
      status = -2
  return
    end if
    if (tabPosition .eq. 1) then ! the buffer starts with a tab. Not allowed.
      status = -3
  return
    end if

    nonTabStart = 1 ! hopefully it is the start of an integer.
    setNum = 0
!print *, '      len(buffer):', len(buffer)
!print *, '      setNum:', setNum
    do
!print *, '      '
!print *, '      Start of do.'
      tabPosition = index(buffer(nonTabStart:), achar(9))
!print *, '      tabPosition:', tabPosition
      if (tabPosition .eq. len(buffer)) then ! not allowed.
        status = -4
!print *, '      ooo status -4, tabPosition .eq. len(buffer).'
  return
      else if (tabPosition .le. 0) then ! no further tabs found.
        nonTabEnd = len(buffer)
!print *, '      tabPosition .le. 0. nonTabEnd:', nonTabEnd
      else
        nonTabEnd = tabPosition - 1
!print *, '      tabPosition .gt. 0. nonTabEnd:', nonTabEnd
      end if

!print *, '      nonTabStart .gt. nonTabEnd? exit loop if so:', (nonTabStart .gt. nonTabEnd)
    if (nonTabStart .gt. nonTabEnd) exit ! because there are no more integers.

      setNum = setNum + 1
!print *, '      setNum, numSets:', setNum, numSets
      if (setNum .gt. numSets) then
        status = -5
!print *, '      ooo status -5, setNum .gt. numSets.'
  return
      end if

!print *, '      just before call to strToInt. buffer(nonTabStart:nonTabEnd), ie buffer('&
!,nonTabStart,':',nonTabEnd,'), is ', buffer(nonTabStart:nonTabEnd)
      call strToInt(buffer(nonTabStart:nonTabEnd), indices(setNum), status)
!print *, '      indices(setNum), status:', indices(setNum), status
  if (status .lt. 0) return
      indices(setNum) = indices(setNum) - 1

    if (tabPosition .le. 0 .or. nonTabEnd .ge. len(buffer)) exit
      nonTabStart = tabPosition + 1
!print *, '      nonTabStart:', nonTabStart
    end do

    if (setNum .lt. numSets) then
      status = -6
  return
    end if

  end subroutine pickBufferApart

  !---------------------------------------------------------------------
  subroutine copyColumnInCorrectOrder(inTab, inColName&
  , indices, outTab, outColName)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: inTab, outTab
    character(*), intent(in) :: inColName, outColName
    integer(int32), intent(in) :: indices(:)

    type(ColumnT) :: inCol, outCol
    integer(int8),  dimension(:), pointer ::&
      inColDataInt8, outColDataInt8
    integer(int16), dimension(:), pointer ::&
      inColDataInt16, outColDataInt16
    integer(int32), dimension(:), pointer ::&
      inColDataInt32, outColDataInt32
    real(single),   dimension(:), pointer ::&
      inColDataReal32, outColDataReal32
    real(double),   dimension(:), pointer ::&
      inColDataReal64, outColDataReal64

    !-------------------------------------------------------------------
    ! Start

    inCol  = column( inTab,  inColName)

    select case(dataType(inCol))
    case(INTEGER8)
      outCol = addColumn(outTab, outColName, INTEGER8)
      inColDataInt8    => int8Data(inCol)
      outColDataInt8   => int8Data(outCol)
      outColDataInt8   = inColDataInt8(indices)
    case(INTEGER16)
      outCol = addColumn(outTab, outColName, INTEGER16)
      inColDataInt16   => int16Data(inCol)
      outColDataInt16  => int16Data(outCol)
      outColDataInt16  = inColDataInt16(indices)
    case(INTEGER32)
      outCol = addColumn(outTab, outColName, INTEGER32)
      inColDataInt32   => int32Data(inCol)
      outColDataInt32  => int32Data(outCol)
      outColDataInt32  = inColDataInt32(indices)
    case(REAL32)
      outCol = addColumn(outTab, outColName, REAL32)
      inColDataReal32  => real32Data(inCol)
      outColDataReal32 => real32Data(outCol)
      outColDataReal32 = inColDataReal32(indices)
    case default ! this gives a DAL error if data type is not REAL64 either:
      outCol = addColumn(outTab, outColName, REAL64)
      inColDataReal64  => real64Data(inCol)
      outColDataReal64 => real64Data(outCol)
      outColDataReal64 = inColDataReal64(indices)
    end select

    call release(inCol)
    call release(outCol)

  end subroutine copyColumnInCorrectOrder

  !---------------------------------------------------------------------
  subroutine saveOutImageReal32(newImage, imageName, newImageName&
  , xStartArg, yStartArg)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) ::&
      newImage(:,:)
    character(*), intent(in) ::&
      imageName, newImageName
    integer(int16), intent(in), optional :: xStartArg, yStartArg

    type(DataSetT) :: inSet, outSet
    type(ArrayT) :: inArray, outArray
!    type(BlockT) :: inBlock, outBlock
!    type(DssT) ::&
!      my_dss
    real(single), dimension(:,:), pointer ::&
      arrayDataReal32
    integer(int32), dimension(:), pointer ::&
      dimVector
    integer(int16) :: xStart, yStart, xFinis, yFinis

    !-------------------------------------------------------------------
    ! Start:

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    inSet = dataSet(imageName, READ)
    inArray = array(inSet, 0, READ)
!    inBlock = block(inArray)
    outSet = dataSet(newImageName, CREATE)
    allocate(dimVector(0:1))
    dimVector(0) = size(newImage, 1)
    dimVector(1) = size(newImage, 2)
    xFinis = xStart + size(newImage, 1) - 1
    yFinis = yStart + size(newImage, 2) - 1
    outArray = addArray(outSet, 'PRIMARY', REAL32, dimVector, position=0)
!    outBlock = block(outArray)
    call setAttributes(attributable(outSet), attributable(inSet))

    call copyVitalArrayKwds(inArray, outArray)
    ! This is done instead of a total copy of all array-header keywords because
    ! doing the latter stuffs up the DSS.

    arrayDataReal32 => real32Array2Data(outArray)
    arrayDataReal32 = 0.0
    arrayDataReal32(xStart-1:xFinis-1, yStart-1:yFinis-1) = newImage

    call release(outSet)
    call release(inSet)

    deallocate(dimVector)

  end subroutine saveOutImageReal32

  !---------------------------------------------------------------------
  subroutine saveOutImageInt32(newImage, imageName, newImageName&
  , xStartArg, yStartArg)
  !---------------------------------------------------------------------

    implicit none

    integer(int32), intent(in) ::&
      newImage(:,:)
    character(*), intent(in) ::&
      imageName, newImageName
    integer(int16), intent(in), optional :: xStartArg, yStartArg

    type(DataSetT) :: inSet, outSet
    type(ArrayT) :: inArray, outArray
!    type(BlockT) :: inBlock, outBlock
!    type(DssT) ::&
!      my_dss
    integer(int32), dimension(:,:), pointer ::&
      arrayDataInt32
    integer(int32), dimension(:), pointer ::&
      dimVector
    integer(int16) :: xStart, yStart, xFinis, yFinis

    !-------------------------------------------------------------------
    ! Start:

    if (present(xStartArg)) then
      xStart = xStartArg
    else
      xStart = 1
    end if
    if (present(yStartArg)) then
      yStart = yStartArg
    else
      yStart = 1
    end if

    inSet = dataSet(imageName, READ)
    inArray = array(inSet, 0, READ)
!    inBlock = block(inArray)
    outSet = dataSet(newImageName, CREATE)
    allocate(dimVector(0:1))
    dimVector(0) = size(newImage, 1)
    dimVector(1) = size(newImage, 2)
    xFinis = xStart + size(newImage, 1) - 1
    yFinis = yStart + size(newImage, 2) - 1
    outArray = addArray(outSet, 'PRIMARY', INTEGER32, dimVector, position=0)
!    outBlock = block(outArray)
    call setAttributes(attributable(outSet), attributable(inSet))

    call copyVitalArrayKwds(inArray, outArray)
    ! This is done instead of a total copy of all array-header keywords because
    ! doing the latter stuffs up the DSS.

    arrayDataInt32 => int32Array2Data(outArray)
    arrayDataInt32 = 0
    arrayDataInt32(xStart-1:xFinis-1, yStart-1:yFinis-1) = newImage

    call release(outSet)
    call release(inSet)

    deallocate(dimVector)

  end subroutine saveOutImageInt32

  !---------------------------------------------------------------------
  subroutine copyVitalArrayKwds(inArray, outArray)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT), intent(in) :: inArray, outArray

    integer :: status

    !-------------------------------------------------------------------
    ! Start

    call copyAttribute(inArray, 'SLCTEXPR', outArray, status)
    call copyAttribute(inArray, 'HDUCLASS', outArray, status)
    call copyAttribute(inArray, 'HDUCLAS1', outArray, status)
    call copyAttribute(inArray, 'HDUCLAS2', outArray, status)
    call copyAttribute(inArray, 'HDUVERS1', outArray, status)

    call copyAttribute(inArray, 'CTYPE1', outArray, status)
    call copyAttribute(inArray, 'CRPIX1', outArray, status)
    call copyAttribute(inArray, 'CRVAL1', outArray, status)
    call copyAttribute(inArray, 'CUNIT1', outArray, status)
    call copyAttribute(inArray, 'CDELT1', outArray, status)
    call copyAttribute(inArray, 'CTYPE2', outArray, status)
    call copyAttribute(inArray, 'CRPIX2', outArray, status)
    call copyAttribute(inArray, 'CRVAL2', outArray, status)
    call copyAttribute(inArray, 'CUNIT2', outArray, status)
    call copyAttribute(inArray, 'CDELT2', outArray, status)

    call copyAttribute(inArray, 'EXPOSURE', outArray, status)

  end subroutine copyVitalArrayKwds

  !---------------------------------------------------------------------
  subroutine andKeywordsSetScalar(masterSet, set, readAlsoFromMaster, commonKwds)
  !---------------------------------------------------------------------

!*** see also the routine checkAndCopyKeywords().

    implicit none

    type(DataSetT), intent(in)           :: masterSet, set
    logical,        intent(in), optional :: readAlsoFromMaster
    character(20),  pointer,    optional :: commonKwds(:)

    logical :: localReadFromMaster
    type(DataSetT) :: sets(1)

    !-------------------------------------------------------------------
    ! Start

    if (present(readAlsoFromMaster)) then
      localReadFromMaster = readAlsoFromMaster ! ie, include masterSet in
      ! the ANDing process.
    else
      localReadFromMaster = .false. ! assume the keywords in masterSet
      ! are to be completely overwritten by the ANDed kwds from the other sets.
    end if

    sets(1) = set

    if (present(commonKwds)) then
      call andKeywordsSetVector(masterSet, sets, localReadFromMaster&
        , commonKwds)
    else
      call andKeywordsSetVector(masterSet, sets, localReadFromMaster)
    end if

  end subroutine andKeywordsSetScalar

  !---------------------------------------------------------------------
  subroutine andKeywordsTabScalar(masterTab, tab, readAlsoFromMaster, commonKwds)
  !---------------------------------------------------------------------

!*** see also the routine checkAndCopyKeywords().

    implicit none

    type(TableT),   intent(in)           :: masterTab, tab
    logical,        intent(in), optional :: readAlsoFromMaster
    character(20),  pointer,    optional :: commonKwds(:)

    logical :: localReadFromMaster
    type(TableT) :: tabs(1)

    !-------------------------------------------------------------------
    ! Start

    if (present(readAlsoFromMaster)) then
      localReadFromMaster = readAlsoFromMaster ! ie, include masterTab in
      ! the ANDing process.
    else
      localReadFromMaster = .false. ! assume the keywords in masterTab
      ! are to be completely overwritten by the ANDed kwds from the other tabs.
    end if

    tabs(1) = tab

    if (present(commonKwds)) then
      call andKeywordsTabVector(masterTab, tabs, localReadFromMaster&
        , commonKwds)
    else
      call andKeywordsTabVector(masterTab, tabs, localReadFromMaster)
    end if

  end subroutine andKeywordsTabScalar

  !---------------------------------------------------------------------
  subroutine andKeywordsSetVector(masterSet, sets, readAlsoFromMaster, commonKwds)
  !---------------------------------------------------------------------

!*** see also the routine checkAndCopyKeywords().

    implicit none

    type(DataSetT), intent(in)           :: masterSet, sets(:)
    logical,        intent(in), optional :: readAlsoFromMaster
    character(20),  pointer,    optional :: commonKwds(:)

    integer :: numSets, numKwds, setNum, kwdNum, startSet, i
    logical :: localReadFromMaster
    character(20), allocatable :: kwdName(:)
    logical, allocatable :: isCommonKwd(:)
    type(AttributeT) :: gaugeAttribute, testAttribute

    !-------------------------------------------------------------------
    ! Start

    numSets = size(sets)
  if (numSets <= 0) return

    if (present(readAlsoFromMaster)) then
      localReadFromMaster = readAlsoFromMaster ! ie, include masterSet in
      ! the ANDing process.
    else
      localReadFromMaster = .false. ! assume the keywords in masterSet
      ! are to be completely overwritten by the ANDed kwds from the other sets.
    end if

    ! First job is to get a list of common keyword names. After that is
    ! done, the values will be checked.

    ! Load starting list of keywords:
    if (localReadFromMaster) then ! may as well load it from master:
      numKwds = numberOfAttributes(masterSet)
      allocate(kwdName(numKwds), isCommonKwd(numKwds))
      do kwdNum = 1, numKwds
        kwdName(kwdNum) = name(attribute(masterSet, kwdNum - 1))
      end do

      startSet = 1

    else ! load the starting list of kwds from the first set:
      numKwds = numberOfAttributes(sets(1))
      allocate(kwdName(numKwds), isCommonKwd(numKwds))
      do kwdNum = 1, numKwds
        kwdName(kwdNum) = name(attribute(sets(1), kwdNum - 1))
      end do

      startSet = 2
    end if

    ! Now check to see if the keyword names are present in the other sets:
    isCommonKwd = .true. ! default
    do setNum = startSet, numSets
      do kwdNum = 1, numKwds
        if (.not. hasAttribute(sets(setNum), kwdName(kwdNum)))&
          isCommonKwd(kwdNum) = .false.
      end do
    end do

    ! Now we check the values:

    KWDLOOP: do kwdNum = 1, numKwds
    if (.not. isCommonKwd(kwdNum)) cycle KWDLOOP

      if (localReadFromMaster) then
        gaugeAttribute = attribute(masterSet, kwdName(kwdNum))
      else
        gaugeAttribute = attribute(sets(1), kwdName(kwdNum))
      end if

      do setNum = startSet, numSets
        testAttribute = attribute(sets(setNum), kwdName(kwdNum))
        if (.not. attributeValuesAreEqual(testAttribute, gaugeAttribute)) then
          isCommonKwd(kwdNum) = .false.
    cycle KWDLOOP
        end if
      end do
    end do KWDLOOP

    ! The final task is to copy the common set into masterSet:

    if (.not. localReadFromMaster) then ! replace attributes in masterSet
    ! by those in sets(1). It is more convenient to do it this way (ie,
    ! copy everything, then delete non-common kwds) than to delete all
    ! the masterSet attributes, then copy the common ones.

      call setAttributes(attributable(masterSet), attributable(sets(1)))
    end if

    do kwdNum = 1, numKwds
      if (.not. isCommonKwd(kwdNum))&
        call deleteAttribute(masterSet, kwdName(kwdNum))
    end do

    if (present(commonKwds)) then
      allocate(commonKwds(count(isCommonKwd)))
      i = 0
      do kwdNum = 1, numKwds
      if (.not. isCommonKwd(kwdNum)) cycle
        i = i + 1
        commonKwds(i) = kwdName(kwdNum)
      end do
    end if

  end subroutine andKeywordsSetVector

  !---------------------------------------------------------------------
  subroutine andKeywordsTabVector(masterTab, tabs, readAlsoFromMaster, commonKwds)
  !---------------------------------------------------------------------

!*** see also the routine checkAndCopyKeywords().

    implicit none

    type(TableT),   intent(in)           :: masterTab, tabs(:)
    logical,        intent(in), optional :: readAlsoFromMaster
    character(20),  pointer,    optional :: commonKwds(:)

    integer :: numTabs, numKwds, tabNum, kwdNum, startTab, i
    logical :: localReadFromMaster
    character(20), allocatable :: kwdName(:)
    logical, allocatable :: isCommonKwd(:)
    type(AttributeT) :: gaugeAttribute, testAttribute

    !-------------------------------------------------------------------
    ! Start

    numTabs = size(tabs)
  if (numTabs <= 0) return

    if (present(readAlsoFromMaster)) then
      localReadFromMaster = readAlsoFromMaster ! ie, include masterTab in
      ! the ANDing process.
    else
      localReadFromMaster = .false. ! assume the keywords in masterTab
      ! are to be completely overwritten by the ANDed kwds from the other tabs.
    end if

    ! First job is to get a list of common keyword names. After that is
    ! done, the values will be checked.

    ! Load starting list of keywords:
    if (localReadFromMaster) then ! may as well load it from master:
      numKwds = numberOfAttributes(masterTab)
      allocate(kwdName(numKwds), isCommonKwd(numKwds))
      do kwdNum = 1, numKwds
        kwdName(kwdNum) = name(attribute(masterTab, kwdNum - 1))
      end do

      startTab = 1

    else ! load the starting list of kwds from the first set:
      numKwds = numberOfAttributes(tabs(1))
      allocate(kwdName(numKwds), isCommonKwd(numKwds))
      do kwdNum = 1, numKwds
        kwdName(kwdNum) = name(attribute(tabs(1), kwdNum - 1))
      end do

      startTab = 2
    end if

    ! Now check to see if the keyword names are present in the other sets:
    isCommonKwd = .true. ! default
    do tabNum = startTab, numTabs
      do kwdNum = 1, numKwds
        if (.not. hasAttribute(tabs(tabNum), kwdName(kwdNum)))&
          isCommonKwd(kwdNum) = .false.
      end do
    end do

    ! Now we check the values:

    KWDLOOP: do kwdNum = 1, numKwds
    if (.not. isCommonKwd(kwdNum)) cycle KWDLOOP

      if (localReadFromMaster) then
        gaugeAttribute = attribute(masterTab, kwdName(kwdNum))
      else
        gaugeAttribute = attribute(tabs(1), kwdName(kwdNum))
      end if

      do tabNum = startTab, numTabs
        testAttribute = attribute(tabs(tabNum), kwdName(kwdNum))
        if (.not. attributeValuesAreEqual(testAttribute, gaugeAttribute)) then
          isCommonKwd(kwdNum) = .false.
    cycle KWDLOOP
        end if
      end do
    end do KWDLOOP

    ! The final task is to copy the common set into masterTab:

    if (.not. localReadFromMaster) then ! replace attributes in masterTab
    ! by those in tabs(1). It is more convenient to do it this way (ie,
    ! copy everything, then delete non-common kwds) than to delete all
    ! the masterTab attributes, then copy the common ones.

      call setAttributes(attributable(masterTab), attributable(tabs(1)))
    end if

    do kwdNum = 1, numKwds
      if (.not. isCommonKwd(kwdNum))&
        call deleteAttribute(masterTab, kwdName(kwdNum))
    end do

    if (present(commonKwds)) then
      allocate(commonKwds(count(isCommonKwd)))
      i = 0
      do kwdNum = 1, numKwds
      if (.not. isCommonKwd(kwdNum)) cycle
        i = i + 1
        commonKwds(i) = kwdName(kwdNum)
      end do
    end if

  end subroutine andKeywordsTabVector

  !---------------------------------------------------------------------
  subroutine sortTable(tab, sortExpression)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: tab
    character(*), intent(in) :: sortExpression

    integer(int32) :: numRows, i
    character(20) :: sortColName
    type(ColumnT) :: sortCol
    real(single), pointer :: sortColData(:)
    integer, allocatable :: index(:)

    !-------------------------------------------------------------------
    ! Start

    numRows = numberOfRows(tab)
    if (numRows <= 0) then
      call warning('noRowsInTable','')
  return
    end if

    sortColName = 'RESULT'
    if (hasColumn(tab, sortColName)) then
call error('resultColAlreadyExists', "fits_utils wants to make a column &
&in the table to hold the result of its column arithmetic. &
&However it is not yet smart enough to choose an unused column name &
&- it insists on it being called RESULT. But alas, there is already &
&a column of that name in the file.")
    end if
!*** write routine to generate a spare column name.

    call initExprColGen(sortExpression, sortColName, REAL32)
    call doExprColGen(tab, tab)

    sortCol = column(tab, sortColName)
    sortColData => real32Data(sortCol)
    allocate(index(numRows))
    call quick_sort(sortColData, index)
    call deleteColumn(tab, sortColName)

    do i = 0, numberOfColumns(tab) - 1
      call sortColumn(tab, i, index)
    end do

  end subroutine sortTable

  !---------------------------------------------------------------------
  subroutine sortColumn(tab, colNum, index)
  !---------------------------------------------------------------------
!*** overload it so that (i) index becomes optional (sort if not present) and (ii) can alternately give col name.

    implicit none

    type(TableT), intent(in) :: tab
    integer(int32), intent(in) :: colNum, index(:)

    integer(int32) :: numRows, row, stringLength
    type(ColumnT) :: col
    integer(int8),    pointer :: int8Ptr(:)
    integer(int16),   pointer :: int16Ptr(:)
    integer(int32),   pointer :: int32Ptr(:)
    real(single),     pointer :: real32Ptr(:)
    real(double),     pointer :: real64Ptr(:)
    logical(bool), pointer :: boolPtr(:)
    integer(int8),    allocatable :: tempInt8Vector(:)
    integer(int16),   allocatable :: tempInt16Vector(:)
    integer(int32),   allocatable :: tempInt32Vector(:)
    real(single),     allocatable :: tempReal32Vector(:)
    real(double),     allocatable :: tempReal64Vector(:)
    logical(bool), allocatable :: tempBoolVector(:)
    integer(int16), parameter :: tempStringLength = 255
    character(tempStringLength), allocatable :: tempStringVector(:)

    !-------------------------------------------------------------------
    ! Start

    numRows = numberOfRows(tab)
    if (size(index) .ne. numRows)&
call error('noRowsInTable','')

    if (numRows <= 0) then
      call warning('','')
  return
    end if

    col = column(tab, colNum)
    select case(dataType(col))
    case(INTEGER8)
      int8Ptr => int8Data(col)
      allocate(tempInt8Vector(numRows))
      do row = 0, numRows - 1
        tempInt8Vector(row+1) = int8Ptr(index(row+1)-1)
      end do
      int8Ptr = tempInt8Vector

    case(INTEGER16)
      int16Ptr => int16Data(col)
      allocate(tempInt16Vector(numRows))
      do row = 0, numRows - 1
        tempInt16Vector(row+1) = int16Ptr(index(row+1)-1)
      end do
      int16Ptr = tempInt16Vector

    case(INTEGER32)
      int32Ptr => int32Data(col)
      allocate(tempInt32Vector(numRows))
      do row = 0, numRows - 1
        tempInt32Vector(row+1) = int32Ptr(index(row+1)-1)
      end do
      int32Ptr = tempInt32Vector

    case(REAL32)
      real32Ptr => real32Data(col)
      allocate(tempReal32Vector(numRows))
      do row = 0, numRows - 1
        tempReal32Vector(row+1) = real32Ptr(index(row+1)-1)
      end do
      real32Ptr = tempReal32Vector

    case(REAL64)
      real64Ptr => real64Data(col)
      allocate(tempReal64Vector(numRows))
      do row = 0, numRows - 1
        tempReal64Vector(row+1) = real64Ptr(index(row+1)-1)
      end do
      real64Ptr = tempReal64Vector

    case(BOOLEAN)
      boolPtr => boolData(col)
      allocate(tempBoolVector(numRows))
      do row = 0, numRows - 1
        tempBoolVector(row+1) = boolPtr(index(row+1)-1)
      end do
      boolPtr = tempBoolVector

    case(STRING)
      stringLength = len(stringCell(col, 0))
      if (stringLength > tempStringLength) then
        call warning('','')
        stringLength = tempStringLength
      end if

      allocate(tempStringVector(numRows))
      tempStringVector = ''
      do row = 0, numRows - 1
        tempStringVector(row+1) = stringCell(col, index(row+1)-1)
      end do
      do row = 0, numRows - 1
        call setStringCell(col, row&
          , tempStringVector(row+1)(:stringLength))
      end do

    case default
call error('','')
    end select

    call release(col)

  end subroutine sortColumn

  !---------------------------------------------------------------------
  function addOrOpenColumnDataReal32(tab, colName, units, comment)&
  result(dataPointer)
  !---------------------------------------------------------------------
  ! (Notes about how the function ought to be called, with pointer deallocation in mind - these hang off how the dal release() call works, whuch I have to check out some more.)

    implicit none

    type(TableT), intent(in)           :: tab
    character(*), intent(in)           :: colName
    character(*), intent(in), optional :: units, comment
    real(single), pointer :: dataPointer(:)

    character(255) :: localUnits, localComment

    !-------------------------------------------------------------------
    ! Start

    if (present(units)) then
      localUnits = units
    else
      localUnits = ''
    end if

    if (present(comment)) then
      localComment = comment
    else
      localComment = ''
    end if

    if (hasColumn(tab, colName)) then
      dataPointer => real32Data(column(tab, colName))
    else
      dataPointer => real32Data(addColumn(tab, colName, REAL32, localUnits&
        , comment=localComment))
    end if

  end function addOrOpenColumnDataReal32

  !---------------------------------------------------------------------
  function addOrOpenColumnDataInt32(tab, colName, units, comment)&
  result(dataPointer)
  !---------------------------------------------------------------------
  ! (Notes about how the function ought to be called, with pointer deallocation in mind - these hang off how the dal release() call works, whuch I have to check out some more.)

    implicit none

    type(TableT), intent(in)           :: tab
    character(*), intent(in)           :: colName
    character(*), intent(in), optional :: units, comment
    integer(int32), pointer :: dataPointer(:)

    character(255) :: localUnits, localComment

    !-------------------------------------------------------------------
    ! Start

    if (present(units)) then
      localUnits = units
    else
      localUnits = ''
    end if

    if (present(comment)) then
      localComment = comment
    else
      localComment = ''
    end if

    if (hasColumn(tab, colName)) then
      dataPointer => int32Data(column(tab, colName))
    else
      dataPointer => int32Data(addColumn(tab, colName, INTEGER32, localUnits&
        , comment=localComment))
    end if

  end function addOrOpenColumnDataInt32

  !---------------------------------------------------------------------
  function addOrOpenColumn(tab, colName, dataType, units, comment)&
  result(col)
  !---------------------------------------------------------------------
  ! (Notes about how the function ought to be called, with pointer deallocation in mind - these hang off how the dal release() call works, whuch I have to check out some more.)

    implicit none

    type(TableT),   intent(in)           :: tab
    character(*),   intent(in)           :: colName
    integer(int32), intent(in)           :: dataType
    character(*),   intent(in), optional :: units, comment
    type(ColumnT) :: col

    character(255) :: localUnits, localComment

    !-------------------------------------------------------------------
    ! Start

    if (present(units)) then
      localUnits = units
    else
      localUnits = ''
    end if

    if (present(comment)) then
      localComment = comment
    else
      localComment = ''
    end if

    if (hasColumn(tab, colName)) then
      col = column(tab, colName)
    else
      col = addColumn(tab, colName, dataType, localUnits&
        , comment=localComment)
    end if

  end function addOrOpenColumn

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module fits_utils
