module sort_mod
  !  Non-recursive stack versions of quicksort from N.Wirth's Pascal
  !  book, "Algorithms + data structures = programs".
  !  Converted to Fortran90 by Alan Miller, CSIRO.
  !  Module constructed by Clive Page, Leicester University.
  implicit none

  interface quick_sort
     module procedure &
          quick_sort_c,  quick_sort_d,  quick_sort_i,  quick_sort_r, &
          quick_sort_co, quick_sort_do, quick_sort_io, quick_sort_ro
  end interface

contains

  subroutine quick_sort_c(array)
    character(*), intent(INOUT)    :: array(:)
    character(LEN=len(array)) :: temp, x

    !     Local Variables

    integer  :: i, j, k, l, r, s, stackl(30), stackr(30)

    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
                if (array(i) >= x) exit  
                i = i + 1
             end do

             do                          ! Search from upper end
                if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
                temp    = array(i)
                array(i) = array(j)
                array(j) = temp
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
                exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do
    return
  end subroutine quick_sort_c

  subroutine quick_sort_d(array)
    double precision, intent(INOUT)    :: array(:)
    double precision :: temp, x

    !     Local Variables

    integer  :: i, j, k, l, r, s, stackl(30), stackr(30)

    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
                if (array(i) >= x) exit  
                i = i + 1
             end do

             do                          ! Search from upper end
                if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
                temp    = array(i)
                array(i) = array(j)
                array(j) = temp
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
                exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do
    return
  end subroutine quick_sort_d

  subroutine quick_sort_i(array)
    integer, intent(INOUT)    :: array(:)
    integer :: temp, x

    !     Local Variables

    integer  :: i, j, k, l, r, s, stackl(30), stackr(30)

    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
                if (array(i) >= x) exit  
                i = i + 1
             end do

             do                          ! Search from upper end
                if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
                temp    = array(i)
                array(i) = array(j)
                array(j) = temp
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
                exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do
    return
  end subroutine quick_sort_i

  subroutine quick_sort_r(array)
    real, intent(INOUT)    :: array(:)
    real     :: temp, x

    !     Local Variables

    integer  :: i, j, k, l, r, s, stackl(30), stackr(30)

    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
                if (array(i) >= x) exit  
                i = i + 1
             end do

             do                          ! Search from upper end
                if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
                temp    = array(i)
                array(i) = array(j)
                array(j) = temp
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
                exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do
    return
  end subroutine quick_sort_r

  subroutine quick_sort_co(array, order)
    character(*), intent(INOUT)    :: array(:)
    integer, intent(OUT) :: order(:)  ! returns original position
    character(LEN=len(array)) :: temp, x

    !     Local Variables

    integer  :: i, j, k, l, r, s, stackl(15), stackr(15), ww

    do j = lbound(array,1),ubound(array,1)
       order(j) = j
    end do
    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
                if (array(i) >= x) exit
                i = i + 1
             end do

             do                          ! Search from upper end
                if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
                temp    = array(i)
                ww   = order(i)
                array(i) = array(j)
                order(i) = order(j)
                array(j) = temp
                order(j) = ww
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
                exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do
    return
  end subroutine quick_sort_co

  subroutine quick_sort_do_new(array, order)
    double precision, intent(INOUT)    :: array(:)
    integer, intent(OUT) :: order(:)  ! returns original position
    double precision :: temp, x

    !     Local Variables

    integer  :: i, j, k, l, r, s, stackl(15), stackr(15), ww

!print *, 'lbound(array,1),ubound(array,1),lbound(order,1),ubound(order,1)'&
!,lbound(array,1),ubound(array,1),lbound(order,1),ubound(order,1) 

    do j = lbound(array,1),ubound(array,1)
       order(j) = j
    end do
    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

print *, 's=', s
    do
print *, ''
print *, '  start loop A'
       l = stackl(s)
       r = stackr(s)
print *, '  l = stackl(s=',s,')=', l
print *, '  r = stackr(s=',s,')=', r
       s = s - 1
print *, '  s=', s

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
print *, ''
print *, '    start loop B'
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)
print *, '    i = l=', i
print *, '    j = r=', j
print *, '    k = (l+r) / 2=', k
print *, '    x = array(k=',k,')=', x
!!
          !     REPEAT UNTIL I > J.

          do
print *, ''
print *, '      start loop C'
             do                          ! Search from lower end
print *, ''
print *, '        start loop D1 - low search loop'
print *, '        i = ', i
print *, '        array(i) = ', array(i)
print *, '        exit loop D1 (ie, array(i) >= x)?', (array(i) >= x)
             if (array(i) >= x) exit
!!
                i = i + 1
             end do

             do                          ! Search from upper end
print *, ''
print *, '        start loop D2 - high search loop'
print *, '        j = ', j
print *, '        array(j) = ', array(j)
print *, '        exit loop D2 (ie, x >= array(j))?', (x >= array(j))
             if( x >= array(j)) exit
!!
                j = j - 1
             end do

print *, '      exit loop C (ie, i=',i,' > j=',j,')?', (i > j)
          if (i > j) exit

             ! Otherwise, swap positions i & j
print *, '      Otherwise, swap positions i & j'

             temp    = array(i)
             ww   = order(i)
             array(i) = array(j)
             order(i) = order(j)
             array(j) = temp
!!
             order(j) = ww
print *, '      increment i & j'
             i = i + 1
             j = j - 1
          end do

          if (j-l >= r-i) then
print *, '      j-l=',j-l,' >= r-i=',r-i
             if (l < j) then
print *, '      l=',l,' < j=',j,': increment s'
                s = s + 1
                stackl(s) = l
                stackr(s) = j
print *, '      s now =', s
print *, '      stackl(s=',s,') = i =', i
print *, '      stackr(s=',s,') = r =', r
             end if
             l = i
          else
print *, '      j-l=',j-l,' < r-i=',r-i
             if (i < r) then
print *, '      i=',i,' < r=',r,': increment s'
                s = s + 1
                stackl(s) = i
                stackr(s) = r
print *, '      s now =', s
print *, '      stackl(s=',s,') = i =', i
print *, '      stackr(s=',s,') = r =', r
             end if
             r = j
          end if

print *, '    exit loop B (ie, l=',l,' >= r=',r,')?', (l >= r)
       if(l >= r) exit
       end do
print *, '  exit loop A (ie, s == 0)?', (s == 0)
    if (s == 0) exit
    end do

  end subroutine quick_sort_do_new

  subroutine quick_sort_do(array, order)
    double precision, intent(INOUT)    :: array(:)
    integer, intent(OUT) :: order(:)  ! returns original position
    double precision :: temp, x

    !     Local Variables

    integer  :: i, j, k, l, r, s, stackl(15), stackr(15), ww

    do j = lbound(array,1),ubound(array,1)
       order(j) = j
    end do
    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
             if (array(i) >= x) exit
                i = i + 1
             end do

             do                          ! Search from upper end
             if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
                temp    = array(i)
                ww   = order(i)
                array(i) = array(j)
                order(i) = order(j)
                array(j) = temp
                order(j) = ww
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
          exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do
    return
  end subroutine quick_sort_do

  subroutine quick_sort_io(array, order)
    integer, intent(INOUT)    :: array(:)
    integer, intent(OUT) :: order(:)  ! returns original position
    integer  :: temp, x

    !     Local Variables

    integer  :: i, j, k, l, r, s, stackl(15), stackr(15), ww

    do j = lbound(array,1),ubound(array,1)
       order(j) = j
    end do
    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
                if (array(i) >= x) exit
                i = i + 1
             end do

             do                          ! Search from upper end
                if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
                temp    = array(i)
                ww   = order(i)
                array(i) = array(j)
                order(i) = order(j)
                array(j) = temp
                order(j) = ww
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
                exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do
    return
  end subroutine quick_sort_io

  subroutine quick_sort_ro(array, order)
    real, intent(INOUT)    :: array(:)
    integer, intent(OUT) :: order(:)  ! returns original position
    real     :: temp, x

    !     Local Variables

    integer  :: i, j, k, l, r, s, stackl(15), stackr(15), ww

    do j = lbound(array,1),ubound(array,1)
       order(j) = j
    end do
    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
                if (array(i) >= x) exit
                i = i + 1
             end do

             do                          ! Search from upper end
                if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
                temp    = array(i)
                ww   = order(i)
                array(i) = array(j)
                order(i) = order(j)
                array(j) = temp
                order(j) = ww
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
                exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do
    return
  end subroutine quick_sort_ro

end module sort_mod
