#!/usr/local/bin/perl -w

my ($testflag);

sub eimsimchain {

  use SAS;

  my $taskname = "eimsimchain";

  my ($command, $new_fvtmp, $i, $stopfile);#, $instexpid, $prefix, $band);

  # Prints all output to stderr:
  select(STDERR);

  #......................................................................
  # Read and check non-slave parameters:

  my $num_entries = parameterCount('obsidroots');
SAS::error('noObsIds', "There must be at least one member in the "
."list --obsidroots")
  if ($num_entries <= 0);

  my @obsid_roots = ();
  foreach $i (0 .. $num_entries-1) {
    $obsid_roots[$i] = stringParameter('obsidroots', $i)
  }

  my $stream_num   = intParameter('streamnumber');
  my $nfields      = intParameter('nfields');
  my $start_num    = intParameter('startatn');
  my $apply_corr   = booleanParameter('applycorr');
  my $next_on_fail = booleanParameter('nextonfail');
  $testflag        = booleanParameter('astest');

  $stream_num =~ s/^0+//; # chop off any leading zeros

  #......................................................................
  # General setup:

  my $lockfile = "lock_$stream_num";
SAS::error('lockFileExists', "Lock file $lockfile exists!") if (-e "$lockfile");
SAS::error('writeLockFileFailed', "Couldn't make lock file $lockfile.")
  if (!$testflag && system("touch $lockfile"));

  my $old_fvtmp = $ENV{"FVTMP"};
  if (defined($old_fvtmp) && -e $old_fvtmp){
    $new_fvtmp = "$old_fvtmp/$stream_num";
  } else {
    $new_fvtmp = "./temp/$stream_num";
  }

SAS::error('mkdirFailed', "Can't mkdir $new_fvtmp")
  if (!-e "$new_fvtmp" && system("mkdir -p $new_fvtmp"));

  $ENV{"FVTMP"} = $new_fvtmp;

  #......................................................................
  # Main routine:

  my $stop_num = $start_num + $nfields - 1;
  for($i=$start_num; $i<=$stop_num; $i++) {

    #....................................................................
    # Check stop file:

    $stopfile = "stop_$stream_num";
    if (-e "$stopfile") {
      unlink("$lockfile");
SAS::error('gracefulStop', "Stop file $stopfile found.");
    }

    $command = "eimsimslave "
    ."obsidroots='".join(' ', @obsid_roots)."' "
    ."entrystage=".stringParameter('entrystage')." "
    ."finalstage=".stringParameter('finalstage')." "
    ."prdssubdir=".stringParameter('prdssubdir')." "
    ."simopsubdir=".stringParameter('simopsubdir')." "
    ."streamnumber=$stream_num "
    ."idnumber=$i "
    ."templateset=".stringParameter('templateset')." "
    ."energyfraction=".realParameter('energyfraction')." "
    ."psfmaxcoarseflux=".realParameter('psfmaxcoarseflux')." "
    ."applycorr=".booleanParameter('applycorr')." "
    ."ecffile=".stringParameter('ecffile')." "
    ."likemin=".realParameter('likemin')." "
    ."mlmin=".realParameter('mlmin')." "
    ."cleanup=".booleanParameter('cleanup')." "
    ."astest=".booleanParameter('astest')." "
    ;

    print "\ninvoke $command\n\n";
    if (system("$command")) {
SAS::error('eimsimSlaveFailed', "Couldn't run eimsimslave, iteration $i.")
      if (!$next_on_fail);

      SAS::warning('eimsimSlaveFailed', "Couldn't run eimsimslave, iteration $i.");
    }
  }

  ENDD:

  if (!$apply_corr) {
    print "Don't forget! Param --applycorr was not set. Your figures "
    ."for completeness and/or reliability will therefore not be optimum. "
    ."To get the best figures you should run eimsimfinalize, "
    ."check the correction models against the results, then run "
    ."eimsimchain with --entrystage=compare and --applycorr=yes.\n";
  }

  unlink("$lockfile");

}

  #---------------------------------------------------------------------
  sub run {
  #---------------------------------------------------------------------
  # $testflag should be defined globally.

    my ($command) = @_;
    my $status = 0;

    print "\ninvoke $command\n\n";
    if (!$testflag) {
      $status = system($command);
    }

  return $status;
  }

