#!/usr/local/bin/perl -w

# badEcfKeys

my ($testflag);

sub eimsimslave {

  use SAS;

  my $taskname = "eimsimslave";

  my ($i, $key, $obsid_root, $prod_dir, @prod_files, $already_found_obsid);
  my ($file, $obsid, $inst_uc, @files, @chunks, $instexpid, $simop_dir);
  my (%cifs, %merged_det_list, %htmlSummary, %padmasksets, %boxmapsets);
  my (%mlsets, %filtered_mlsets, %sens_maps, $evlist, $filter, %filters);
  my ($band, %expmaps, %ctsimages, %sim_images, %bkgmaps, %rateimages);
  my (%ximaerrcorrsets, %yimaerrcorrsets, %ratecorrsets, $band_id);
  my ($edetect_first_stage, $edetect_last_stage, $edetect_astest);
  my ($command, $padmaskset, $cif, $expmapset, $rateimage, $ecf);
  my ($ctsimage, $bkgmapset, $imageset, $mlset, $ximaerrcorrset);
  my ($yimaerrcorrset, $ratecorrset, $filtered_mlset, $sens_map);

  # Prints all output to stderr:
  select(STDERR);

  #......................................................................
  # Read and check parameters:

  my $num_entries = parameterCount('obsidroots');
SAS::error('noObsIds', "There must be at least one member in the "
."list --obsidroots")
  if ($num_entries <= 0);

  my @obsid_roots = ();
  foreach $i (0 .. $num_entries-1) {
    $obsid_roots[$i] = stringParameter('obsidroots', $i)
  }

  my $first_stage      = stringParameter('entrystage');
  my $last_stage       = stringParameter('finalstage');
  my $prod_subdir      = stringParameter('prdssubdir');
  my $simop_subdir     = stringParameter('simopsubdir');
  my $stream_num       = intParameter('streamnumber');
  my $id_number        = intParameter('idnumber');
  my $template_set     = stringParameter('templateset');
  my $energyfraction   = realParameter('energyfraction');
  my $psfmaxcoarseflux = realParameter('psfmaxcoarseflux');
  my $apply_corr       = booleanParameter('applycorr');
  my $ecf_file         = stringParameter('ecffile');
#  my $do_eddington     = booleanParameter('witheddington');
my $do_eddington = 'no';
  my $likemin          = realParameter('likemin');
  my $mlmin            = realParameter('mlmin');
  my $cleanup          = booleanParameter('cleanup');
  $testflag            = booleanParameter('astest');

  #......................................................................
  # Misc setups:

  my %ecf_info = &getEcfs($ecf_file); # contains band edge info too.
  my @used_insts_uc = ();
  foreach $key (keys(%ecf_info)) {
    if ($key =~ /^(M1|M2|PN)/) {
      push @used_insts_uc, $key;
    }
  }

  my @idbandlo = @{$ecf_info{'lo'}};
  my @idbandhi = @{$ecf_info{'hi'}};
#  my @used_bands = (1 .. @idbandlo);
my @used_bands = @{$ecf_info{'id'}};

  my %inst_inter = ('M1'=>'emos1', 'M2'=>'emos2', 'PN'=>'epn');

  #......................................................................
  # Set up obsid hash:

  my @expmap_vector = ();
  my %obs_ids = ();
  foreach $obsid_root (@obsid_roots) {
    $prod_dir  = "$obsid_root/$prod_subdir";
SAS::error('noProductSubdirectory', "Directory $prod_dir not found.")
    if (!-d "$prod_dir");

    @prod_files = `ls -1 $prod_dir`;

SAS::error('noProducts', "No product files in $prod_dir.")
    if (@prod_files <= 0);

    $already_found_obsid = 0;
    foreach $file (@prod_files) {
      chomp($file);

      if ($file =~ /^P(\d{10})/) {
        if ($already_found_obsid) {
          if ($1 != $obsid) {
SAS::error('tooManyPathsThisObsid', "Only 1 --obsidroots entry "
."allowed per obsid.");
          }
        } else {
          $obsid = $1;
          $already_found_obsid = 1;
        }

      } # end if $file =~ /^P(\d{10})/
    } # end foreach $file (@prod_files)

SAS::error('noProducts', "No product files in $prod_dir.")
    if (!$already_found_obsid);

    $obs_ids{$obsid_root} = $obsid;
  } # end foreach $obsid_root (@obsid_roots)

  #......................................................................
  # Setup instexpid arrays from existence of any files for that obsid and
  # instrument:

  my %temp_hash = ();
  my %instexpids = ();
  foreach $obsid_root (@obsid_roots) {
    $prod_dir  = "$obsid_root/$prod_subdir";
    $obsid = $obs_ids{$obsid_root};

    %temp_hash = ();
    foreach $inst_uc (@used_insts_uc) {
      @files = `ls -1 $prod_dir/P$obsid$inst_uc*`;
      foreach $file (@files) {
        @chunks = split('\/', $file);
        if ($chunks[$#chunks] =~ /^P${obsid}${inst_uc}((S|U)\d{3})/) {
          $instexpid = "$inst_uc$1";
          $temp_hash{$instexpid} = 1;
        }
      }
    }

    push @{$instexpids{$obsid}}, keys(%temp_hash)
  }

  #......................................................................
  # Setup file names:

  $stream_num =~ s/^0+//; # chop off any leading zeros
  $id_number  =~ s/^0+//; # chop off any leading zeros
  my $stream_num_str = '0'x(2-length($stream_num))."$stream_num";
  my $id_number_str  = '0'x(4-length($id_number)) ."$id_number";
  my $id_str = "S$stream_num_str"."F$id_number_str";
  my $temp_set = "$stream_num_str"."_temp.ds";

  my $simsrclistset   = "Sim_srcList_$id_str.fits";

  foreach $obsid_root (@obsid_roots) {
    $prod_dir  = "$obsid_root/$prod_subdir";
    $simop_dir = "$obsid_root/$simop_subdir";

    $obsid = $obs_ids{$obsid_root};

    $cifs{$obsid} = "$prod_dir/P$obsid"."OBX000CALIND0000.FIT";
    $merged_det_list{$obsid} = "$simop_dir/P$id_str"."OBX000OBSMLI0000.FIT";
    $htmlSummary{$obsid}     = "$simop_dir/P$id_str"."OBX000OBSMLI0000.HTM";

    foreach $instexpid (@{$instexpids{$obsid}}) {
      $inst_uc = substr($instexpid, 0, 2);

      $padmasksets{$obsid}{$instexpid} = "$simop_dir/Sim_$instexpid"
        ."_paddedMask.fits";

      $boxmapsets{$obsid}{$instexpid} = "$simop_dir/P$id_str$instexpid"
        ."EBMSLI0000.FIT";

      $mlsets{$obsid}{$instexpid} = "$simop_dir/P$id_str$instexpid"
        ."EMSRLI0000.FIT";

      $filtered_mlsets{$obsid}{$instexpid} = "$mlsets{$obsid}{$instexpid}.filt";

      $sens_maps{$obsid}{$instexpid} = "P$obsid$instexpid"."SNSMAP8000.FIT";
#***** should go in $simop_dir. Modify my_esensmap stuff to make this possible.

      if ($inst_uc eq 'PN') {
        $evlist = "$prod_dir/P$obsid$instexpid"."PIEVLI0000.FIT";
      } else {
        $evlist = "$prod_dir/P$obsid$instexpid"."MIEVLI0000.FIT";
      }
SAS::error('noEvListSet', "Couldn't find $evlist") if (!-e "$evlist");

      $filter=&getExposureProperty($evlist, 'FILTER');

      unless(uc($filter) =~ /^(THIN|MEDIUM|THICK|OPEN)/ ){
SAS::error('badFilter', "Could not match filter for energy conversion factor.");
      }
      $filter = $1;

      $filters{$obsid}{$instexpid} = $filter;

      foreach $band (@used_bands) {
        $expmaps{$obsid}{$instexpid}{$band} = "$prod_dir/P$obsid$instexpid"
          ."EXPMAP$band"."000.FIT";

        $ctsimages{$obsid}{$instexpid}{$band} = "$simop_dir/"
          ."Sim_$instexpid"."_CTSIMG_$band.fits";

        $sim_images{$obsid}{$instexpid}{$band} = "$simop_dir/P$id_str$instexpid"
          ."IMAGE_$band"."000.FIT";

        $bkgmaps{$obsid}{$instexpid}{$band} = "$simop_dir/"
          ."Sim_$instexpid"."_BKGMAP_$band.fits";

        $rateimages{$obsid}{$instexpid}{$band}
          = "$simop_dir/Sim_$id_str$instexpid"
          ."_rateImage_$band".".fits";
      } # end foreach $band (@used_bands)
    } # end foreach $instexpid (@{$instexpids{$obsid}})
  } # end foreach $obsid_root (@obsid_roots)

  foreach $inst_uc ('M1','M2','PN') {
    $ximaerrcorrsets{$inst_uc} = 'xima_err_corrector.ds';
    $yimaerrcorrsets{$inst_uc} = 'yima_err_corrector.ds';
    $ratecorrsets{   $inst_uc} = "rate_corrector_$inst_inter{$inst_uc}.ds";
  }

  #......................................................................
  # Decide where to go:

  if ($first_stage eq 'makeimsimlist') {
    goto MAKESIMLIST;
  } elsif($first_stage eq 'makerateimg') {
    goto MAKERATEIMG;
  } elsif($first_stage eq 'makectsimg') {
    goto MAKECTSIMG;
  } elsif($first_stage eq 'poissonize') {
    goto POISSONIZE;
  } elsif ($first_stage eq 'edetect'
#        || $first_stage eq 'emask'
        || $first_stage eq 'boxlocal'
        || $first_stage eq 'splinemap'
        || $first_stage eq 'boxmap'
        || $first_stage eq 'mldetect'
        || $first_stage eq 'finalize') {
    goto EDETECT;
#  } elsif($first_stage eq 'eddington') {
#    goto EDDINGTON;
  } elsif($first_stage eq 'correct') {
    goto CORRECT;
  } elsif($first_stage eq 'ratetofluxrand') {
    goto RATETOFLUXRAND;
  } elsif($first_stage eq 'compare') {
    goto COMPARE;
  } elsif($first_stage eq 'addsimrates') {
    goto ADDSIMRATES;
  } elsif($first_stage eq 'filter') {
    goto FILTER;
  } elsif($first_stage eq 'sensmaps') {
    goto SENSMAPS;
  } elsif($first_stage eq 'srcmatch') {
    goto SRCMATCH;
  } elsif($first_stage eq 'endd') {
    goto ENDD;
  } else {
SAS::error('badEntryStage', "Unrecognized entry stage $first_stage.");
  }

  #......................................................................
  # Start main routine:

  MAKESIMLIST:

  # Params accessed:
###### all these are now out of date.
  #	- templateset

  #	- streamnumber
  #	- idnumber
  #	- entrystage
  #	- finalstage
  #	- astest

SAS::error('noTemplate', "Couldn't find $template_set.") if (!-e "$template_set");

  $command = "eimsim "
  ."function=srclist "
  ."templateset=$template_set "
  ."simsrclistset=$simsrclistset "
  ."obsids='".join(' ', values(%obs_ids))."' "
  ;

SAS::error('makeSimListFailed', "Couldn't make sim source list.")
  if (&run("$command"));

  $command = "eimsim "
  ."function=addcols "
  ."tab=$simsrclistset".":SRCLIST "
  ."newcolumnlist='STREAM_N FIELD_N' "
  ."newcolvaluelist='$stream_num $id_number' "
  ;

SAS::error('addColsFailed', "Couldn't add stream & field numbers to "
."sim source list.")
  if (&run("$command"));

#  $command = "fcalc "
#  ."infile=$simsrclistset"."'[SRCLIST]' "
#  ."outfile=$simsrclistset "
#  ."clname=STREAM_N "
#  ."expr='STREAM_N=$stream_num' "
#  ."clobber=yes"
#  ;
#
#&quit("Couldn't add stream number to sim source list.", 12)
#if (&run("$command"));
#
#  $command = "fcalc "
#  ."infile=$simsrclistset"."'[SRCLIST]' "
#  ."outfile=$simsrclistset "
#  ."clname=FIELD_N "
#  ."expr='FIELD_N=$id_number' "
#  ."clobber=yes"
#  ;
#
#&quit("Couldn't add field number to sim source list.", 12)
#if (&run("$command"));

  goto ENDD if ($last_stage eq 'makeimsimlist');

  MAKERATEIMG:

  # Params accessed:
  #	- energyfraction
  #	- psfmaxcoarseflux
  #     - ecffile

  #	- prdsdir
  #	- interdir
  #	- obsnumber
  #	- instexpids
  #	- streamnumber
  #	- idnumber
  #	- entrystage
  #	- finalstage
  #	- astest

SAS::error('noSimSrcListSet', "Couldn't find $simsrclistset.")
  if (!$testflag && !-e "$simsrclistset");

  foreach $obsid_root (@obsid_roots) {
    $obsid = $obs_ids{$obsid_root};

    $cif = $cifs{$obsid};
SAS::error('noCif', "Couldn't find CIF $cif.") if (!-e "$cif");
    $ENV{'SAS_CCF'} = "$cif";

    foreach $instexpid (@{$instexpids{$obsid}}) {
      $inst_uc = substr($instexpid, 0, 2);
      $filter = $filters{$obsid}{$instexpid};
      $padmaskset = $padmasksets{$obsid}{$instexpid};
SAS::error('noPadMask', "Couldn't find $padmaskset.")
      if (!$testflag && !-e "$padmaskset");

      foreach $band_id (0 .. $#used_bands) {
        $band = $used_bands[$band_id];
        $expmapset  = $expmaps{$obsid}{$instexpid}{$band};
SAS::error('noExpMap', "Couldn't find $expmapset.") if (!-e "$expmapset");
        $rateimage  = $rateimages{$obsid}{$instexpid}{$band};
        $ecf = $ecf_info{$inst_uc}{$filter}{$band};
SAS::error('badEcfKeys', "Unrecognized inst $inst_uc, filter "
."$filter or band $band")
        if (!defined($ecf));

        $command = "eimsim "
        ."function=image "
        ."expmapset=$expmapset "
        ."simsrclistset=$simsrclistset "
        ."energyfraction=$energyfraction "
        ."psfmaxcoarseflux=$psfmaxcoarseflux "
        ."ecf=$ecf "
        ."elo=$idbandlo[$band_id] "
        ."ehi=$idbandhi[$band_id] "
        ."rateimageset=$rateimage "
        ."paddedmaskset=$padmaskset "
        ."obsids='$obsid' "
        ."idband=$band "
        ;

SAS::error('makeRateImageFailed', "Couldn't make sim image.")
        if (&run("$command"));
      }
    }
  }

  goto ENDD if ($last_stage eq 'makerateimg');

  MAKECTSIMG:

#??  # Params accessed:

  foreach $obsid_root (@obsid_roots) {
    $obsid = $obs_ids{$obsid_root};
    foreach $instexpid (@{$instexpids{$obsid}}) {
      foreach $band (@used_bands) {
        $expmapset   = $expmaps{   $obsid}{$instexpid}{$band};
SAS::error('noExpMap', "Couldn't find $expmapset.") if (!-e "$expmapset");
        $rateimage   = $rateimages{$obsid}{$instexpid}{$band};
SAS::error('noRateImage', "Couldn't find $rateimage.")
        if (!$testflag && !-e "$rateimage");

        $ctsimage    = $ctsimages{$obsid}{$instexpid}{$band};
        $bkgmapset   = $bkgmaps{  $obsid}{$instexpid}{$band};
SAS::error('noBkgmap', "Couldn't find $bkgmapset.")
        if (!$testflag && !-e "$bkgmapset");

        $command = "eimsim "
        ."function=makectsimg "
        ."expmapset=$expmapset "
        ."bkgmapset=$bkgmapset "
        ."rateimageset=$rateimage "
        ."avcountsimageset=$ctsimage "
        ;

SAS::error('makeCountsImageFailed', "Couldn't make the average counts image.")
        if (&run("$command"));
      }
    }
  }

  goto ENDD if ($last_stage eq 'makectsimg');

  POISSONIZE:

### can be done by imageutils task poissonize as soon as this is part of the sas.

  # Params accessed:
  #	- prdsdir
  #	- interdir
  #	- obsnumber
  #	- instexpids
  #	- streamnumber
  #	- idnumber
  #	- entrystage
  #	- finalstage
  #	- astest

  foreach $obsid_root (@obsid_roots) {
    $obsid = $obs_ids{$obsid_root};
    foreach $instexpid (@{$instexpids{$obsid}}) {
      foreach $band (@used_bands) {
        $imageset    = $sim_images{$obsid}{$instexpid}{$band};
        $ctsimage    = $ctsimages{ $obsid}{$instexpid}{$band};
SAS::error('noAvCountsImage', "Couldn't find $ctsimage.")
        if (!$testflag && !-e "$ctsimage");

        $command = "eimsim "
        ."function=poissonize "
        ."avcountsimageset=$ctsimage "
        ."imageset=$imageset "
        ;

SAS::error('poissonizeFailed', "Couldn't poissonize average counts image.")
        if (&run("$command"));
      }
    }
  }

  goto ENDD if ($last_stage eq 'poissonize');

  EDETECT:

  # Params accessed:
  #	- prdsdir
  #	- interdir
  #	- obsnumber
  #	- instexpids
  #	- streamnumber
  #	- idnumber
  #	- entrystage
  #	- finalstage
  #	- astest

  if  ($first_stage eq 'boxlocal'
    || $first_stage eq 'splinemap'
    || $first_stage eq 'boxmap'
    || $first_stage eq 'mldetect') {

    $edetect_first_stage = $first_stage;
  } else {
    $edetect_first_stage = 'boxlocal';
  }

  if  ($last_stage eq 'boxlocal'
    || $last_stage eq 'splinemap'
    || $last_stage eq 'boxmap'
    || $last_stage eq 'mldetect') {

    $edetect_last_stage = $last_stage;
  } else {
    $edetect_last_stage = 'mldetect';
  }

  foreach $obsid_root (@obsid_roots) {
    $obsid = $obs_ids{$obsid_root};
    $prod_dir = "$obsid_root/$prod_subdir";
    $simop_dir = "$obsid_root/$simop_subdir";

    if ($testflag) {
      $edetect_astest = 'yes';
    } else {
      $edetect_astest = 'no';
    }

    foreach $instexpid (@{$instexpids{$obsid}}) {
      $command = "eimsimdetect "
      ."prdsdir=$prod_dir "
      ."simopdir=$simop_dir "
      ."instexpid=$instexpid "
      ."obsid=$obsid "
      ."streamnumber=$stream_num "
      ."idnumber=$id_number "
      ."entrystage=$edetect_first_stage "
      ."finalstage=$edetect_last_stage "
      ."likemin=$likemin "
      ."mlmin=$mlmin "
      ."bandlist='".join(' ', @used_bands)."' "
      ."astest=$edetect_astest "
      ;

### should pass it the filenames and the ecf file.
### ...but! The ecfs are read from the image!

      print "\ninvoke $command\n\n";

SAS::error('edetectFailed', "Couldn't run eimsimdetect chain.")
      if (system($command));
      # Not using &run() here because we want eimsim_e5detect to be run whether
      # $params{'astest'} is set or not. $params{'astest'} is passed to
      # eimsim_e5detect and so is also effective therein.
    }
  }

  goto ENDD if ($last_stage eq 'edetect'
        || $last_stage eq 'emask'
        || $last_stage eq 'boxlocal'
        || $last_stage eq 'splinemap'
        || $last_stage eq 'boxmap'
        || $last_stage eq 'mldetect'
#        || $last_stage eq 'finalize'
  );

#  EDDINGTON: # correct for Eddington bias.
#
#  # Params accessed:
#  # ???
#
#  foreach $obsid_root (@obsid_roots) {
#    $obsid = $obs_ids{$obsid_root};
#    foreach $instexpid (@{$instexpids{$obsid}}) {
#      $mlset = "$mlsets{ $obsid}{$instexpid}";
#
##      if (&boolTranslate($do_eddington)) {
#      if ($do_eddington) {
#        $command = "eimsim "
#        ."function=eddington "
#        ."detsrclistset=$mlset "
#        ."likemin=$likemin "
#        ;
#
#&quit("Couldn't compute Eddington correction.", 21)
#if (&run("$command"));
#      }
#    }
#  }
#
#  goto ENDD if ($last_stage eq 'eddington');

  CORRECT: # correct rate, position and their respective uncerts for
  # systematic errors. Normally one would run eimsim_single twice, the
  # first time to measure the syserrors and the second time (tho only
  # from 'correct' onward) to correct for them.

  # Params accessed:
  # ???

  if ($apply_corr) {
    foreach $obsid_root (@obsid_roots) {
      $obsid = $obs_ids{$obsid_root};
      foreach $instexpid (@{$instexpids{$obsid}}) {
        $inst_uc = substr($instexpid, 0, 2);
        $mlset = $mlsets{$obsid}{$instexpid};

SAS::error('noMlSrcListSet', "Couldn't find $mlset.")
        if (!$testflag && !-e "$mlset");

        $ximaerrcorrset = $ximaerrcorrsets{$inst_uc};
SAS::error('noXimaCorrSet', "Couldn't find $ximaerrcorrset.")
        if (!$testflag && !-e "$ximaerrcorrset");

        $yimaerrcorrset = $yimaerrcorrsets{$inst_uc};
SAS::error('noYimaCorrSet', "Couldn't find $yimaerrcorrset.")
        if (!$testflag && !-e "$yimaerrcorrset");

        $ratecorrset = $ratecorrsets{$inst_uc};
SAS::error('noRateCorrSet', "Couldn't find $ratecorrset.")
        if (!$testflag && !-e "$ratecorrset");

        $command = "eimsim "
        ."function=correct "
        ."detsrclistset=$mlset "
        ."ximaerrcorrectorset=$ximaerrcorrset "
        ."yimaerrcorrectorset=$yimaerrcorrset "
        ."ratecorrectorset=$ratecorrset "
        ."usecorrected=$do_eddington "
        ;

SAS::error('correctSysErrorsFailed', "Couldn't correct systematic errors.")
        if (&run("$command"));
      }
    }
  }

  goto ENDD if ($last_stage eq 'correct');

  RATETOFLUXRAND:

  # Params accessed:
  #	- templateset
  #     - withecf
  #     - ecffile

  #     - obsnumber
  #	- interdir
  #	- instexpids
  #	- streamnumber
  #	- idnumber
  #	- entrystage
  #	- finalstage
  #	- astest

  foreach $obsid_root (@obsid_roots) {
    $obsid = $obs_ids{$obsid_root};
    foreach $instexpid (@{$instexpids{$obsid}}) {
      $mlset = $mlsets{$obsid}{$instexpid};
SAS::error('noMlSrcListSet', "Couldn't find $mlset.") if (!$testflag && !-e "$mlset");

      $command = "eimsim "
      ."function=ratetofluxrand "
      ."detsrclistset=$mlset "
      ."usecorrected=$apply_corr "
      ;

SAS::error('convertRateToFluxRandFailed', "Couldn't convert RATE to FLUXRAND.")
      if (&run("$command"));
    }
  }

  goto ENDD if ($last_stage eq 'ratetofluxrand');

  COMPARE:

  # Params accessed:
  #	- templateset

  #     - obsnumber
  #	- prdsdir
  #	- interdir
  #	- instexpids
  #	- streamnumber
  #	- idnumber
  #	- entrystage
  #	- finalstage
  #	- astest

SAS::error('noSimSrcListSet', "Couldn't find $simsrclistset.")
  if (!$testflag && !-e "$simsrclistset");

  foreach $obsid_root (@obsid_roots) {
    $obsid = $obs_ids{$obsid_root};

    $cif = $cifs{$obsid};
SAS::error('noCif', "Couldn't find CIF $cif.") if (!-e "$cif");
    $ENV{'SAS_CCF'} = "$cif";

    foreach $instexpid (@{$instexpids{$obsid}}) {
      $mlset = $mlsets{$obsid}{$instexpid};
SAS::error('noMlSrcListSet', "Couldn't find $mlset.") if (!$testflag && !-e "$mlset");

      $command = "eimsim "
      ."function=compare "
      ."detsrclistset=$mlset "
      ."simsrclistset=$simsrclistset "
      ."obsids='$obsid' "
      ."idband=0 "
      ;

SAS::error('matchDetWithSimFailed', "Couldn't compare detected to "
."simulated source lists.")
      if (&run("$command"));
    }
  }

  goto ENDD if ($last_stage eq 'compare');

  ADDSIMRATES:

  # Params accessed:

  #	- prdsdir
  #	- interdir
  #	- obsnumber
  #	- instexpids
  #	- streamnumber
  #	- idnumber
  #	- entrystage
  #	- finalstage
  #	- astest

  foreach $obsid_root (@obsid_roots) {
    $obsid = $obs_ids{$obsid_root};
    foreach $instexpid (@{$instexpids{$obsid}}) {
      $mlset = $mlsets{$obsid}{$instexpid};
SAS::error('noMlSrcListSet', "Couldn't find $mlset.") if (!$testflag && !-e "$mlset");

      $command = "eimsim "
      ."function=addsimratecols "
      ."detsrclistset=$mlset "
      ;

SAS::error('addSimRateColsFailed', "Couldn't add sim_rate columns to det list.")
      if (&run("$command"));
    }
  }

  goto ENDD if ($last_stage eq 'addsimrates');

  FILTER:

  # Params accessed:

  #     - obsnumber
  #	- instexpids
  #	- streamnumber
  #	- idnumber
  #	- entrystage
  #	- finalstage
  #	- astest

  foreach $obsid_root (@obsid_roots) {
    $obsid = $obs_ids{$obsid_root};
    foreach $instexpid (@{$instexpids{$obsid}}) {
      $mlset = $mlsets{$obsid}{$instexpid};
SAS::error('noMlSrcListSet', "Couldn't find $mlset.") if (!$testflag && !-e "$mlset");
      $filtered_mlset = $filtered_mlsets{$obsid}{$instexpid};

      $command = "fselect "
#***** try evselect to avoid cross-talk problems between streams????
      ."infile=$mlset'[SRCLIST]' "
      ."outfile=$filtered_mlset "
      ."expr='(ID_BAND==0)&&(!(isnull(RATE)))&&(RATE_ERR>0.0)' "
      ."clobber=yes "
      ;

SAS::error('filterSrcListFailed', "Couldn't filter $mlset")
      if (&run("$command"));
    }
  }

  goto ENDD if ($last_stage eq 'filter');

  SENSMAPS:

  # Params accessed:
#****

  foreach $obsid_root (@obsid_roots) {
    $obsid = $obs_ids{$obsid_root};
    foreach $instexpid (@{$instexpids{$obsid}}) {
      $sens_map = "$sens_maps{$obsid}{$instexpid}";
SAS::error('noSensMapSet', "Couldn't find $sens_map.") if (!$testflag && !-e "$sens_map");
      $filtered_mlset = $filtered_mlsets{$obsid}{$instexpid};
SAS::error('noFilteredMlSrcListSet', "Couldn't find $filtered_mlset.")
      if (!$testflag && !-e "$filtered_mlset");

      $command = "imsample "
      ."imageset=$sens_map "
      ."srclisttab=$filtered_mlset:SRCLIST "
      ."outcol=SENSY "
      ;

SAS::error('sampleSensMapFailed', "Couldn't add SENSY column to source list.")
      if (&run("$command"));
    }
  }

  goto ENDD if ($last_stage eq 'sensmaps');

  SRCMATCH:

  # Params accessed:

  # ??

  foreach $obsid_root (@obsid_roots) {
    $obsid = $obs_ids{$obsid_root};

#    $cif = $cifs{$obsid};
#&quit("Couldn't find CIF $cif.", 12.5) if (!-e "$cif");
#    $ENV{'SAS_CCF'} = "$cif";
$ENV{'SAS_CCF'} = "fred"; ######## just to test whether it really needs a cif.

    $command = "srcmatch "
    ."inputlistsets='".join(" ", values(%{$filtered_mlsets{$obsid}}))."' "
    ."outputlistset=$merged_det_list{$obsid} "
    ."htmloutput=$htmlSummary{$obsid} "
    ."maxerr=4 "
    ."systerr=4 "
    ;

SAS::error('srcmatchFailed', "Couldn't run srcmatch.")
    if (&run("$command"));
  }

#  goto ENDD if ($last_stage eq 'srcmatch');

#........................................................................
# Clean up.

  ENDD:

  if ($cleanup) {
    foreach $obsid_root (@obsid_roots) {
      $obsid = $obs_ids{$obsid_root};
      foreach $instexpid (@{$instexpids{$obsid}}) {
        foreach $band (@used_bands) {
          $imageset = $sim_images{$obsid}{$instexpid}{$band};
        next if (!-e "$imageset");

          $command = "rm $imageset";

SAS::error('rmFailed', "Couldn't delete image set $imageset.")
          if (&run("$command"));
        }
      }
    }
  }

}

  #---------------------------------------------------------------------
  sub run {
  #---------------------------------------------------------------------
  # $testflag should be defined globally.

    my ($command) = @_;
    my $status = 0;

    print "\ninvoke $command\n\n";
    if (!$testflag) {
      $status = system($command);
    }

  return $status;
  }

  #----------------------------------------------------------------------
  sub getExposureProperty {
  #----------------------------------------------------------------------

    my ($file, $kwd) = @_;
    my (@fields);

    my $command = "fkeyprint $file"."+0 $kwd | grep =";
    my @lines = `$command`;
    if (defined($lines[0])) {
      @fields = split('\'', $lines[0]);
      if (defined($fields[1]) && $fields[0] =~ /=/) {
  return $fields[1];
      } else {
  return 0;
      }
    } else {
  return 0;
    }
  }

  #---------------------------------------------------------------------
  sub getEcfs {
  #---------------------------------------------------------------------
  # Loading default parameter values:

    my ($ecf_file) = @_;
    my ($line, %bands, $filter, @ecf_line, $num_bands, $band, $inst_uc, $band_index);

SAS::error('noEcfFile', "Can't find $ecf_file") if (!-e "$ecf_file");

    open (ECFS, "$ecf_file");

    while(){
      $line=<ECFS>;
SAS::error('ecfTooFewDataLines', "Too few data lines in $ecf_file") if (!defined($line));
    next if ($line =~ /^\s*$/);
    next if ($line =~ /^#/);
    last;
    }
    if ($line =~ /^numBands\s*=\s*(\d+)/) {
      $num_bands = $1;
    } else {
SAS::error('ecfBadStructureFile', "First non-comment line must contain the number of bands.");
    }

SAS::error('ecfTooFewBands', "Too few bands in $ecf_file") if ($num_bands <= 0);

    while(){
      $line=<ECFS>;
SAS::error('ecfTooFewDataLines', "Too few data lines in $ecf_file") if (!defined($line));
    next if ($line =~ /^\s*$/);
    next if ($line =~ /^#/);
    last;
    }
    $line =~ s/^\s+//;
    chomp($line);
    my @band_line = split('\s+', $line);
SAS::error('ecfWrongNumBandIds', "There should be $num_bands entries in the band-id line in $ecf_file") if ($num_bands != @band_line);

    @{$bands{'id'}} = ();
    foreach $band (1 .. $num_bands) {
      push @{$bands{'id'}}, shift @band_line;
    }

    while(){
      $line=<ECFS>;
SAS::error('ecfTooFewDataLines', "Too few data lines in $ecf_file") if (!defined($line));
    next if ($line =~ /^\s*$/);
    next if ($line =~ /^#/);
    last;
    }
    $line =~ s/^\s+//;
    chomp($line);
    @band_line = split('\s+', $line);
SAS::error('ecfWrongNumBandLos', "There should be $num_bands entries in the band-lo-E line in $ecf_file") if ($num_bands != @band_line);

    @{$bands{'lo'}} = ();
    foreach $band (1 .. $num_bands) {
      push @{$bands{'lo'}}, shift @band_line;
    }

    while(){
      $line=<ECFS>;
SAS::error('ecfTooFewDataLines', "Too few data lines in $ecf_file") if (!defined($line));
    next if ($line =~ /^\s*$/);
    next if ($line =~ /^#/);
    last;
    }
    $line =~ s/^\s+//;
    chomp($line);
    @band_line = split('\s+', $line);
SAS::error('ecfWrongNumBandHis', "There should be $num_bands entries in the band-hi-E line in $ecf_file") if ($num_bands != @band_line);

    @{$bands{'hi'}} = ();
    foreach $band_index (1 .. $num_bands) {
      push @{$bands{'hi'}}, shift @band_line;
    }

    #....................................................................
    # Now look for ecfs:

    my %ecf_info = ();
    my $num_instruments = 0;
    while(defined($line=<ECFS>)){
      if ($line =~ /^instrument\s*=\s*(\w\w)/) {
        ++$num_instruments;
        $inst_uc = $1;
        foreach $filter ('OPEN', 'THIN', 'MEDIUM', 'THICK') {
          while(){
            $line=<ECFS>;
SAS::error('ecfNoFilterData', "There was no data for filter=$filter, instrument=$inst_uc in $ecf_file") if (!defined($line));
          next if ($line =~ /^\s*$/);
          next if ($line =~ /^#/);
          last;
          }

          $line =~ s/^\s+//;
          chomp($line);

          @ecf_line = split('\s+', $line);

SAS::error('ecfWrongNumEcfs', "There should be $num_bands entries in the line for filter=$filter of instrument=$inst_uc in $ecf_file") if ($num_bands != @ecf_line);

#          foreach $band (1 .. $num_bands) {
#          foreach $band (@{$bands{'id'}}) {
          foreach $band_index (0 .. @{$bands{'id'}} - 1) {
$band = $bands{'id'}[$band_index];
            $ecf_info{$inst_uc}{$filter}{$band} = $ecf_line[$band_index];
          }
        }
      }
    }

    close(ECFS);

SAS::error('ecfNoInstrumentData', "There was no data for instrument=$inst_uc in $ecf_file") if ($num_instruments==0);

  return (%ecf_info, %bands);
  }
