!
! Module: eimsimtest_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: This is a collection of routines used to test the making of simulated xmm images.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module eimsimtest_mod

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.

  use dal
  use cal	! contains 'use caloaldefs' and 'use caltypes'
  use param  
  use errorHandling

  use eimsim_utils
  use constants
  use dal_aux	! for addOrOpenColumn().
use save_image ! for saveOutImage()
  use ssc_misc
  use flux_transforms
  use cal_aux
  use array_utils ! for trapezoidSum
  use exposure ! for getQe and sampleEnergyRange
  use wcs_aux
  use coordinate ! for tanToInst

  implicit none 

  private 

  public eimsimtest

!  type(WcsT) :: wcs
  integer(int32) ::&
    instrumentId
  type(SpacecraftAttitudeType) ::&
    scAttitude, instrumAttitude
  real(double) ::&
    expStartTimeStamp

contains 

  !---------------------------------------------------------------------
  subroutine eimsimtest
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    select case(stringParameter('function'))
!    case ('fluxtest')
!      call testFluxConversions
    case ('lognlogs')
      call compileLogIntNlogS
    case ('template')
      call makeDummyLogNlogStemplate
    case ('fluxtrans')
      call testFluxTransforms
    case ('printecf')
      call testPrintEcf
    case ('qe')
      call testQe
    case ('thetamap')
      call makeThetaMap
!    case ('testnumv')
!      call testNumVintegral
    case ('anticorrsrclist')
!      call testAntiCorrSrcList
      call testMakeUncorrSrcListSet
!    case ('recalcrand')
!      call recalcFluxRandoms
case ('temp')
  call tempMakeRate0
    case default
call error('badStringPar','')
    end select

  end subroutine eimsimtest
  
  !---------------------------------------------------------------------
  subroutine testMakeUncorrSrcListSet!(srcListSetName, numSources, maskSetName&
!  , antiCorrLength, avoidEdges)
  !---------------------------------------------------------------------
  ! function=anticorrsrclist
  !
  ! Makes a fake eboxdetect output table of source RAs and decs. The source
  ! positions are evenly distributed in the plane of the mask image and are
  ! restricted to the true-valued mask pixels. In addition, the positions
  ! are anti-correlated and also optionally avoid the mask edges.

    implicit none

!    character(*),   intent(in) :: srcListSetName, maskSetName
!    integer(int32), intent(in) :: numSources
!    real(single),   intent(in) :: antiCorrLength
!    logical,        intent(in) :: avoidEdges

character(255) :: srcListSetName, maskSetName
integer(int32) :: numSources
real(single) :: antiCorrLength
logical :: avoidEdges
    integer(int32) :: bandId
    type(DataSetT) :: set
    type(TableT) :: tab
    real(single), pointer :: srcX(:), srcY(:), rate(:)
!    real(double) ::&
!      tanX(numSources), tanY(numSources), srcRa(numSources), srcDec(numSources)
real(double), allocatable ::&
  tanX(:), tanY(:), srcRa(:), srcDec(:)
    real(double), pointer :: srcRaDeg(:), srcDecDeg(:)
    logical(bool), pointer :: mask(:,:)
    type(WcsT) :: wcs

    !-------------------------------------------------------------------
    ! Start

    srcListSetName = stringParameter('simsrclistset')
    maskSetName    = stringParameter('expmapset')
    numSources     = intParameter('ints', 0)
    antiCorrLength = realParameter('reals', 0)
    avoidEdges     = booleanParameter('avoid')

    call getCalInfo(maskSetName, instrumentId, expStartTimeStamp&
      , scAttitude, instrumAttitude)

bandId = 2

    call makeEmptyEboxdetectSrclist(srcListSetName, numSources&
      , instrumentId, bandId)

    call readArrayData(maskSetName, mask) ! in fits_utils

    set = dataSet(srcListSetName, MODIFY)
    tab = table(set, 'SRCLIST')

    srcX      => real32Data(column(tab, 'X_IMA'))
    srcY      => real32Data(column(tab, 'Y_IMA'))
    rate      => real32Data(column(tab, 'RATE'))
rate = 1.0
    srcRaDeg  => real64Data(column(tab, 'RA'))
    srcDecDeg => real64Data(column(tab, 'DEC'))

    call makeUncorrPositions(mask, antiCorrLength, avoidEdges&
      , srcX, srcY)!, maskSetName)
    call readWcs(maskSetName, wcs)
allocate(tanX(numSources), tanY(numSources))
!allocate(srcRa(numSources), srcDec(numSources))
!    tanX = (srcX - wcs%x%refPixels) * abs(wcs%x%pixelSizeWorld) * PiOn180
!    tanY = (srcY - wcs%y%refPixels) * abs(wcs%y%pixelSizeWorld) * PiOn180
call pixelsToTan(srcX, srcY, wcs, tanX, tanY) ! in ssclib/wcs_aux
!    call tanToRaDec(wcs%x%refWorld, wcs%y%refWorld&
!    , tanX, tanY, srcRaDeg, srcDecDeg) ! in ssclib/coordinate.
call tanToRaDec(tanX, tanY, wcs%x%refWorld*PiOn180, wcs%y%refWorld*PiOn180&
 , srcRa, srcDec) ! in ssclib/coordinate.

srcRaDeg  = srcRa  / PiOn180
srcDecDeg = srcDec / PiOn180

    call release(set)

  end subroutine testMakeUncorrSrcListSet

  !---------------------------------------------------------------------
  subroutine makeEmptyEboxdetectSrclist(srcListSetName, numSources&
  , instrumentId, bandId)
  !---------------------------------------------------------------------

    implicit none

    character(*),   intent(in) :: srcListSetName
    integer(int32), intent(in) :: numSources, instrumentId, bandId

    type(DataSetT) ::&
      set
    type(TableT) ::&
      tab
    type(ColumnT) ::&
      col_BOX_ID_SRC,&
      col_ID_INST,&
      col_ID_BAND,&
      col_SCTS,&
      col_SCTS_ERR,&
      col_BOX_CTS,&
      col_X_IMA,&
      col_X_IMA_ERR,&
      col_Y_IMA,&
      col_Y_IMA_ERR,&
      col_LIKE,&
      col_BG_MAP,&
      col_BG_RAW,&
      col_EXP_MAP,&
      col_FLUX,&
      col_FLUX_ERR,&
      col_RATE,&
      col_RATE_ERR,&
      col_RA,&
      col_DEC,&
      col_RADEC_ERR,&
      col_LII,&
      col_BII,&
      col_HR1,&
      col_HR1_ERR,&
      col_HR2,&
      col_HR2_ERR,&
      col_HR3,&
      col_HR3_ERR,&
      col_VIGNETTING,&
      col_BOX_SIZE,&
      col_EEF,&
      col_DIST_NN
    integer(int32) ::&
      row
    integer(int32), pointer ::&
      data_BOX_ID_SRC(:), data_ID_INST(:), data_ID_BAND(:)
!, indices(:)
    real(single), pointer, dimension(:) ::&
      data_SCTS,&
      data_SCTS_ERR,&
      data_BOX_CTS,&
      data_X_IMA ,&
      data_X_IMA_ERR,&
      data_Y_IMA ,&
      data_Y_IMA_ERR ,&
      data_LIKE,&
      data_BG_MAP,&
      data_BG_RAW,&
      data_EXP_MAP,&
      data_FLUX ,&
      data_FLUX_ERR,&
      data_RATE,&
      data_RATE_ERR,&
      data_RADEC_ERR,&
      data_VIGNETTING,&
      data_BOX_SIZE,&
      data_EEF,&
      data_DIST_NN
    real(double), pointer ::&
      data_RA(:), data_DEC(:), data_LII(:), data_BII(:)
integer(int32) :: i
    !-------------------------------------------------------------------
    ! Start

    set = dataSet(srcListSetName, CREATE)
    tab = addTable(set, 'SRCLIST', numSources)

    col_BOX_ID_SRC = addColumn(tab, 'BOX_ID_SRC', INTEGER32)
    col_ID_INST    = addColumn(tab, 'ID_INST',    INTEGER32)
    col_ID_BAND    = addColumn(tab, 'ID_BAND',    INTEGER32)
    col_SCTS       = addColumn(tab, 'SCTS',       REAL32, 'counts')
    col_SCTS_ERR   = addColumn(tab, 'SCTS_ERR',   REAL32, 'counts')
    col_BOX_CTS    = addColumn(tab, 'BOX_CTS',    REAL32, 'counts')
    col_X_IMA      = addColumn(tab, 'X_IMA',      REAL32, 'image pixels')
    col_X_IMA_ERR  = addColumn(tab, 'X_IMA_ERR',  REAL32, 'image pixels')
    col_Y_IMA      = addColumn(tab, 'Y_IMA',      REAL32, 'image pixels')
    col_Y_IMA_ERR  = addColumn(tab, 'Y_IMA_ERR',  REAL32, 'image pixels')
    col_LIKE       = addColumn(tab, 'LIKE',       REAL32)
    col_BG_MAP     = addColumn(tab, 'BG_MAP',     REAL32, 'counts/pixel')
    col_BG_RAW     = addColumn(tab, 'BG_RAW',     REAL32, 'counts/pixel')
    col_EXP_MAP    = addColumn(tab, 'EXP_MAP',    REAL32, 'seconds')
    col_FLUX       = addColumn(tab, 'FLUX',       REAL32, 'cgs')
    col_FLUX_ERR   = addColumn(tab, 'FLUX_ERR',   REAL32, 'cgs')
    col_RATE       = addColumn(tab, 'RATE',       REAL32, 'counts/sec')
    col_RATE_ERR   = addColumn(tab, 'RATE_ERR',   REAL32, 'counts/sec')
    col_RA         = addColumn(tab, 'RA',         REAL64, 'deg')
    col_DEC        = addColumn(tab, 'DEC',        REAL64, 'deg')
    col_RADEC_ERR  = addColumn(tab, 'RADEC_ERR',  REAL32, 'arcsec')
    col_LII        = addColumn(tab, 'LII',        REAL64, 'deg')
    col_BII        = addColumn(tab, 'BII',        REAL64, 'deg')
    col_HR1        = addColumn(tab, 'HR1',        REAL32)
    col_HR1_ERR    = addColumn(tab, 'HR1_ERR',    REAL32)
    col_HR2        = addColumn(tab, 'HR2',        REAL32)
    col_HR2_ERR    = addColumn(tab, 'HR2_ERR',    REAL32)
    col_HR3        = addColumn(tab, 'HR3',        REAL32)
    col_HR3_ERR    = addColumn(tab, 'HR3_ERR',    REAL32)
    col_VIGNETTING = addColumn(tab, 'VIGNETTING', REAL32)
    col_BOX_SIZE   = addColumn(tab, 'BOX_SIZE',   REAL32, 'pixels')
    col_EEF        = addColumn(tab, 'EEF',        REAL32)
    col_DIST_NN    = addColumn(tab, 'DIST_NN',    REAL32, 'arcsec')

    data_BOX_ID_SRC => int32Data(col_BOX_ID_SRC)
!      indices => indicize(numSources)
!      data_BOX_ID_SRC = indices
data_BOX_ID_SRC = (/(i,i=1,numSources)/)
!      deallocate(indices)

    data_ID_INST    => int32Data(col_ID_INST)
      select case(instrumentId)
      case(EPN)
        data_ID_INST = 1
      case(EMOS1)
        data_ID_INST = 2
      case(EMOS2)
        data_ID_INST = 3
      case default
call error('badInstrument','')
      end select

    data_ID_BAND    => int32Data(col_ID_BAND)
      data_ID_BAND    = bandId

    data_SCTS       => real32Data(col_SCTS)
      data_SCTS       = 0.0

    data_SCTS_ERR   => real32Data(col_SCTS_ERR)
      data_SCTS_ERR   = 0.0

    data_BOX_CTS    => real32Data(col_BOX_CTS)
      data_BOX_CTS    = 0.0

    data_X_IMA      => real32Data(col_X_IMA)
      data_X_IMA      = 0.0

    data_X_IMA_ERR  => real32Data(col_X_IMA_ERR)
      data_X_IMA_ERR  = 0.0

    data_Y_IMA      => real32Data(col_Y_IMA)
      data_Y_IMA      = 0.0

    data_Y_IMA_ERR  => real32Data(col_Y_IMA_ERR)
      data_Y_IMA_ERR  = 0.0

    data_LIKE       => real32Data(col_LIKE)
      data_LIKE       = 0.7

    data_BG_MAP     => real32Data(col_BG_MAP)
      data_BG_MAP     = 0.0

    data_BG_RAW     => real32Data(col_BG_RAW)
      data_BG_RAW     = 0.0

    data_EXP_MAP    => real32Data(col_EXP_MAP)
      data_EXP_MAP    = 0.0

    data_FLUX       => real32Data(col_FLUX)
      data_FLUX       = 0.0

    data_FLUX_ERR   => real32Data(col_FLUX_ERR)
      data_FLUX_ERR   = 0.0

    data_RATE       => real32Data(col_RATE)
      data_RATE       = 0.0

    data_RATE_ERR   => real32Data(col_RATE_ERR)
      data_RATE_ERR   = 0.0

    data_RA         => real64Data(col_RA)
      data_RA         = 0.0d0

    data_DEC        => real64Data(col_DEC)
      data_DEC        = 0.0d0

    data_RADEC_ERR  => real32Data(col_RADEC_ERR)
      data_RADEC_ERR  = 0.0

    data_LII        => real64Data(col_LII)
      data_LII        = 0.0d0

    data_BII        => real64Data(col_BII)
      data_BII        = 0.0d0

    do row = 0, numSources - 1
      call setToNull(col_HR1,     row)
      call setToNull(col_HR1_ERR, row)
      call setToNull(col_HR2,     row)
      call setToNull(col_HR2_ERR, row)
      call setToNull(col_HR3,     row)
      call setToNull(col_HR3_ERR, row)
    end do

    data_VIGNETTING => real32Data(col_VIGNETTING)
      data_VIGNETTING = 0.0

    data_BOX_SIZE   => real32Data(col_BOX_SIZE)
      data_BOX_SIZE   = 0.0

    data_EEF        => real32Data(col_EEF)
      data_EEF        = 0.0

    data_DIST_NN    => real32Data(col_DIST_NN)
      data_DIST_NN    = 0.0


    call release(set)

  end subroutine makeEmptyEboxdetectSrclist

  !---------------------------------------------------------------------
  subroutine testFluxConversions
  !---------------------------------------------------------------------

    implicit none

    real(single) :: fluxErgCmSec, amplitude, newFlux
    real(double) :: photonSpectralIndex, EeVlo, EeVhi

    !-------------------------------------------------------------------
    ! Start

    photonSpectralIndex = -1.4d0
    EeVlo = 500.0d0
    EeVhi = 2000.0d0

    fluxErgCmSec = 1.0e-15

    amplitude = fluxToFluxDensity0(photonSpectralIndex, EeVlo, EeVhi&
      , fluxErgCmSec)

    print *, 'amplitude: ', amplitude

    newFlux = fluxDensity0ToFlux(photonSpectralIndex, EeVlo, EeVhi&
      , amplitude)

    print *, 'old flux, new flux: ', fluxErgCmSec, newFlux

  end subroutine testFluxConversions

  !---------------------------------------------------------------------
  subroutine makeDummyLogNlogStemplate
  !---------------------------------------------------------------------
  ! This makes a dummy (although a plausible) template containing information about the required logN-logS distribution of simulated sources. The output provides a convenient basis via editing with fv or similar for constructing a genuine logN-logS template.
  !
  ! The format for such files can be found in subroutine makeSimulatedSourceList of eimsim/src/eimsim_utils.f90.

    implicit none

    character(255) :: logNlogSSetName = 'dummy_lognlogs.ds'
    type(DataSetT) :: logNlogSset
    type(TableT) :: logNlogStab
    real(single), pointer :: setFluxErgCmSec(:), setCumDensityDeg(:)

    !-------------------------------------------------------------------
    ! Start

    logNlogSset = dataSet(logNlogSSetName, CREATE)
    logNlogStab = addTable(logNlogSset, 'LOGNLOGS', 1)

    call setAttribute(logNlogStab, 'EUNIT', 'eV')
    call setAttribute(logNlogStab, 'E_MIN', 500.0d0)
    call setAttribute(logNlogStab, 'E_MAX', 2000.0d0)
    call setAttribute(logNlogStab, 'SPECINDX', -1.4d0)
    call setAttribute(logNlogStab, 'LO_SLOPE', -0.8d0)
    call setAttribute(logNlogStab, 'HI_SLOPE', -1.5d0)
    call setAttribute(logNlogStab, 'MINFLUX', 5.0d-17)

    setFluxErgCmSec  => real32Data(addColumn(logNlogStab, 'FLUX',    REAL32))
    setCumDensityDeg => real32Data(addColumn(logNlogStab, 'DENSITY', REAL32))

    setFluxErgCmSec(0)  = 7e-15
    ! erg cm^-2 s^-1 (within the specified energy band).

    setCumDensityDeg(0) = 200.0
    ! sources per square degree.

    call release(logNlogSset)

  end subroutine makeDummyLogNlogStemplate

  !---------------------------------------------------------------------
  subroutine compileLogIntNlogS
  !---------------------------------------------------------------------
  ! This compiles a log(intN) vs logS histogram from a list of sources.
  !
  ! The subroutine requires two datasets: (i) A template table which should have a single column of flux values + 4 keywords E_MIN, E_MAX, EUNIT and SPECINDX. (ii) A set containing a table of source amplitudes.
  !
  ! Each source is assumed to have a flux density F that obeys the power law
  !
  !	F = A * E**gamma
  !
  ! where F is the flux density in counts s^-1 cm^-2 eV^-1, A is the amplitude stored in the output source list, E is the photon energy in eV and gamma is a spectral index SPECINDX. The flux S in ergs s^-1 cm^-2 can be recovered through the following relation:
  !
  !	              EeVhi
  !	               /
  !	S = 10**7 A qe | dE E E**gamma
  !	               /
  !	             EeVlo
  !
  !	     10**7 A qe
  !	  = ------------ (EeVhi**(gamma + 2) - EeVlo**(gamma + 2))	(1)
  !	     gamma + 2
  !
  ! where qe is the electronic charge in coulombs. This is done for all sources.
  !
  ! For each value of flux S in the template table, the number of sources brighter than that value is calculated. This number is divided by the mask area SKY_AREA (in square degrees) which is obtained from the source list, then written to a column DENSITY in the template table.
  !.....................................................................

    implicit none

    character(255) :: sourceListSetName
    type(DataSetT) :: sourceSet
    type(TableT) :: sourceTab
    integer(int32) :: numSources, i
    type(ColumnT) :: fluxCol, integDensityCol, integDensityUncertCol
    real(single), pointer :: flux(:), integDensity(:), integDensityUncert(:)
    real(single), allocatable :: amplitude(:)
    character(20) :: energyUnit
    real(double) :: photonSpectralIndex, EeVlo, EeVhi
    real(single) :: maskAreaDeg
    integer, pointer :: indices(:)

    !-------------------------------------------------------------------
    ! Start

    sourceListSetName = stringParameter('simsrclistset')
    sourceSet = dataSet(sourceListSetName, MODIFY)
    sourceTab = table(sourceSet, 'SRCLIST')
    numSources = numberOfRows(sourceTab)
if (numSources <= 0) call error('','')

    call sortTable(sourceTab, '-AMPLITUD') ! in ssclib/src/fits_utils.f90

    energyUnit = stringAttribute(sourceTab, 'EUNIT')
    select case(ucase(energyUnit))
    case('EV')
      EeVlo = real64Attribute(sourceTab, 'E_MIN')
      EeVhi = real64Attribute(sourceTab, 'E_MAX')
    case('KEV')
      EeVlo = real64Attribute(sourceTab, 'E_MIN') * 1000.0
      EeVhi = real64Attribute(sourceTab, 'E_MAX') * 1000.0
    case('ERG')
      EeVlo = real64Attribute(sourceTab, 'E_MIN') * 1.0e-7 / ElectronQ
      EeVhi = real64Attribute(sourceTab, 'E_MAX') * 1.0e-7 / ElectronQ
    case default
call error('','')
    end select

    photonSpectralIndex = real64Attribute(sourceTab, 'SPECINDX')
!    maskAreaDeg = real32Attribute(sourceTab, 'AREASQDG')
maskAreaDeg = real32Attribute(sourceTab, 'SKY_AREA')

    fluxCol = addColumn(sourceTab, 'FLUX', REAL32)
    flux => real32Data(fluxCol)

    allocate(amplitude(numSources))
    call readColDataToFixed(sourceTab, 'AMPLITUD', amplitude)

!print *, 'amp -> flux conversion:'
!print *, 'photonSpectralIndex, EeVlo, EeVhi: ', photonSpectralIndex, EeVlo, EeVhi
    do i = 1, numSources
      flux(i-1) = fluxDensity0ToFlux(photonSpectralIndex, EeVlo, EeVhi&
        , amplitude(i))
!print *, 'i, amplitude(i), -> flux(i-1): ', i, amplitude(i), '->',  flux(i-1)
    end do
!print *, ''
!print *, ''

    integDensityCol = addColumn(sourceTab, 'DENSITY', REAL32)
    integDensity => real32Data(integDensityCol)
    integDensityUncertCol = addColumn(sourceTab, 'DENSYERR', REAL32)
    integDensityUncert => real32Data(integDensityUncertCol)
allocate(indices(numSources))
!    indices => indicize(numSources)
indices = (/(i,i=1,numSources)/)
    integDensity = real(indices) / maskAreaDeg
    integDensityUncert = sqrt(real(indices)) / maskAreaDeg
    deallocate(indices)

    call release(sourceSet)

  end subroutine compileLogIntNlogS

  !---------------------------------------------------------------------
  subroutine convertFluxToRateDensity
  !---------------------------------------------------------------------

    implicit none

    real(single) :: flux, amplitude, rateDensity
    real(double) :: eventSpectrumPowerIndex, EeVlo, EeVhi, spotEeV
    character(20) :: filterName
    integer(int32) :: filtrId
    character(255) :: calInfoSetName

    !-------------------------------------------------------------------
    ! Start

    flux = realParameter('flux')
    eventSpectrumPowerIndex = realParameter('specindex')
    EeVlo = realParameter('elo')
    EeVhi = realParameter('ehi')
    spotEeV = realParameter('psfenergy')
    filterName = stringParameter('filter')
    calInfoSetName = stringParameter('calinfoset')

    call getCalInfo(calInfoSetName, instrumentId, expStartTimeStamp&
      , scAttitude, instrumAttitude)

    filtrId = filterId(filterName) ! in caloalutils/src/caloaldefs.f90
    call CAL_setState(filterId=filtrId)

    amplitude = fluxToFluxDensity0(eventSpectrumPowerIndex, EeVlo, EeVhi, flux)
    rateDensity = fluxDensity0ToRateDensity(eventSpectrumPowerIndex, spotEeV&
      , amplitude)!, filtrId)

    print *, 'rateDensity is', rateDensity

  end subroutine convertFluxToRateDensity

  !---------------------------------------------------------------------
  subroutine convertRateDensityToFlux
  !---------------------------------------------------------------------

    implicit none

    real(single) :: flux, amplitude, rateDensity
    real(double) :: eventSpectrumPowerIndex, EeVlo, EeVhi, spotEeV
    character(20) :: filterName
    integer(int32) :: filtrId
    character(255) :: calInfoSetName

    !-------------------------------------------------------------------
    ! Start

    rateDensity = realParameter('psfmaxcoarserate')
    eventSpectrumPowerIndex = realParameter('specindex')
    EeVlo = realParameter('elo')
    EeVhi = realParameter('ehi')
    spotEeV = realParameter('psfenergy')
    filterName = stringParameter('filter')
    calInfoSetName = stringParameter('calinfoset')

    call getCalInfo(calInfoSetName, instrumentId, expStartTimeStamp&
      , scAttitude, instrumAttitude)

    filtrId = filterId(filterName) ! in caloalutils/src/caloaldefs.f90
    call CAL_setState(filterId=filtrId)

    amplitude = rateDensityToFluxDensity0(eventSpectrumPowerIndex, spotEeV&
      , rateDensity)!, filtrId)
    flux = fluxDensity0ToFlux(eventSpectrumPowerIndex, EeVlo, EeVhi&
      , amplitude)

    print *, 'flux is', flux

  end subroutine convertRateDensityToFlux

  !---------------------------------------------------------------------
  subroutine convertRateToRateDensity
  !---------------------------------------------------------------------

    implicit none

    real(single) :: rate, amplitude, rateDensity
    real(double) :: eventSpectrumPowerIndex, EeVlo, EeVhi, spotEeV
    character(20) :: filterName
    integer(int32) :: filtrId
    character(255) :: calInfoSetName

    !-------------------------------------------------------------------
    ! Start

    rate = realParameter('rate')
    eventSpectrumPowerIndex = realParameter('specindex')
    EeVlo = realParameter('elo')
    EeVhi = realParameter('ehi')
    spotEeV = realParameter('psfenergy')
    filterName = stringParameter('filter')
    calInfoSetName = stringParameter('calinfoset')

    call getCalInfo(calInfoSetName, instrumentId, expStartTimeStamp&
      , scAttitude, instrumAttitude)

    filtrId = filterId(filterName) ! in caloalutils/src/caloaldefs.f90
    call CAL_setState(filterId=filtrId)

    amplitude = rateToFluxDensity0(eventSpectrumPowerIndex, EeVlo, EeVhi, rate)
!&      , filtrId)
    rateDensity = fluxDensity0ToRateDensity(eventSpectrumPowerIndex, spotEeV&
      , amplitude)!, filtrId)

    print *, 'rateDensity is', rateDensity

  end subroutine convertRateToRateDensity

  !---------------------------------------------------------------------
  subroutine convertFluxToRate
  !---------------------------------------------------------------------

    implicit none

    real(single) :: flux, amplitude, rate
    real(double) :: eventSpectrumPowerIndex, EeVlo, EeVhi
    character(20) :: filterName
    integer(int32) :: filtrId
    character(255) :: calInfoSetName

    !-------------------------------------------------------------------
    ! Start

    flux = realParameter('flux')
    eventSpectrumPowerIndex = realParameter('specindex')
    EeVlo = realParameter('elo')
    EeVhi = realParameter('ehi')
    filterName = stringParameter('filter')
    calInfoSetName = stringParameter('calinfoset')

    call getCalInfo(calInfoSetName, instrumentId, expStartTimeStamp&
      , scAttitude, instrumAttitude)

    filtrId = filterId(filterName) ! in caloalutils/src/caloaldefs.f90
    call CAL_setState(filterId=filtrId)

    amplitude = fluxToFluxDensity0(eventSpectrumPowerIndex, EeVlo, EeVhi, flux)
    rate = fluxDensity0ToRate(eventSpectrumPowerIndex, EeVlo, EeVhi&
      , amplitude)!, filtrId)

    print *, 'rate is', rate

  end subroutine convertFluxToRate

  !---------------------------------------------------------------------
  subroutine testFluxTransforms
  !---------------------------------------------------------------------
  ! when function=fluxtrans.

    implicit none

    real(single) ::&
      flux, fluxDensity, amplitude, rate, rateDensity
    real(double) :: eventSpectrumPowerIndex, EeVlo, EeVhi, EeV
    character(20) :: filterName, style
    integer(int32) :: filtrId
    character(20) :: instrumentName

    !-------------------------------------------------------------------
    ! Start

    style = stringParameter('style')
    select case(style)
    !...................................................................
    case ('fluxtoamp')
      eventSpectrumPowerIndex = realParameter('specindex')
      EeVlo = realParameter('elo')
      EeVhi = realParameter('ehi')
      flux  = realParameter('flux')

      amplitude = fluxToFluxDensity0(eventSpectrumPowerIndex, EeVlo, EeVhi&
        , flux)

      print *, 'result: amplitude =', amplitude

    !...................................................................
    case ('fluxdensitytoamp')
      eventSpectrumPowerIndex = realParameter('specindex')
      EeV         = realParameter('psfenergy')
      fluxDensity = realParameter('fluxdensity')

      amplitude = fluxDensityToFluxDensity0(eventSpectrumPowerIndex, EeV&
        , fluxDensity)

      print *, 'result: amplitude =', amplitude

    !...................................................................
    case ('amptoflux')
      eventSpectrumPowerIndex = realParameter('specindex')
      EeVlo     = realParameter('elo')
      EeVhi     = realParameter('ehi')
      amplitude = realParameter('amplitude')

      flux = fluxDensity0ToFlux(eventSpectrumPowerIndex, EeVlo, EeVhi&
        , amplitude)

      print *, 'result: flux =', flux, 'erg cm^-2 s^-1'

    !...................................................................
    case ('amptofluxdensity')
      eventSpectrumPowerIndex = realParameter('specindex')
      EeV       = realParameter('psfenergy')
      amplitude = realParameter('amplitude')

      fluxDensity = fluxDensity0ToFluxDensity(eventSpectrumPowerIndex, EeV&
        , amplitude)

      print *, 'result: fluxDensity =', fluxDensity, 'erg cm^-2 s^-1 eV^-1'

    !...................................................................
    case ('ratetoamp')
      eventSpectrumPowerIndex = realParameter('specindex')
      EeVlo          =   realParameter('elo')
      EeVhi          =   realParameter('ehi')
      filterName     = stringParameter('filter')
      instrumentName = stringParameter('inst')
      rate           =   realParameter('rate')

      instrumentId = instId(instrumentName) ! in caloalutils/src/caloaldefs.f90
      call CAL_setState(instrument=instrumentId)
      filtrId = filterId(filterName) ! in caloalutils/src/caloaldefs.f90
      call CAL_setState(filterId=filtrId)

      amplitude = rateToFluxDensity0(eventSpectrumPowerIndex&
        , EeVlo, EeVhi, rate)!, filtrId)

      print *, 'result: amplitude =', amplitude

    !...................................................................
    case ('ratedensitytoamp')
      eventSpectrumPowerIndex = realParameter('specindex')
      EeV            =   realParameter('psfenergy')
      filterName     = stringParameter('filter')
      instrumentName = stringParameter('inst')
      rateDensity    =   realParameter('ratedensity')

      instrumentId = instId(instrumentName) ! in caloalutils/src/caloaldefs.f90
      call CAL_setState(instrument=instrumentId)
      filtrId = filterId(filterName) ! in caloalutils/src/caloaldefs.f90
      call CAL_setState(filterId=filtrId)

      amplitude = rateDensityToFluxDensity0(eventSpectrumPowerIndex, EeV&
        , rateDensity)!, filtrId)

      print *, 'result: amplitude =', amplitude

    !...................................................................
    case ('amptorate')
      eventSpectrumPowerIndex = realParameter('specindex')
      EeVlo          =   realParameter('elo')
      EeVhi          =   realParameter('ehi')
      filterName     = stringParameter('filter')
      instrumentName = stringParameter('inst')
      amplitude      =   realParameter('amplitude')

      instrumentId = instId(instrumentName) ! in caloalutils/src/caloaldefs.f90
      call CAL_setState(instrument=instrumentId)
      filtrId = filterId(filterName) ! in caloalutils/src/caloaldefs.f90
      call CAL_setState(filterId=filtrId)

      rate = fluxDensity0ToRate(eventSpectrumPowerIndex, EeVlo, EeVhi&
        , amplitude)!, filtrId)

      print *, 'result: detected rate =', rate, 's^-1'

    !...................................................................
    case ('amptoratedensity')
      eventSpectrumPowerIndex = realParameter('specindex')
      EeV            =   realParameter('psfenergy')
      filterName     = stringParameter('filter')
      instrumentName = stringParameter('inst')
      amplitude      =   realParameter('amplitude')

      instrumentId = instId(instrumentName) ! in caloalutils/src/caloaldefs.f90
      call CAL_setState(instrument=instrumentId)
      filtrId = filterId(filterName) ! in caloalutils/src/caloaldefs.f90
      call CAL_setState(filterId=filtrId)

      rateDensity = fluxDensity0ToRateDensity(eventSpectrumPowerIndex, EeV&
        , amplitude)!, filtrId)

      print *, 'result: detected rateDensity =', rateDensity, 's^-1 eV^-1'

    case default
call error('badFluxTransStyle','')
    end select

  end subroutine testFluxTransforms

  !---------------------------------------------------------------------
  subroutine testPrintEcf
  !---------------------------------------------------------------------

    implicit none

    real(double) :: eventSpectrumPowerIndex, EeVlo, EeVhi, ecf
    character(20) :: filterName, instrumentName

    !-------------------------------------------------------------------
    ! Start

    eventSpectrumPowerIndex = realParameter('specindex')
    instrumentName = stringParameter('inst')
    filterName     = stringParameter('filter')
    EeVlo          =   realParameter('elo')
    EeVhi          =   realParameter('ehi')

    ecf = calcEcf(eventSpectrumPowerIndex, EeVlo, EeVhi&
      , instrumentName, filterName)

    print *, 'result: ecf =', ecf, 'counts s^-1 per 10^-11 cgs'

  end subroutine testPrintEcf

  !---------------------------------------------------------------------
  subroutine testQe
  !---------------------------------------------------------------------
  ! NOTE! The cal MUST be set up for the correct instrument before this
  ! function is called. The cal state is set up in here for CCDNR==1.
  !
  ! This routine calculates the spectrum-weighted QE for pattern 0, at
  ! pixel (50,100) of chip 1.
  !.....................................................................

    implicit none

    integer(int16), parameter :: numEsamples = 100
    real(double) ::&
      eventSpectrumPowerIndex, EeVlo, EeVhi, EeVsamples(numEsamples)
    real(single), dimension(numEsamples) ::&
      qe, spectrum
    integer(int16) :: Ei
    real(single) :: multiplier
    logical(bool), allocatable :: usePattern(:)
    integer(int8), pointer ::&
      patterns(:,:,:)
    integer(int32) ::&
      numPatterns

    !-------------------------------------------------------------------
    ! Start

    eventSpectrumPowerIndex = realParameter('specindex')
    EeVlo = realParameter('elo')
    EeVhi = realParameter('ehi')

    call sampleEnergyRange(EeVlo, EeVhi, EeVsamples) ! in ssclib/exposure
    call CAL_setState(ccdChipId=1)
    call CAL_getEventPatterns(patterns)
    numPatterns = size(patterns, 1)
    allocate(usePattern(0:numPatterns-1))
    usePattern = .false. ! default
    usePattern(0) = .true.
    call getQe(EeVsamples, usePattern, qe) ! in ssclib/exposure

    do Ei = 1, numEsamples
      spectrum(Ei) = EeVsamples(Ei)**eventSpectrumPowerIndex
    end do

    multiplier = trapezoidSum(real(EeVsamples, single), spectrum * qe)
    ! in ssclib/array_utils

    print *, multiplier / trapezoidSum(real(EeVsamples, single), spectrum)

    call CAL_releaseMemory(patterns)

  end subroutine testQe

  !---------------------------------------------------------------------
  subroutine makeThetaMap
  !---------------------------------------------------------------------

    implicit none

    character(255) :: expMapSetName
    real(double), pointer :: expMap(:,:)
    integer(int16) :: xsize, ysize, xi, yi
    type(WcsT) :: imageWcs
    real(double) :: tanX, tanY
!, refRa, refDec
    integer(int32) ::&
      instrumentId
    real(double) ::&
      expStartTimeStamp, phi
    type(SpacecraftAttitudeType) ::&
      scAttitude, instrumAttitude
!    integer :: foreignStatus

    !-------------------------------------------------------------------
    ! Start

    expMapSetName = stringParameter('expmapset')
    call readArrayData(expMapSetName, expMap)
    xsize = size(expMap, 1)
    ysize = size(expMap, 2)

    call getCalInfo(expMapSetName, instrumentId, expStartTimeStamp&
      , scAttitude, instrumAttitude)

    call readWcs(expMapSetName, imageWcs)
!    refRa  = imageWcs%x%refWorld * PiOn180
!    refDec = imageWcs%y%refWorld * PiOn180

    do xi = 1, xsize
      do yi = 1, ysize
      if (expMap(xi, yi) <= 0.0) cycle

!        tanX = (real(xi) - imageWcs%x%refPixels)&
!          * abs(imageWcs%x%pixelSizeWorld) * PiOn180
!        tanY = (real(yi) - imageWcs%y%refPixels)&
!          * abs(imageWcs%y%pixelSizeWorld) * PiOn180
call pixelsToTan(real(xi, single), real(yi, single), imageWcs, tanX, tanY)
! in ssclib/wcs_aux

!        call tanToInst(tanX, tanY, imageWcs%x%refWorld, imageWcs%y%refWorld&
!          , scAttitude&
!          , expStartTimeStamp, thetaArcsec=expMap(xi, yi), phi=phi)
call tanToTel(tanX, tanY, imageWcs%x%refWorld*PiOn180&
 , imageWcs%y%refWorld*PiOn180&
 , scAttitude, expStartTimeStamp, expMap(xi, yi), phi)

      end do
    end do

    call saveImage(real(expMap, single), expMapSetName, 'theta_map.ds', 'temp.ds')

    deallocate(expMap)

  end subroutine makeThetaMap

!  !---------------------------------------------------------------------
!  subroutine testNumVintegral
!  !---------------------------------------------------------------------
!
!    implicit none
!
!    integer(int16) :: numSegments, numLsamples, i, segmentNum
!    real(double), allocatable :: simLsamples(:), simLpowers(:), simLamps(:),&
!      L(:), aV(:), q(:)
!    real(double) :: srcL, deltaX, deltaY, deltaL, Ldiff
!    character(12) :: style
!    real(single) :: numV
!    integer(int16), pointer :: indices(:)
!
!    !-------------------------------------------------------------------
!    ! Start
!
!    numSegments = 4
!    allocate(simLsamples(numSegments), simLpowers(numSegments), simLamps(numSegments))
!
!    srcL = 3.17d0
!    deltaX = 0.701d0
!    deltaY = 0.592d0
!    deltaL = 1.36d0
!
!    simLsamples = (/0.84d0, 1.05d0, 4.73d0, 5.19d0/)
!    simLpowers  = (/-1.2d0, -1.7d0, -0.6d0, -1.5d0/)
!    simLamps    = (/390.0d0, 258.0d0, 317.0d0, 594.0d0/)
!
!style = 'ellipse' ! can't really do numerical test of half-ellipse without some fiddling with the linearity of the bins.
!
!    numV = calcNumSrcsInVolume(srcL, deltaX, deltaY, deltaL, style&
!      , simLsamples, simLpowers, simLamps)
!
!    print *, 'analytical numV = :', numV
!
!    ! Now do numerical approx:
!
!    numLsamples = 10000
!    allocate(L(numLsamples), q(numLsamples), aV(numLsamples))
!    indices => indicize(numLsamples)
!    L = fracVal(srcL - deltaL, srcL + deltaL, (indices - 1)&
!      / real(numLsamples - 1, double))
!
!    deallocate(indices)
!
!    do i = 1, numLsamples
!      Ldiff = L(i) - srcL
!      aV(i) = Pi * deltaX * deltaY * (1.0d0 - Ldiff * Ldiff / deltaL / deltaL)
!      segmentNum = getBinNum(real(simLsamples), real(L(i)))
!      if (segmentNum <= 0) then
!        q(i) = 0.0d0
!      else
!        q(i) = simLamps(segmentNum) * L(i)**simLpowers(segmentNum)
!      end if
!!print *, 'i, L(i), aV(i), q(i): ', i, L(i), aV(i), q(i)
!    end do
!
!    numV = trapezoidSum(real(L), real(aV * q))
!
!    print *, ' numerical numV = :', numV
!
!!*** might be of interest to at least look at halfellipse value:
!
!!numV = calcNumSrcsInVolume(srcL, deltaX, deltaY, deltaL, 'halfellipse'&
!!  , simLsamples, simLpowers, simLamps)
!!
!!print *, 'halfellipse numV = :', numV
!
!  end subroutine testNumVintegral

  !---------------------------------------------------------------------
!  subroutine testAntiCorrSrcList
!  !---------------------------------------------------------------------
!  ! function = 'anticorrsrclist'
!
!    implicit none
!
!    character(255) :: srcListSetName, maskSetName
!    integer(int32) :: numSources
!    real(single) :: antiCorrLength
!    logical :: avoidEdges
!
!    !-------------------------------------------------------------------
!    ! Start
!
!    srcListSetName = stringParameter('simsrclistset')
!    maskSetName    = stringParameter('expmapset')
!    numSources     = intParameter('ints', 0)
!    antiCorrLength = realParameter('reals', 0)
!    avoidEdges     = booleanParameter('avoid')
!
!    call makeUncorrSrcListSet(srcListSetName, numSources, maskSetName&
!      , antiCorrLength, avoidEdges)
!
!  end subroutine testAntiCorrSrcList

  !---------------------------------------------------------------------
  subroutine tempMakeRate0
  !---------------------------------------------------------------------

    implicit none

    character(255) :: calInfoSetName, sourceListSetName
    character(20) :: filterName
    integer(int32) :: filtrId
    real(double) :: photonSpectralIndex, EeVlo, EeVhi
    real(single), pointer :: simSrcL(:), amplitude(:)
    type(DataSetT) :: sourceSet
    type(TableT) :: sourceTab

    !-------------------------------------------------------------------
    ! Start

    calInfoSetName = stringParameter('calinfoset')
    sourceListSetName = stringParameter('simsrclistset')

    call getCalInfo(calInfoSetName, instrumentId, expStartTimeStamp&
      , scAttitude, instrumAttitude)

    filterName     = 'Thin1'
    filtrId = filterId(filterName) ! in caloalutils/src/caloaldefs.f90
    call CAL_setState(filterId=filtrId)
    EeVlo = 200.0
    EeVhi = 12000.0

    sourceSet = dataSet(sourceListSetName, MODIFY)
    sourceTab = table(sourceSet, 'SRCLIST', MODIFY)
    photonSpectralIndex = real64Attribute(sourceTab, 'SPECINDX')
!    simSrcL => addOrOpenColumnDataReal32(sourceTab, 'RATE_M2S003_8'&
!      , 'counts/sec')
simSrcL => real32Data(addOrOpenColumn(sourceTab, 'RATE_M2S003_8', REAL32&
      , 'counts/sec'))

    amplitude => real32Data(column(sourceTab, 'AMPLITUD'))
    simSrcL  = fluxDensity0ToRate(photonSpectralIndex, EeVlo, EeVhi&
      , amplitude)!, filtrId)

    if (hasAttribute(sourceTab, 'AREASQDG'))&
      call rename(attribute(sourceTab, 'AREASQDG'), 'SKY_AREA')

    call release(sourceSet)

  end subroutine tempMakeRate0

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module eimsimtest_mod
