!
! Module: esrclistpatch_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Adds columns SRC_BOX_CTS, SRC_BOX_CTS_ERR and BKG_BOX_CTS to the emldetect output (taking the eboxdetect output as input) and also copies the LIKE column from eboxdetect.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module esrclistpatch_mod

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.

  use dal
  use param  
  use errorHandling

  use array_utils	! smallestTrueIndex().
  use dal_aux		! addOrOpenColumn().

  implicit none 

  private 

  character(255) ::&
    boxSrcListSetName,&
    mlSrcListSetName

  public esrclistpatch

contains 

  !---------------------------------------------------------------------
  subroutine esrclistpatch
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT) :: boxSet, mlSet
    type(TableT) :: boxTab, mlTab
    type(ColumnT) ::&
      mlLikeCol, mlBoxSizeCol, mlBkgBoxCtsCol, mlSrcBoxCtsCol, mlSrcBoxCtsErrCol
    integer(int32), pointer ::&
      boxIndex(:), boxInstrum(:), boxIdband(:), mlBoxIndex(:),&
      mlInstrum(:), mlIdband(:)
    real(single), pointer, dimension(:) ::&
      boxBoxCts, boxBgMap, boxBoxSize, boxLike, mlLike, mlBkgBoxCts,&
      mlSrcBoxCts, mlSrcBoxCtsErr, mlBoxSize
    integer :: numBoxSrcs, numMlSrcs, mlRow, boxRow
    logical(bool), allocatable :: mask(:)

    !-------------------------------------------------------------------
    ! Start

    call readParameters

    boxSet = dataSet(boxSrcListSetName, READ)
    boxTab = table(boxSet, 'SRCLIST')
    numBoxSrcs = numberOfRows(boxTab)
    allocate(mask(numBoxSrcs))

    boxIndex   => int32Data(column(boxTab, 'BOX_ID_SRC'))
    boxInstrum => int32Data(column(boxTab, 'ID_INST'))
    boxIdband  => int32Data(column(boxTab, 'ID_BAND'))
    boxBoxCts  => real32Data(column(boxTab, 'BOX_CTS'))
    boxBgMap   => real32Data(column(boxTab, 'BG_MAP'))
    boxBoxSize => real32Data(column(boxTab, 'BOX_SIZE'))
    boxLike    => real32Data(column(boxTab, 'LIKE'))

    mlSet = dataSet(mlSrcListSetName, MODIFY)
    mlTab = table(mlSet, 'SRCLIST')
    numMlSrcs = numberOfRows(mlTab)
!    allocate(mask(numMlSrcs))
    mlBoxIndex => int32Data(column(mlTab, 'BOX_ID_SRC'))
    mlInstrum  => int32Data(column(mlTab, 'ID_INST'))
    mlIdband   => int32Data(column(mlTab, 'ID_BAND'))

    mlLikeCol         = addOrOpenColumn(mlTab, 'LIKE'&
      , REAL32, '',       'eboxdetect detection likelihood.')
    mlBoxSizeCol      = addOrOpenColumn(mlTab, 'BOX_SIZE'&
      , REAL32, 'pixels', 'eboxdetect square box dimension.')
    mlBkgBoxCtsCol    = addOrOpenColumn(mlTab, 'BKG_BOX_CTS'&
      , REAL32, 'counts', 'BG_MAP * BOX_SIZE * BOX_SIZE')
    mlSrcBoxCtsCol    = addOrOpenColumn(mlTab, 'SRC_BOX_CTS'&
      , REAL32, 'counts', 'BOX_CTS - BKG_BOX_CTS')
    mlSrcBoxCtsErrCol = addOrOpenColumn(mlTab, 'SRC_BOX_CTS_ERR'&
      , REAL32, 'counts', 'sqrt(BOX_CTS)')
! - BKG_BOX_CTS')
    ! addOrOpenColumn is in ssclib/dal_aux.

    do mlRow = 0, numMlSrcs - 1
      call setToNull(mlLikeCol,         mlRow)
      call setToNull(mlBkgBoxCtsCol,    mlRow)
      call setToNull(mlSrcBoxCtsCol,    mlRow)
      call setToNull(mlSrcBoxCtsErrCol, mlRow)
    end do

    mlLike         => real32Data(mlLikeCol)
    mlBoxSize      => real32Data(mlBoxSizeCol)
    mlBkgBoxCts    => real32Data(mlBkgBoxCtsCol)
    mlSrcBoxCts    => real32Data(mlSrcBoxCtsCol)
    mlSrcBoxCtsErr => real32Data(mlSrcBoxCtsErrCol)

    do mlRow = 0, numMlSrcs - 1
    if (mlIdband(mlRow) == 9) cycle ! don't want the xid band.

      mask = (boxIndex   == mlBoxIndex(mlRow)&
        .and. boxInstrum == mlInstrum(mlRow)&
        .and. boxIdband  == mlIdband(mlRow))
if (count(mask) .ne. 1) call error('boxIndexNotFound','')

      boxRow = smallestTrueIndex(mask) - 1

      mlLike(mlRow) = boxLike(boxRow)
      mlBoxSize(mlRow) = boxBoxSize(boxRow)
      mlBkgBoxCts(mlRow) = boxBgMap(boxRow) * boxBoxSize(boxRow)&
        * boxBoxSize(boxRow)
      mlSrcBoxCts(mlRow) = boxBoxCts(boxRow) - mlBkgBoxCts(mlRow)
    if (mlSrcBoxCts(mlRow) < 0.0) cycle ! leave the err at null.
      mlSrcBoxCtsErr(mlRow) = sqrt(boxBoxCts(boxRow))
! - mlBkgBoxCts(mlRow))
    end do

    call release(mlSet)
    call release(boxSet)

  end subroutine esrclistpatch
  
  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

    boxSrcListSetName = stringParameter('boxset')
    mlSrcListSetName = stringParameter('mlset')

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module esrclistpatch_mod
