!+
! Module: histogram_2
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains some assorted routines.
! 
! Comments: TODO:
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module histogram_2

!  use types
  use dal
!  use param  
!  use errorHandling

!  use constants
!  use sort_mod
  use ssc_misc	! for fracVal().

  implicit none 

  private 

  public makeHistogram

  type, public :: histoType
    real(double) ::&
      xLo,&
      xHi,&
      yAve,&
      yStdDev,&
      yUncert
    integer(int32) ::&
      n
  end type histoType

  contains

  !---------------------------------------------------------------------
  subroutine makeHistogram(x, histo, y, binEdges, binSize)
  !---------------------------------------------------------------------
  ! Presence of binEdges takes precedence over binSize.

    implicit none

    real(single), intent(in)           :: x(:)
    type(histoType), pointer         :: histo(:)
    real(single), intent(in), optional :: y(size(x))
    real(single), intent(in), optional :: binEdges(:),&
                                          binSize

    logical :: mask(size(x))
    integer(int32) :: numBins, bin
    real(single) :: overShoot, weight(size(x))
!integer :: i, summN
!real :: summW, aveg

    !-------------------------------------------------------------------
    ! Start

    if (size(x) <= 0) then
      allocate(histo(0))
print *, 'size(x) <= 0'
  return
    end if
    if (.not. (present(binEdges) .or. present(binSize))) then
      allocate(histo(0))
print *, '.not. (present(binEdges) .or. present(binSize))'
  return
    end if

    if (present(y)) then
      weight = y
    else
      weight = 1.0
    end if

    ! Construct bin edges:

    if (present(binEdges)) then
      numBins = size(binEdges) - 1

      if (numBins <= 0) then
        allocate(histo(0))
print *, 'present(binEdges), but numBins <= 0'
  return
      end if
      if (any(binEdges(:numBins) >= binEdges(2:))) then ! bin edges are
      ! either not all distinct values, or are not in ascending order.
        allocate(histo(0))
print *, 'present(binEdges), but bad order.'
do bin = 1, numBins
  print *, 'bin, binEdges(bin), binEdges(bin+1): ', bin, binEdges(bin), binEdges(bin+1)
  if (binEdges(bin) >= binEdges(bin+1)) print *, 'WHOA BABY!!'
end do
  return
      end if

      allocate(histo(numBins))
      histo%xLo = binEdges(:numBins)
      histo%xHi = binEdges(2:)

    else ! then binSize must be present from the test applied earlier.
      if (binSize <= 0) then
        allocate(histo(0))
print *, 'present(binSize), but binSize <= 0'
  return
      end if

      numBins = max(1, ceiling((maxval(x) - minval(x)) / binSize)) ! there is
      ! at least 1 bin, even if all x values are the same.

      allocate(histo(numBins))
      overShoot = ((numBins * binSize) - (maxval(x) - minval(x))) / 2.0

!      bin = 1
!      histo(bin)%xLo = minval(x) - overShoot
!      do bin = 2, numBins
!        histo(bin-1)%xHi = histo(1)%xLo + (bin - 1) * binSize
!        histo(bin  )%xLo = histo(bin-1)%xHi
!      end do
!      histo(numBins)%xHi = histo(1)%xLo + numBins * binSize

do bin = 1, numBins
  histo(bin)%xLo = fracval(minval(x) - overShoot, maxval(x) + overShoot&
  , (bin - 1) / real(numBins))
end do
histo(1:numBins-1)%xHi = histo(2:numBins)%xLo
histo(numBins)%xHi = maxval(x) + overShoot
    end if ! present(binEdges)

    ! Now make the histogram (all but the last bin):

!print *, ''
!bin = 31
!summN = 0
!summW = 0.0
!do i = 1, size(x)
!  if (x(i) >= histo(bin)%xLo .and. x(i) < histo(bin)%xHi) then
!    print *, 'i, x(i), weight(i): ', i, x(i), weight(i)
!    summN = summN + 1
!    summW = summW + weight(i)
!  end if
!end do
!aveg = summW/real(summN)
!print *, 'summN, summW, ave: ', summN, summW, aveg
!summW = 0.0
!do i = 1, size(x)
!  if (x(i) >= histo(bin)%xLo .and. x(i) < histo(bin)%xHi) then
!!    print *, 'i, err(i): ', i, weight(i) - aveg
!    summW = summW + (weight(i) - aveg)*(weight(i) - aveg)
!  end if
!end do
!print *, 'summW, aveErr: ', summW, sqrt(summW/real(summN))

    do bin = 1, numBins - 1
      mask = (x >= histo(bin)%xLo .and. x < histo(bin)%xHi)
      histo(bin)%n = count(mask)
      histo(bin)%yAve = sum(weight, mask)

!print *, 'bin, x, n, total_weight, ave_weight: '&
!, (histo(bin)%xLo+histo(bin)%xHi)/2.0, bin, histo(bin)%n, histo(bin)%yAve&
!, histo(bin)%yAve / real(histo(bin)%n)
      if (histo(bin)%n > 0) then
        histo(bin)%yAve = histo(bin)%yAve / real(histo(bin)%n)
      else
        histo(bin)%yAve = 0.0
      end if

!      histo(bin)%yErr = sum((weight - histo(bin)%yAve)&
histo(bin)%yStdDev = sum((weight - histo(bin)%yAve)&
        * (weight - histo(bin)%yAve), mask)

      if (histo(bin)%n > 0) then
!        histo(bin)%yErr = sqrt(histo(bin)%yErr / real(histo(bin)%n))
histo(bin)%yStdDev = sqrt(histo(bin)%yStdDev / real(histo(bin)%n))
histo(bin)%yUncert = histo(bin)%yStdDev / sqrt(real(histo(bin)%n))
      else
!        histo(bin)%yErr = 0.0
histo(bin)%yStdDev = 0.0
histo(bin)%yUncert = 0.0
      end if
    end do

    ! Sort out the top bin (it has a different mask condition):

    bin = numBins
    mask = (x >= histo(bin)%xLo .and. x <= histo(bin)%xHi)
    histo(bin)%n = count(mask)
    histo(bin)%yAve = sum(weight, mask)

    if (histo(bin)%n > 0) then
      histo(bin)%yAve = histo(bin)%yAve / real(histo(bin)%n)
    else
      histo(bin)%yAve = 0.0
    end if

!    histo(bin)%yErr = sum((weight - histo(bin)%yAve)&
histo(bin)%yStdDev = sum((weight - histo(bin)%yAve)&
      * (weight - histo(bin)%yAve), mask)

    if (histo(bin)%n > 0) then
!      histo(bin)%yErr = sqrt(histo(bin)%yErr / real(histo(bin)%n))
histo(bin)%yStdDev = sqrt(histo(bin)%yStdDev / real(histo(bin)%n))
histo(bin)%yUncert = histo(bin)%yStdDev / sqrt(real(histo(bin)%n))
    else
!      histo(bin)%yErr = 0.0
histo(bin)%yStdDev = 0.0
histo(bin)%yUncert = 0.0
    end if

  end subroutine makeHistogram

  !---------------------------------------------------------------------
  subroutine saveHistogram(histo, histoSet, histoTabName)
  !---------------------------------------------------------------------
!*** NOTE!! the dataset must have been opened for MODIFY or CREATE.

    implicit none

    type(histoType), intent(in) :: histo(:)
    type(DataSetT), intent(in) :: histoSet
    character(*), intent(in) :: histoTabName

    integer(int32) :: numRows
    type(TableT) :: tab
    type(ColumnT) :: xLoCol, xHiCol, nCol, yAveCol&
, yStdDevCol, yUncertCol
!, yErrCol
    real(single), pointer :: xLo(:), xHi(:), yAve(:)&
, yStdDev(:), yUncert(:)
!, yErr(:)
    integer(int32), pointer :: n(:)

    !-------------------------------------------------------------------
    ! Start

print *, trim(histoTabName)

    numRows = size(histo)
  if (numRows <= 0) return
    tab = addTable(histoSet, trim(histoTabName), numRows)
    xLoCol     = addColumn(tab, 'X_LOW',      REAL32&
      , comment='The low edge of the bin.')
    xHiCol     = addColumn(tab, 'X_HIGH',     REAL32&
      , comment='The high edge of the bin.')
    nCol       = addColumn(tab, 'N',          INTEGER32&
      , comment='The number of entities in the bin.')
    yAveCol    = addColumn(tab, 'Y_AVE',      REAL32&
      , comment='The total weight in the bin divided by N.')
    yStdDevCol = addColumn(tab, 'Y_STD_DEV',  REAL32&
      , comment='The standard deviation of Y minus Y_AVE.')
    yUncertCol = addColumn(tab, 'Y_UNCERT',   REAL32&
      , comment='The uncertainty in Y_AVE.')

    xLo     => real32Data(xLoCol)
    xHi     => real32Data(xHiCol)
    n       => int32Data(nCol)
    yAve    => real32Data(yAveCol)
    yStdDev => real32Data(yStdDevCol)
    yUncert => real32Data(yUncertCol)

    xLo  = histo%xLo
    xHi  = histo%xHi
    n    = histo%n
    yAve = histo%yAve
    yStdDev = histo%yStdDev
    yUncert = histo%yUncert

    call release(tab)

  end subroutine saveHistogram

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module histogram_2
