!
! Module: ml_corr
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: A grab-bag of general-purpose routines.
! 
! Comments: TODO:
! 
! $Id: ml_corr.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module ml_corr

  implicit none 

  private 

  public corr_single_band
  public corr_xid_band
  public corr_sum_band
  public corr_ep_from_new
  public corr_ep_from_old

contains 

!------------------------------------------------------------------------
!  subroutine corr_single_band(det_ml_old,det_ml_new,nrows)
subroutine corr_single_band(det_ml_old,det_ml_new,nrows, dataPathName)
!------------------------------------------------------------------------

! input: old single-band DET_ML value.

character(*), intent(in) :: dataPathName
  integer, intent(in) :: nrows
  integer           :: dof(nrows)
  real, intent(in)  :: det_ml_old(nrows)
  real, intent(out) :: det_ml_new(nrows)
  real              :: det_ml_raw_old(nrows)

integer :: irow

  do irow=1,nrows
     dof(irow)=3
  end do
!  call invert_gamma(dof,det_ml_old,det_ml_raw_old,nrows)
!  call forward_gamma(dof,det_ml_raw_old/2.,det_ml_new,nrows) 
call invert_gamma(dof,det_ml_old,det_ml_raw_old,nrows, dataPathName)
call forward_gamma(dof,det_ml_raw_old/2.,det_ml_new,nrows, dataPathName) 

 end subroutine corr_single_band
!------------------------------------------------------------------------
  

!------------------------------------------------------------------------
!  subroutine corr_xid_band(det_ml_old,det_ml_new,nrows)
subroutine corr_xid_band(det_ml_old,det_ml_new,nrows, dataPathName)
!------------------------------------------------------------------------

! input: old XID band DET_ML value.

character(*), intent(in) :: dataPathName
  integer, intent(in) :: nrows
  integer           :: dof(nrows)
  real, intent(in)  :: det_ml_old(nrows)
  real, intent(out) :: det_ml_new(nrows)
  real              :: det_ml_raw_old(nrows)

integer :: irow

  do irow=1,nrows
     dof(irow)=4
  end do
!  call invert_gamma(dof,det_ml_old,det_ml_raw_old,nrows)
!  call forward_gamma(dof,det_ml_raw_old/2.,det_ml_new,nrows) 
call invert_gamma(dof,det_ml_old,det_ml_raw_old,nrows, dataPathName)
call forward_gamma(dof,det_ml_raw_old/2.,det_ml_new,nrows, dataPathName) 
  
 end subroutine corr_xid_band
!------------------------------------------------------------------------


!------------------------------------------------------------------------
!  subroutine corr_sum_band(det_ml_old,det_ml_new,nrows)
subroutine corr_sum_band(det_ml_old,det_ml_new,nrows, dataPathName)
!------------------------------------------------------------------------

! input: old 5-band DET_ML value.

character(*), intent(in) :: dataPathName
  integer, intent(in) :: nrows
  integer           :: dof(nrows)
  real, intent(in)  :: det_ml_old(nrows)
  real, intent(out) :: det_ml_new(nrows)
  real              :: det_ml_raw_old(nrows)

  dof = 7

!  call invert_gamma(dof,det_ml_old,det_ml_raw_old,nrows)
!  call forward_gamma(dof,det_ml_raw_old/2.,det_ml_new,nrows) 
call invert_gamma(dof,det_ml_old,det_ml_raw_old,nrows, dataPathName)
call forward_gamma(dof,det_ml_raw_old/2.,det_ml_new,nrows, dataPathName) 
  
 end subroutine  corr_sum_band
!------------------------------------------------------------------------



!------------------------------------------------------------------------
!  subroutine corr_ep_from_new(pn_det_ml,m1_det_ml,m2_det_ml,ep_det_ml,nrows)
subroutine corr_ep_from_new(pn_det_ml,m1_det_ml,m2_det_ml,ep_det_ml,nrows, dataPathName)
!------------------------------------------------------------------------

! provides new value for the (catalog) column EP_DET_ML
! input: NEW (corrected) values of PN_DET_ML, M1_DET_ML, M2_DET_ML 

character(*), intent(in) :: dataPathName
  integer ::   nrows,irow
  real    ::   pn_det_ml(nrows),m1_det_ml(nrows),m2_det_ml(nrows)
  real, intent(out) ::   ep_det_ml(nrows)    
  real ::      pn_det_ml_raw(nrows),m1_det_ml_raw(nrows),m2_det_ml_raw(nrows)

  real ::      ep_det_ml_raw(nrows)
  integer::    n_detections,dof(nrows),dof7(nrows)

  ! work out the number of detections:
  do irow=1,nrows 
    n_detections=0
    if (pn_det_ml(irow) > 0) then
      n_detections=n_detections+1
    else
      pn_det_ml(irow) = 0.0
    endif
    if (m1_det_ml(irow) > 0) then
      n_detections=n_detections+1
    else
      m1_det_ml(irow) = 0.0
    endif
    if (m2_det_ml(irow) > 0) then
      n_detections=n_detections+1
    else
      m2_det_ml(irow) = 0.0
    endif
    dof(irow)=n_detections*7
    end do
  dof7 = 7

!  call invert_gamma(dof7,pn_det_ml,pn_det_ml_raw,nrows)
!  call invert_gamma(dof7,m1_det_ml,m1_det_ml_raw,nrows)
!  call invert_gamma(dof7,m2_det_ml,m2_det_ml_raw,nrows)
call invert_gamma(dof7,pn_det_ml,pn_det_ml_raw,nrows, dataPathName)
call invert_gamma(dof7,m1_det_ml,m1_det_ml_raw,nrows, dataPathName)
call invert_gamma(dof7,m2_det_ml,m2_det_ml_raw,nrows, dataPathName)

  ep_det_ml_raw = pn_det_ml_raw + m1_det_ml_raw +  m2_det_ml_raw

!  call forward_gamma(dof,ep_det_ml_raw,ep_det_ml,nrows)
call forward_gamma(dof,ep_det_ml_raw,ep_det_ml,nrows, dataPathName)

!  Overwrite this value with a simple sum of the new camera det_mls (RDS)
  ep_det_ml = m1_det_ml + m2_det_ml + pn_det_ml


end subroutine corr_ep_from_new
!------------------------------------------------------------------------ 

!------------------------------------------------------------------------
!  subroutine corr_ep_from_old(pn_det_ml,m1_det_ml,m2_det_ml,ep_det_ml,nrows)
subroutine corr_ep_from_old(pn_det_ml,m1_det_ml,m2_det_ml,ep_det_ml,nrows, dataPathName)
!------------------------------------------------------------------------

! provides new value for the column EP_DET_ML
! input: OLD (uncorrected) values of PN_DET_ML, M1_DET_ML, M2_DET_ML 

character(*), intent(in) :: dataPathName
  integer ::    nrows,irow
  real    ::    pn_det_ml(nrows),m1_det_ml(nrows),m2_det_ml(nrows)
  real, intent(out) ::   ep_det_ml(nrows)    
  real ::      pn_det_ml_raw(nrows),m1_det_ml_raw(nrows),m2_det_ml_raw(nrows)

  real ::      ep_det_ml_raw(nrows)
  integer::    n_detections,dof(nrows),dof7(nrows)

  ! work out the number of detections:
  do irow=1,nrows 
    n_detections=0
    if (pn_det_ml(irow) > 0) then
      n_detections=n_detections+1
    else
      pn_det_ml(irow) = 0.0
    endif
    if (m1_det_ml(irow) > 0) then
      n_detections=n_detections+1
    else
    m1_det_ml(irow) = 0.0
    endif
    if (m2_det_ml(irow) > 0) then
      n_detections=n_detections+1
    else
      m2_det_ml(irow) = 0.0
    endif
    dof(irow)=n_detections*7
  end do  
  dof7 = 7

!  call invert_gamma(dof7,pn_det_ml,pn_det_ml_raw,nrows)
!  call invert_gamma(dof7,m1_det_ml,m1_det_ml_raw,nrows)
!  call invert_gamma(dof7,m2_det_ml,m2_det_ml_raw,nrows)
call invert_gamma(dof7,pn_det_ml,pn_det_ml_raw,nrows, dataPathName)
call invert_gamma(dof7,m1_det_ml,m1_det_ml_raw,nrows, dataPathName)
call invert_gamma(dof7,m2_det_ml,m2_det_ml_raw,nrows, dataPathName)

  ep_det_ml_raw = (pn_det_ml_raw + m1_det_ml_raw +  m2_det_ml_raw)/2.

!  call forward_gamma(dof,ep_det_ml_raw,ep_det_ml,nrows)
call forward_gamma(dof,ep_det_ml_raw,ep_det_ml,nrows, dataPathName)



end subroutine corr_ep_from_old
!------------------------------------------------------------------------ 


 


!-------------------------------------------------------------------- 
!     subroutine invert_gamma(dof,ml_old,ml_raw,nrows)
subroutine invert_gamma(dof,ml_old,ml_raw,nrows, dataPathName)
!-------------------------------------------------------------------- 

character(*), intent(in) :: dataPathName
      integer, intent(in) ::  nrows
      integer, intent(in) :: dof(nrows)
      integer :: irow, j, flag
      real, intent(in)  :: ml_old(nrows)
      real, intent(out) :: ml_raw(nrows)
!      parameter(ntab = 36952)
integer, parameter :: ntab = 36952
      real :: raw(ntab),ml3(ntab),ml4(ntab),ml7(ntab)
      real :: ml14(ntab),ml21(ntab),ml(ntab)
      real :: xa(4),ya(4),ddy 

integer :: nop, i

      nop=34331
!      open(unit=1,file='newtab.dat',status='old')
open(unit=1,file=dataPathName,status='old')
      read(1,*) (raw(i),ml3(i),ml4(i),ml7(i),ml14(i),ml21(i),i=1,nop)
      close(1)

      ml_raw=0.

      do irow=1,nrows
        flag=0
        if (ml_old(irow) > 0) then
          select case (dof(irow))
          case (3)
            if (ml_old(irow) < 4.E-2) then
              ml_raw(irow)=3.85275*ml_old(irow)
              flag=1
            endif 
            ml=ml3
          case (4)
            if (ml_old(irow) < 1.E-2) then
              ml_raw(irow)=16.1328*ml_old(irow)
              flag=1
            endif 
            ml=ml4
          case (7)
            if (ml_old(irow) < 1.E-4) then
              ml_raw(irow)=1928.30*ml_old(irow)
              flag=1
            endif 
            ml=ml7
          case (14)
            if (ml_old(irow) < 5.E-10) then
              ml_raw(irow)=9.364E8*ml_old(irow)
              flag=1
            endif 
            ml=ml14
          case (21)
            if (ml_old(irow) < 2.E-14) then
              ml_raw(irow)=1.19541E13*ml_old(irow)
              flag=1
            endif 
            ml=ml21
          end select 
            call locate(ml,nop,ml_old(irow),j)
          xa(1)=raw(j-1)
          xa(2)=raw(j)
          xa(3)=raw(j+1)
          xa(4)=raw(j+2)
          ya(1)=ml(j-1)
          ya(2)=ml(j)
          ya(3)=ml(j+1)
          ya(4)=ml(j+2)
          if (flag == 0) then
            call polint(ya,xa,4,ml_old(irow),ml_raw(irow),ddy)
          endif
!          write(*,*) "invert:",ml_old(irow),ml_raw(irow)
        endif
      end do

     end subroutine invert_gamma
!-------------------------------------------------------------------- 


!-------------------------------------------------------------------- 
!     subroutine forward_gamma(dof,ml_raw,ml_new,nrows)
subroutine forward_gamma(dof,ml_raw,ml_new,nrows,dataPathName)
!-------------------------------------------------------------------- 

character(*), intent(in) :: dataPathName
      integer, intent(in) ::  nrows
      integer, intent(in) :: dof(nrows)
      integer irow, j, flag
      real, intent(in)  :: ml_raw(nrows)
      real, intent(out) :: ml_new(nrows)
!      parameter (ntab = 36952)
integer, parameter :: ntab = 36952
      real :: raw(ntab),ml3(ntab),ml4(ntab),ml7(ntab)
      real :: ml14(ntab),ml21(ntab),ml(ntab)
      real :: xa(4),ya(4),ddy 

integer :: nop, i

      nop=34331
!      open(unit=1,file='newtab.dat',status='old')
open(unit=1,file=dataPathName,status='old')
      read(1,*) (raw(i),ml3(i),ml4(i),ml7(i),ml14(i),ml21(i),i=1,nop)
      close(1)
    
      ml_new = 0.

      do irow=1,nrows
        flag=0
        if (ml_raw(irow) > 0) then
!          write(*,*) "forward input:", ml_raw(irow)
          select case (dof(irow))
          case (3)
            if (ml_raw(irow) < 0.140) then
              ml_new(irow) = ml_raw(irow)/3.85275
              flag=1
            endif 
            ml=ml3
          case (4)
            if (ml_raw(irow) < 0.140) then
              ml_new(irow) = ml_raw(irow)/16.1328
              flag=1
            endif 
            ml=ml4
          case (7)
            if (ml_raw(irow) < 0.140) then
              ml_new(irow) = ml_raw(irow)/1928.30
              flag=1
            endif 
            ml=ml7
          case (14)
            if (ml_raw(irow) < 0.140) then
              ml_new(irow) = ml_raw(irow)/9.364E8
              flag=1
            endif 
            ml=ml14
          case (21)
            if (ml_raw(irow) < 0.240) then
              ml_new(irow) = ml_raw(irow)/1.1954E13
              flag=1
            endif 
            ml=ml21
          end select 
          call locate(raw,nop,ml_raw(irow),j)
          xa(1)=raw(j-1)
          xa(2)=raw(j)
          xa(3)=raw(j+1)
          xa(4)=raw(j+2)
          ya(1)=ml(j-1)
          ya(2)=ml(j)
          ya(3)=ml(j+1)
          ya(4)=ml(j+2)
          if (flag == 0) then
            call polint(xa,ya,4,ml_raw(irow),ml_new(irow),ddy)
          endif
!          write(*,*) "forward:",ml_raw(irow),ml_new(irow)
        endif
      end do

     end subroutine forward_gamma
!-------------------------------------------------------------------- 





! Numerical recipes stuff:
!--------------------------------------------------------------------


      SUBROUTINE polint(xa,ya,n,x,y,dy)
      INTEGER n,NMAX
      REAL dy,x,y,xa(n),ya(n)
      PARAMETER (NMAX=10)
      INTEGER i,m,ns
      REAL den,dif,dift,ho,hp,w,c(NMAX),d(NMAX)
      ns=1
      dif=abs(x-xa(1))
      do 11 i=1,n
        dift=abs(x-xa(i))
        if (dift.lt.dif) then
          ns=i
          dif=dift
        endif
        c(i)=ya(i)
        d(i)=ya(i)
11    continue
      y=ya(ns)
      ns=ns-1
      do 13 m=1,n-1
        do 12 i=1,n-m
          ho=xa(i)-x
          hp=xa(i+m)-x
          w=c(i+1)-d(i)
!          print *,ho,hp
          den=ho-hp
          if(den.eq.0.d0) write(*,*) 'failure in polint'
          den=w/den
          d(i)=hp*den
          c(i)=ho*den
12      continue
        if (2*ns.lt.n-m)then
          dy=c(ns+1)
        else
          dy=d(ns)
          ns=ns-1
        endif
        y=y+dy
13    continue
      return
!      END
end subroutine polint

      SUBROUTINE locate(xx,n,x,j)
      INTEGER j,n
      REAL x,xx(n)
      INTEGER jl,jm,ju
      jl=0
      ju=n+1
10    if(ju-jl.gt.1)then
        jm=(ju+jl)/2
        if((xx(n).gt.xx(1)).eqv.(x.gt.xx(jm)))then
          jl=jm
        else
          ju=jm
        endif
      goto 10
      endif
      j=jl
      return
!      END

end subroutine locate

!--------------------------------------------------------------------




  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module ml_corr
