!
! Module: test_routines
!
! Authors:  Ian Stewart - SSC/LUX
!
! Comments: TODO:
! 
! $Id: test_routines.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- fred
module test_routines

  use dal

  use constants
  use coordinate	! for skyToCartesian
  use array_utils

  implicit none 

  private 

  public test_doConePlot
  public test_doSrcSkyPlot

  contains

  !---------------------------------------------------------------------
  subroutine test_doConePlot(rays, centralRay, coneAngle)
  !---------------------------------------------------------------------
  ! This tests the tightest-cone finding routines in temp_ssclib.

    implicit none

    real(double), intent(in) :: rays(:,:), centralRay(3), coneAngle

    real :: angle, x, y
    integer ::&
      numRays, rayNum, devNum, i, numCircs = 360, numRadials = 12, pgopen
    real(double) ::&
      refRa, refDec, refMatrix_sky(3,3), srcVector_sky(3),&
      theta(size(rays,1)), phi(size(rays,1)), phiCentral, thetaCentral
    real :: magnifier = 100.0

    !-------------------------------------------------------------------
    ! Start

    numRays = size(rays, 1)

    ! Bring the ray coordinates into the basis of centralRay:

    if (centralRay(1) == 0.0 .and. centralRay(2) == 0.0) then
      refRa = 0.0d0
    else
      refRa = atan2(centralRay(2), centralRay(1))
    end if
    refDec = asin(centralRay(3))

    refMatrix_sky(1, :) = skyToCartesian(refRa, refDec)
    refMatrix_sky(2, :) = skyToCartesian(refRa + PiOverTwo, 0.0d0)
    refMatrix_sky(3, :) = vectorCross(refMatrix_sky(1, :)&
      , refMatrix_sky(2, :))
    refMatrix_sky = transpose(refMatrix_sky)

    do rayNum = 1, numRays
      srcVector_sky = matmul(rays(rayNum,:), refMatrix_sky)
      if (srcVector_sky(3) == 0.0 .and. srcVector_sky(2) == 0.0) then
        phi(rayNum) = 0.0d0
      else
        phi(rayNum) = atan2(srcVector_sky(3), srcVector_sky(2))
      end if
      theta(rayNum) = acos(srcVector_sky(1))
    end do
    srcVector_sky = matmul(centralRay, refMatrix_sky)

    if (srcVector_sky(3) == 0.0 .and. srcVector_sky(2) == 0.0) then
      phiCentral = 0.0d0
    else
      phiCentral = atan2(srcVector_sky(3), srcVector_sky(2))
    end if
    thetaCentral = acos(srcVector_sky(1))

    devNum = pgopen('/xs')
    call pgsvp(0.0, 1.0, 0.0, 1.0)
    call pgwnad(real(-Pi) / magnifier, real(Pi) / magnifier&
      , real(-Pi) / magnifier, real(Pi) / magnifier)

    ! Draw the background:

    call pgsfs(1) ! filled
    call pgsci(8)
    call pgscr(8, 0.3, 0.3, 0.3)
    call pgcirc(0.0, 0.0, real(Pi))
    call pgsfs(2) ! open
    call pgsci(1)
    do i = 1, numCircs
      call pgcirc(0.0, 0.0, real(Pi) * (i / real(numCircs)))
    end do
    do i = 1, numRadials
      angle = real(TwoPi) * (i / real(numRadials))
      x = real(Pi) * cos(angle)
      y = real(Pi) * sin(angle)
      call pgmove(0.0, 0.0)
      call pgdraw(x, y)
    end do

    ! Draw the central cone:

    call pgsci(2)
    call pgcirc(0.0, 0.0, real(coneAngle))
    x = real(thetaCentral) * cos(real(phiCentral))
    y = real(thetaCentral) * sin(real(phiCentral))
    call pgpt1(x, y, 5)

    ! Draw the ray points:

    call pgsci(3)
    do rayNum = 1, numrays
      x = real(theta(rayNum)) * cos(real(phi(rayNum)))
      y = real(theta(rayNum)) * sin(real(phi(rayNum)))
      call pgpt1(x, y, 4)
    end do

    call pgend

  end subroutine test_doConePlot

  !---------------------------------------------------------------------
  subroutine test_doSrcSkyPlot(srcRaDeg, srcDecDeg, centralRaDeg&
  , centralDecDeg, coneAngleArcsec)
  !---------------------------------------------------------------------
  ! This plots the positions of the sources on the sky, centred about centralRay. The sources are plotted using an equal-area radius scaling so that the density ought to be uniform within coneAngle from the centre (and of course zero outside it).

    implicit none

    real(double), intent(in) :: srcRaDeg(:), srcDecDeg(size(srcRaDeg)), centralRaDeg, centralDecDeg, coneAngleArcsec

    integer ::&
      numSrcs, srcNum, devNum, pgopen!, X=1, Y=2, Z=3
    real(double) ::&
      srcRa(size(srcRaDeg)), srcDec(size(srcRaDeg)),&
      rays(size(srcRaDeg),3), r(size(srcRaDeg)),&
      theta(size(srcRaDeg)), phi(size(srcRaDeg)),&
      centralRa, centralDec, coneAngle, refMatrix_sky(3,3), srcVector_sky(3)
!      refRa, refDec,&
!      , phiCentral, thetaCentral
    real :: fieldRadius, magnifier = 160.0, x, y

    !-------------------------------------------------------------------
    ! Start

    numSrcs = size(srcRaDeg)

    srcRa = srcRaDeg * PiOn180
    srcDec = srcDecDeg * PiOn180
    rays(:,1) = cos(srcDec) * cos(srcRa)
    rays(:,2) = cos(srcDec) * sin(srcRa)
    rays(:,3) = sin(srcDec)

    centralRa = centralRaDeg * PiOn180
    centralDec = centralDecDeg * PiOn180
    coneAngle = coneAngleArcsec / 3600.0d0 * PiOn180

    ! Bring the source coordinates into the basis of centralRay:

    refMatrix_sky(1, :) = skyToCartesian(centralRa, centralDec)
    refMatrix_sky(2, :) = skyToCartesian(centralRa + PiOverTwo, 0.0d0)
    refMatrix_sky(3, :) = vectorCross(refMatrix_sky(1, :)&
      , refMatrix_sky(2, :))
    refMatrix_sky = transpose(refMatrix_sky)

    do srcNum = 1, numSrcs
      srcVector_sky = matmul(rays(srcNum,:), refMatrix_sky)
      if (srcVector_sky(3) == 0.0 .and. srcVector_sky(2) == 0.0) then
        phi(srcNum) = 0.0d0
      else
        phi(srcNum) = atan2(srcVector_sky(3), srcVector_sky(2))
      end if
      theta(srcNum) = acos(srcVector_sky(1))
      r(srcNum) = sin(theta(srcNum) / 2.0d0)
    end do

    devNum = pgopen('/xs')
    call pgsvp(0.0, 1.0, 0.0, 1.0)
    fieldRadius = real(sin(PiOn2)) / magnifier
    call pgwnad(-fieldRadius, fieldRadius, -fieldRadius, fieldRadius)

    ! Draw the background:

    call pgsfs(1) ! filled
    call pgsci(8)
    call pgscr(8, 0.3, 0.3, 0.3)
    call pgrect(-fieldRadius, fieldRadius, -fieldRadius, fieldRadius)
    call pgsfs(2) ! open

    ! Draw the central cone:

    call pgsci(2)
    call pgcirc(0.0, 0.0, real(sin(coneAngle / 2.0d0)))

    ! Draw the source points:

    call pgsci(1)
    do srcNum = 1, numSrcs
      x = real(r(srcNum)) * cos(real(phi(srcNum)))
      y = real(r(srcNum)) * sin(real(phi(srcNum)))
      call pgpt1(x, y, -1)
    end do

    call pgend

  end subroutine test_doSrcSkyPlot

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module test_routines
